freeze;

/////////////////////////////////////////////////////////////////////////
// strings.m
/////////////////////////////////////////////////////////////////////////
// $Revision: 41104 $
// $Date: 2012-11-27 22:49:11 +1100 (Tue, 27 Nov 2012) $
// $LastChangedBy: kasprzyk $
/////////////////////////////////////////////////////////////////////////
// Author: Alexander Kasprzyk
/////////////////////////////////////////////////////////////////////////
// A collection of useful string utilities.
/////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// Removes all characters appearing in the set "remv" from the start and
// end of the string S. Returns the trimmed string.
function trim(S,remv)
    startidx:=1;
    while startidx le #S and S[startidx] in remv do
        startidx+:=1;
    end while;
    if startidx gt #S then return ""; end if;
    endidx:=#S;
    while endidx ge startidx and S[endidx] in remv do
        endidx-:=1;
    end while;
    return S[startidx..endidx];
end function;

/////////////////////////////////////////////////////////////////////////
// is_string_integer
/////////////////////////////////////////////////////////////////////////
// Input: A string str and a starting index idx.
// Output: Attempts to convert the string, starting at index idx, to an
//      integer. The first return value is true iff the entire string
//      (starting at idx) was successfully parsed. The second return value
//      is the integer parsed so far. The third return value is the index
//      of the first non-parsable character in the string. (If the string
//      was successfully parsed through to the end, then the returned index
//      will be zero.)
/////////////////////////////////////////////////////////////////////////
// Note: This only handles base 10 notation. Perhaps somebody would be
//      interested in extending it to recognising 0x..., etc.?
//      Only allows one '-' prefix before the number. Does not allow a '+'.
//      Does not allow space or ',' within the number.
//      A decimal place '.', even if followed by only zeros, is not allowed.
//      Scientific notation such as "10E3" for 10^3 is not allowed.
/////////////////////////////////////////////////////////////////////////
/*
    is_string_integer("  -    10  ",1);    // true -10 0
    is_string_integer("     ",1);          // false 0 0
    is_string_integer("+10",1);            // false 0 1
    is_string_integer("--10",1);           // false 0 1
    is_string_integer("13,200",1);         // false 13 3
    is_string_integer("13 200",1);         // false 13 3
    is_string_integer("10.0",1);           // false 10 3
*/
function is_string_integer(str,startidx)
    idx:=IndexOfNonWhiteSpace(str,startidx);
    if idx eq 0 then return false,0,0; end if;
    numbers:=["0","1","2","3","4","5","6","7","8","9"];
    // Is there a negative sign?
    negative:=false;
    if str[idx] eq "-" then
        negative:=true;
        idx:=IndexOfNonWhiteSpace(str,idx+1);
    end if;
    // The next character had better be a number
    if idx eq 0 or not str[idx] in numbers then
        return false,0,startidx;
    end if;
    // Read the numbers
    val:=0;
    while idx le #str do
        num:=Index(numbers,str[idx]) - 1;
        if num eq -1 then
            break;
        end if;
        val *:= 10;
        val +:= num;
        idx +:= 1;
    end while;
    // Apply the negation
    if negative then val *:= -1; end if;
    // Check that what follows is white-space only
    if IndexOfNonWhiteSpace(str,idx) eq 0 then
        return true,val,0;
    else
        return false,val,idx;
    end if;
end function;

/////////////////////////////////////////////////////////////////////////
// is_string_rational
/////////////////////////////////////////////////////////////////////////
// Input: A string and an index in that string.
// Output: Attempts to convert the string, starting at index idx, to an
//      rational. The first return value is true iff the entire string
//      (starting at idx) was successfully parsed. The second return value
//      is the rational parsed so far. The third return value is the index
//      of the first non-parsable character in the string. (If the string
//      was successfully parsed through to the end, then the returned index
//      will be zero.)
/////////////////////////////////////////////////////////////////////////
/*
    is_string_rational("   ",1);                        // false 0 0
    is_string_rational("1/2",1);                        // true 1/2 0
    is_string_rational("       1  / 2    ",1);          // true 1/2 0
    is_string_rational("       1  / 2  ,  ",1);         // false 1/2 14
    is_string_rational("       1  / 2      ,  ",1);     // false 1/2 14
    is_string_rational("       1  /       ,  ",1);      // false 1 9
    is_string_rational("       1,  /       ,  ",1);     // false 1 9
*/
function is_string_rational(str,startidx)
    // First read in the numerator
    bool,num,idx:=is_string_integer(str,startidx);
    if bool then
        return true,Rationals()!num,0;
    end if;
    if idx eq startidx or idx eq 0 then
        return false,Rationals()!0,idx;
    end if;
    // Is the next character a '/'?
    newidx:=IndexOfNonWhiteSpace(str,idx);
    if newidx eq 0 then
        return true,Rationals()!num,0;
    end if;
    if not str[newidx] eq "/" or #str eq newidx then
        return false,Rationals()!num,idx;
    end if;
    newidx+:=1;
    // Read in the denominator
    bool,den,idx2:=is_string_integer(str,newidx);
    if den eq 0 then
        return false,Rationals()!num,idx;
    end if;
    if bool then
        return true,Rationals()!(num/den),0;
    end if;
    if idx2 eq newidx then
        return false,Rationals()!num,idx;
    end if;
    return false,Rationals()!(num/den),idx2;
end function;

/////////////////////////////////////////////////////////////////////////
// Intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic StripWhiteSpace( S::MonStgElt ) -> MonStgElt
{Removes all white space from the string}
    white:={" ","\n","\r","\t"};
    newS:="";
    while #S gt 0 do
        idxs:=[Integers() | Index(S,s) : s in white];
        idxs:=[Integers() | i : i in idxs | i ne 0];
        if #idxs eq 0 then
            newS cat:= S;
            S:="";
        else
            idx:=Minimum(idxs);
            newS cat:= S[1..idx - 1];
            S:=S[idx + 1..#S];
        end if;
    end while;
    return newS;
end intrinsic;

intrinsic LTrim( S::MonStgElt ) -> MonStgElt
{Removes all white space from the start of the string}
    white:={" ","\n","\r","\t"};
    startidx:=1;
    while startidx le #S and S[startidx] in white do
        startidx+:=1;
    end while;
    return S[startidx..#S];
end intrinsic;

intrinsic RTrim( S::MonStgElt ) -> MonStgElt
{Removes all white space from the end of the string}
    white:={" ","\n","\r","\t"};
    endidx:=#S;
    while endidx ge 1 and S[endidx] in white do
        endidx-:=1;
    end while;
    return S[1..endidx];
end intrinsic;

intrinsic Trim( S::MonStgElt ) -> MonStgElt
{Removes all white space from the start and end of the string}
    return trim(S,{" ","\n","\r","\t"});
end intrinsic;

intrinsic Trim( S::MonStgElt, R::MonStgElt ) -> MonStgElt
{Removes all characters occurring in the string R from the start and end of the string S}
    if #R eq 0 then return S; end if;
    return trim(S,{R[i] : i in [1..#R]});
end intrinsic;

intrinsic CollateWhiteSpace( S::MonStgElt ) -> MonStgElt
{Collects repeated instances of white space into a single " ". Also trims any white space from the start or end of the string.}
    white:={" ","\n","\r","\t"};
    newstr:="";
    inwhite:=true;
    start:=true;
    for i in [1..#S] do
        c:=S[i];
        if inwhite then
            if not c in white then
                inwhite:=false;
                if start then
                    start:=false;
                    newstr cat:= c;
                else
                    newstr cat:= " " cat c;
                end if;
            end if;
        else
            if c in white then
                inwhite:=true;
            else
                newstr cat:= c;
            end if;
        end if;
    end for;
    return newstr;
end intrinsic;

intrinsic IndexOfNonWhiteSpace( S::MonStgElt ) -> RngIntElt
{}
    return IndexOfNonWhiteSpace(S,1);
end intrinsic;

intrinsic IndexOfNonWhiteSpace( S::MonStgElt, k::RngIntElt ) -> RngIntElt
{The index of the first occurance of a non-white space character (or 0 if no non-white space is found)}
    white:={" ","\n","\r","\t"};
    while k le #S and S[k] in white do
        k+:=1;
    end while;
    return k le #S select k else 0;
end intrinsic;

intrinsic IndexOfFirstWhiteSpace( S::MonStgElt ) -> RngIntElt
{}
    return IndexOfFirstWhiteSpace(S,1);
end intrinsic;

intrinsic IndexOfFirstWhiteSpace( S::MonStgElt, k::RngIntElt ) -> RngIntElt
{The index of the first occurance of white space in the string (or 0 if no white space is found)}
    white:={" ","\n","\r","\t"};
    while k le #S and not S[k] in white do
        k+:=1;
    end while;
    return k le #S select k else 0;
end intrinsic;

intrinsic Index( S::MonStgElt, T::MonStgElt, k::RngIntElt ) -> RngIntElt
{The first occurance of the substring T in S, with offset k (or 0 if T is not found)}
    p:=Index(S[k..#S],T);
    return p eq 0 select 0 else p+k-1;
end intrinsic;

intrinsic Position( S::MonStgElt, T::MonStgElt, k::RngIntElt ) -> RngIntElt
{The first occurance of the substring T in S, with offset k (or 0 if T is not found)}
    p:=Position(S[k..#S],T);
    return p eq 0 select 0 else p+k-1;
end intrinsic;

intrinsic SubstituteString( S::MonStgElt, T::MonStgElt, R::MonStgElt ) -> MonStgElt
{The string formed by replacing every occurance of T in the string S with R}
    idx:=Position(S,T);
    while idx gt 0 do
        S:=S[1..idx-1] cat R cat S[idx+#T..#S];
        idx:=Position(S,T,idx+#R);
    end while;
    return S;
end intrinsic;

intrinsic StringToLower( S::MonStgElt ) -> MonStgElt
{Convert all characters to lower-case}
    src:="ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    dst:="abcdefghijklmnopqrstuvwxyz";
    newS:="";
    for i in [1..#S] do
        idx:=Index(src,S[i]);
        newS cat:= idx eq 0 select S[i] else dst[idx];
    end for;
    return newS;
end intrinsic;

intrinsic StringToUpper( S::MonStgElt ) -> MonStgElt
{Convert all characters to upper-case}
    src:="abcdefghijklmnopqrstuvwxyz";
    dst:="ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    newS:="";
    for i in [1..#S] do
        idx:=Index(src,S[i]);
        newS cat:= idx eq 0 select S[i] else dst[idx];
    end for;
    return newS;
end intrinsic;

intrinsic Join( S::SeqEnum[MonStgElt], x::MonStgElt ) -> MonStgElt
{Concatenates the strings in S, using deliminator x}
    if #S eq 0 then return ""; end if;
    s:=S[1];
    for i in [2..#S] do
        s cat:= x cat S[i];
    end for;
    return s;
end intrinsic;

intrinsic Reverse( S::MonStgElt ) -> MonStgElt
{Reverse the order of the characters in S}
    newstr:="";
    for i in [#S..1 by -1] do
        newstr cat:= S[i];
    end for;
    return newstr;
end intrinsic;

intrinsic WordWrap( S::MonStgElt, k::RngIntElt : tolerance:=8 ) -> MonStgElt
{Splits the string S into lines of length at most k}
    // Sanity check
    require k gt 1: "Argument 2 must be a positive integer";
    require Type(tolerance) eq RngIntElt and tolerance ge 0:
        "Parameter 'tolerance' must be a non-negative integer";
    // Start building the string
    lastspace:=0;
    linelength:=0;
    newstr:="";
    while #S gt 0 do
        l:=Minimum(k,#S);
        chunk:=Reverse(S[1..l]);
        S:=S[l + 1..#S];
        idx:=Index(chunk,"\n");
        if idx eq 0 then
            idx:=Index(chunk," ");
            if idx eq 0 or idx gt tolerance then
                if #S eq 0 then
                    newstr cat:= Reverse(chunk);
                else
                    newstr cat:= Reverse(chunk) cat "\n";
                end if;
            else
                S:=Reverse(chunk[1..idx - 1]) cat S;
                if #S eq 0 then
                    newstr cat:= Reverse(chunk[idx..#chunk]);
                else
                    newstr cat:= Reverse(chunk[idx..#chunk]) cat "\n";
                end if;
            end if;
        else
            S:=Reverse(chunk[1..idx - 1]) cat S;
            newstr cat:= Reverse(chunk[idx..#chunk]);
        end if;
    end while;
    // Return the string
    return newstr;
end intrinsic;

intrinsic StringToRational( S::MonStgElt ) -> FldRatElt
{The rational corresponding to S}
    bool,rat,idx:=is_string_rational(S,1);
    require bool and idx eq 0:
        Sprintf("String \"%o\" does not represent a rational number in base 10",S);
    return rat;
end intrinsic;
