freeze;

/*************************************************************************
    
  Script for running all available machinery for MW and Sha 
  for elliptic curves over Q or over number fields
       
  Steve Donnelly, last modified October 2011

*************************************************************************/

declare verbose MWSha, 1;

debug := false;

// ALWAYS call this procedure before return

procedure Done(~done,rl,ru,ranal,~MWgens,sha,~sha1,RankOnly,Sha,verbosity 
              : final:=false)

  assert rl ge #MWgens and ru ge #MWgens;
  if ranal ge 0 then  // known
    assert rl le ranal and #MWgens le ranal and ru ge ranal; 
  end if;
  if debug then
    assert forall{P : P in MWgens | Order(P) eq 0};
  end if;
  
  done_rank := rl eq ru;
  done_gens := RankOnly or #MWgens eq ru;
  done_sha := {n : n in {2,4,8,3,9} | Sha mod n eq 0} subset Keys(sha);
  done := done_rank and done_gens and done_sha; 

  if done or final then
    if not RankOnly and #MWgens gt 0 and Type(Ring(Universe(MWgens))) eq FldRat then
      // TO DO: work out the correct bound; FldNum case
      MWgens := Saturation(MWgens, 50 : TorsionFree); 
    end if;
    // sha1 = old format for sha
    sha1 := [<n,sha[n]> : n in Sort(Setseq(Keys(sha)))];
    vprintf MWSha: "\n"; 
    SetVerbose("MWSha", verbosity);   // restore original value
  end if;  
end procedure;

procedure refine_sha_bounds(~sha, n, l, u)
  if IsDefined(sha, n) then
    l0, u0 := Explode(sha[n]);
    sha[n] := [Max(l,l0), Min(u,u0)];
  else
    sha[n] := [l, u];
  end if;
end procedure;

function get_sha_bounds(sha, n)
  bool, bounds := IsDefined(sha, n);
  if bool then
    return Explode(bounds);
  else 
    return 0, Infinity();
  end if;
end function;

procedure print_sha_bounds(sha)
  L := "";  
  U := "";

  function plus(s)
    return #s eq 0 select "" else " + ";
  end function;

  bool, bounds8 := IsDefined(sha, 8);
  if bool then
    if bounds8[1] gt 0 then
      L *:= plus(L) * "(Z/8)^"*IntegerToString(bounds8[1]); end if;
    if bounds8[2] gt 0 then
      U *:= plus(U) * "(Z/8)^"*IntegerToString(bounds8[2]); end if;
  else 
    bounds8 := [0,0];
  end if;

  bool, bounds4 := IsDefined(sha, 4);
  if bool then
    b1 := bounds4[1] - bounds8[1];
    b2 := bounds4[2] - bounds8[2];
    if b1 gt 0 then
      L *:= plus(L) * "(Z/4)^"*IntegerToString(b1); end if;
    if b2 gt 0 then
      U *:= plus(U) * "(Z/4)^"*IntegerToString(b2); end if;
  else 
    bounds4 := [0,0];
  end if;

  bool, bounds2 := IsDefined(sha, 2);
  if bool then
    b1 := bounds2[1] - bounds4[1];
    b2 := bounds2[2] - bounds4[2];
    if b1 gt 0 then
      L *:= plus(L) * "(Z/2)^"*IntegerToString(b1); end if;
    if b2 gt 0 then
      U *:= plus(U) * "(Z/2)^"*IntegerToString(b2); end if;
  end if;

  bool, bounds9 := IsDefined(sha, 9);
  if bool then
    if bounds9[1] gt 0 then
      L *:= plus(L) * "(Z/9)^"*IntegerToString(bounds9[1]); end if;
    if bounds9[2] gt 0 then
      U *:= plus(U) * "(Z/9)^"*IntegerToString(bounds9[2]); end if;
  else 
    bounds9 := [0,0];
  end if;

  bool, bounds3 := IsDefined(sha, 3);
  if bool then
    b1 := bounds3[1] - bounds9[1];
    b2 := bounds3[2] - bounds9[2];
    if b1 gt 0 then
      L *:= plus(L) * "(Z/3)^"*IntegerToString(b1); end if;
    if b2 gt 0 then
      U *:= plus(U) * "(Z/3)^"*IntegerToString(b2); end if;
  end if;

  n := LCM(Keys(sha)); assert 72 mod n eq 0;

  if #U eq 0 then 
    vprintf MWSha: "    Sha(E)[%o] is trivial\n", n;
  elif #L eq 0 then
    vprintf MWSha: "    Sha(E)[%o] <= %o\n", n, U;
  else
    vprintf MWSha: "    %o <= Sha(E)[%o] <= %o\n", L, n, U;
  end if;
end procedure;

intrinsic DescentInformation(E::CrvEll : 
          RankOnly:=false, ShaInfo:=false, Sha:=1, Effort:=1, Silent:=false) 
       -> SeqEnum[RngIntElt], SeqEnum[PtEll], GrpAb
{Same as MordellWeilShaInformation}

  require t eq FldRat or ISA(t, FldAlg) where t is Type(BaseRing(E)) :
         "The curve must be over the rationals or a number field";

  require Type(RankOnly) eq BoolElt : "Optional parameter 'RankOnly' should be a boolean";
  require Type(ShaInfo) eq BoolElt : "Optional parameter 'ShaInfo' should be a boolean";
  require Type(Sha) eq RngIntElt : "Optional parameter 'Sha' should be an integer";

  return MordellWeilShaInformation(E : 
         Effort:=Effort, RankOnly:=RankOnly, ShaInfo:=ShaInfo, Sha:=Sha, Silent:=Silent);
end intrinsic;

intrinsic MordellWeilShaInformation(E::CrvEll : 
          RankOnly:=false, ShaInfo:=false, Sha:=1, Effort:=1, Silent:=false) 
       -> SeqEnum[RngIntElt], SeqEnum[PtEll], GrpAb
{Print and return all the information about the Mordell-Weil group and the 
 Tate-Shafarevich group of E that can be obtained from Magma routines.
 Returns: a sequence [rank_lower_bound, rank_upper_bound], a sequence
 of the generators modulo torsion that are known, and a sequence giving
 lower and upper bounds on Sha[n] for integers n for which bounds are known.}

  type := Type(BaseRing(E));
  require type eq FldRat or ISA(type, FldAlg) : 
         "The curve must be over the rationals or a number field";

  require Type(RankOnly) eq BoolElt : "Optional parameter 'RankOnly' should be a boolean";
  require Type(ShaInfo) eq BoolElt : "Optional parameter 'ShaInfo' should be a boolean";
  require Type(Sha) eq RngIntElt : "Optional parameter 'Sha' should be an integer";

  require Effort ge 1 : "Optional paramter 'Effort' should be a positive integer";
  // (Effort not used over FldRat, currently)

  // default behaviour is to print information (the main purpose of the function!)
  verbosity := GetVerbose("MWSha");
  SetVerbose("MWSha", Silent select 0 else 1);

  if ShaInfo then
    Sha := 0; // all possible Sha info
  end if;

  vprintf MWSha: "\n";

  rl := 0;
  ru := Infinity();  
  MWgens := [E(BaseRing(E)) | ];
  // sha stores known bounds on the n-rank of sha, where n is a prime power
  // sha[n] = [lower bound, upper bound]
  sha := AssociativeArray(Integers()); 
  ranal := -1;  // means unknown

  tors := TorsionSubgroup(E);  
  t2 := Ilog(2, #TwoTorsionSubgroup(E));
  if #Invariants(tors) eq 0 then
    tors_string := "is trivial";
  elif #Invariants(tors) eq 1 then
    tors_string := "= Z/"*IntegerToString(Invariants(tors)[1]);
  else 
    tors_string := "= Z/"*IntegerToString(Invariants(tors)[1])*
                  " x Z/"*IntegerToString(Invariants(tors)[2]);
  end if;
  vprintf MWSha: "Torsion Subgroup %o\n", tors_string;

  // AnalyticRank 
  // TO DO: can we increase this to 10^12 ?
  // TO DO: is anything proved for FldNum ?
  if type eq FldRat and Conductor(E) le 10^10 then
    ranal := AnalyticRank(E);
    vprintf MWSha: "Analytic rank = %o\n", ranal;
    if ranal le 1 then
      vprintf MWSha: "     ==> Rank(E) = %o\n", ranal;
      rl := ranal; ru := ranal;
    end if;

    Done(~done,rl,ru,ranal,~MWgens,sha,~sha1,RankOnly,Sha,verbosity);
    if done then return [rl,ru], MWgens, sha1; end if; 
  end if;
  
  // TWO DESCENT
  // TO DO: get TwoCover's only as needed
  if type eq FldRat then
    S2, S2map := TwoSelmerGroup(E : RemoveTorsion);
    S2elts := [s : s in S2 | not IsIdentity(s)];
    TD := [];
    TDmaps := [* *];
    for s in S2elts do 
      C2, C2toE := TwoCover(s @@ S2map : E:=E);
      Append(~TD, C2);
      Append(~TDmaps, C2toE); 
    end for;
  else 
    // TO DO: TwoCover
    TD, TDmaps, TDsel := TwoDescent(E : RemoveTorsion);
    S2 := Domain(TDsel);
    S2elts := [C2@@TDsel : C2 in TD];
  end if;
  ruTD := Ilog(2, 1+#TD);  assert 2^ruTD eq 1+#TD;
  ru := Min(ru, ruTD);
  refine_sha_bounds(~sha, 2, ruTD-ru, ruTD-rl);
  vprintf MWSha: "The 2-Selmer group has rank %o\n", ruTD+t2;

  procedure search_on_TDs( TD, ~pts_on_TDs, ~MWgens, ~rl, ru, Bs, ~usedB : TDsubset:=0);
    inds := (TDsubset cmpeq 0) select [1..#TD]
                                else  [i : i in [1..#TD] | TD[i] in TDsubset];
    for B in Bs do 
      usedB := B;
      n := #MWgens;
      for i in inds do 
        if IsEmpty(pts_on_TDs[i]) then
          pts := type eq FldRat select Points(TD[i] : Bound:=B)
                                 else  Points(TD[i] : Bound:=B, Max:=1);
          if not IsEmpty(pts) then
            pt := Rep(pts);
            pts_on_TDs[i] := [pt];
            MWgens := IndependentGenerators(MWgens cat [pt@TDmaps[i]]);  
            rl := Max(rl, #MWgens);
            if #MWgens gt n then
              n := #MWgens;
              vprintf MWSha: (#MWgens eq 1) 
                     select "Found a point of infinite order.\n"
                      else  "Found %o independent points.\n", #MWgens;
            end if;
            if #MWgens eq ru then break B; end if;
          end if;
        end if;
      end for;
    end for;
  end procedure;

  // quick search for points on the TDs
  pts_on_TDs := [* [] : i in TD *];
  TDsearch := rl ne ru or 
              not RankOnly and ru gt 0 or 
              Sha mod 2 eq 0 and ru gt 0 and rl lt ruTD;
  if TDsearch then
    Bs := (type eq FldRat) 
             select [10^2, 10^3, 10^4] 
              else  [2^4];
    search_on_TDs( TD, ~pts_on_TDs, ~MWgens, ~rl, ru, Bs, ~B);
  end if;
  refine_sha_bounds(~sha, 2, ruTD-ru, ruTD-rl); 
  vprintf MWSha: "After 2-descent: \n    %o <= Rank(E) <= %o\n", rl,ru;
  print_sha_bounds(sha);
  if TDsearch then
    vprintf MWSha: "(Searched up to height %o on the 2-coverings.)\n", B;
  end if;
  
  Done(~done,rl,ru,ranal,~MWgens,sha,~sha1,RankOnly,Sha,verbosity);
  if done then return [rl,ru], MWgens, sha1; end if; 

  // TO DO: make full use of pts_on_TDs

  // CASSELS-TATE (quicker than FourDescent!)
  // Do this if (i) don't know rank 
  //         or (ii) need more MWgens and don't know that all TDs lift 
  //         or (iii) need sha and don't know that all TDs lift
  // Note: we assume all TDs lift when ranal = ruTD, however we check this is true when we do 4-descent.
  //   (ie we use this assumption ONLY to save time now, then prove it before using it in the results.)
  if ranal ne -1 then
    rl_conjectural := ranal;
  else
    rl_conjectural := rl;
    if IsOdd(rl_conjectural - ruTD) then
      rl_conjectural +:= 1;
    end if;
  end if;
  if (rl_conjectural lt ruTD) and  
     (rl lt ru or (not RankOnly and #MWgens lt ru) or IsEven(Sha))
  then
    n := Ngens(S2);
    inds := [Index(S2elts, S2.i) : i in [1..n]];
    CTmat := ZeroMatrix(GF(2), n, n);
    for i in [1..n], j in [1..i-1] do 
      ii := inds[i];
      jj := inds[j];
      if IsEmpty(pts_on_TDs[ii]) and IsEmpty(pts_on_TDs[jj]) then
        CTij := CasselsTatePairing(TD[ii], TD[jj]);
        CTmat[i,j] := CTij;  
        CTmat[j,i] := CTij;
      end if;
    end for; 
    vprintf MWSha: "The Cassels-Tate pairing on Sel(2,E)/E[2] is\n";
    IndentPush();
    vprintf MWSha: "%o\n", CTmat;
    IndentPop();
    error if not IsEven(Rank(CTmat)), "CasselsTatePairing failed the parity check";

    kernel := [v : v in Kernel(CTmat) | not IsZero(v)];
    kernel_inds := [Index(S2elts, S2!Eltseq(v)) : v in kernel];
    S2_4 := [S2elts[ii] : ii in kernel_inds]; // used in 8-descent
    TDCT := [TD[ii] : ii in kernel_inds];
    ruTDCT := Ilog(2, 1+#kernel);  assert 2^ruTDCT eq 1+#kernel;
    ru := Min(ru, ruTDCT);
    refine_sha_bounds(~sha, 2, ruTD-ru, ruTD-rl); 
    refine_sha_bounds(~sha, 4, ruTDCT-ru, ruTDCT-rl); 
    vprintf MWSha: "After using Cassels-Tate: \n    %o <= Rank(E) <= %o\n", rl,ru;
    print_sha_bounds(sha);

    Done(~done,rl,ru,ranal,~MWgens,sha,~sha1,RankOnly,Sha,verbosity);
    if done then return [rl,ru], MWgens, sha1; end if; 
  end if; 
 
  if type ne FldRat then
    // long search for points on the TDs (there's nothing else to do)
    TDsearch := rl ne ru or 
                not RankOnly and ru gt #MWgens or 
                Sha mod 2 eq 0 and ruTDCT gt #MWgens;
    if TDsearch then
      // choose bounds B such that B ^ (deg + 0.5) <= 10^6*Effort
      deg := Degree(BaseRing(E));
      Bmax := Round( (10^6*Effort) ^ (1/(deg+0.5)) );
      b := deg eq 2 select 4 else 2;
      Bs := Reverse([Bmax div b^i : i in [0..Ilog(b,Bmax)] | Bmax div b^i ge 2^5]);

      search_on_TDs( TD, ~pts_on_TDs, ~MWgens, ~rl, ru, Bs, ~B :
                     TDsubset := assigned TDCT select TDCT else 0);
      refine_sha_bounds(~sha, 2, ruTD-ru, ruTD-rl); 
      vprintf MWSha: "After more searching:\n    %o <= Rank(E) <= %o\n", rl,ru;
      print_sha_bounds(sha);
      vprintf MWSha: "(Searched up to height %o on the 2-coverings.)\n", B;
    end if;
    
    Done(~done,rl,ru,ranal,~MWgens,sha,~sha1,RankOnly,Sha,verbosity);
    if done then return [rl,ru], MWgens, sha1; end if; 
  end if;

if type eq FldRat then

  // FOUR DESCENT
  if not assigned TDCT then  // we skipped Cassels-Tate because all TD's will lift
    assert rl eq ruTD or rl_conjectural eq ruTD;  // double check
    S2_4 := S2elts;
    TDCT := TD; 
    ruTDCT := ruTD; 
  end if;
  doFD := #MWgens lt ru or (Sha mod 8 eq 0 and ruTDCT ne ru);
  use_heegner := Conductor(E) le 5*10^6 and ranal eq 1;

  if doFD then
    // TO DO: omit 4-covers that correspond to known points
    FDs := [FourDescent(C2 : RemoveTorsion) : C2 in TDCT];
    assert forall{FD: FD in FDs | not IsEmpty(FD)}; 
    // we MUST check this, otherwise our answers are only conjectural.
    // (We used ranal above to decide whether to skip Cassels-Tate.) 

    FDmaps := [];
    for i := 1 to #TDCT do 
      Append(~FDmaps, [* *]);
      for j := 1 to #FDs[i] do 
        _,C4toE := AssociatedEllipticCurve(FDs[i][j] : E:=E);
        Append(~FDmaps[i], C4toE);
      end for;
    end for;

    pts_on_FDs := [*[* [] : j in [1..#FDs[i]]*] : i in [1..#FDs]*];  // store points found

   if #MWgens lt ru then
    // Search for points on 4-coverings
    // TO DO: where appropriate, try three descent before searching with large B 
    MW_rank_known := #MWgens;
    for b := 1 to (use_heegner select 6 else 7) do 
      B := 10^b;
      for i := 1 to #TDCT do 
        for j := 1 to #FDs[i] do
          if IsEmpty(pts_on_FDs[i][j]) then
            pts := PointsQI(FDs[i][j], B : OnlyOne);
            if not IsEmpty(pts) then
              pts_on_FDs[i][j] := pts;
              MWgens := IndependentGenerators(MWgens cat [pts[1]@FDmaps[i][j]]);  
              assert #MWgens ge rl; // since rl comes from #MWgens or from ranal = 1
              rl := #MWgens;
              if #MWgens gt MW_rank_known then
                MW_rank_known := #MWgens;
                vprintf MWSha: (#MWgens eq 1) 
                       select "Found a point of infinite order.\n"
                        else  "Found %o independent points.\n", #MWgens;
              end if;
              if #MWgens eq ru then break b; end if;
            end if;
          end if;
        end for;
      end for; 
    end for;

    refine_sha_bounds(~sha, 2, ruTD-ru, ruTD-rl);  // rl may have increased during point-searching
    refine_sha_bounds(~sha, 4, ruTDCT-ru, ruTDCT-rl); 
    vprintf MWSha: "After 4-descent: \n    %o <= Rank(E) <= %o\n", rl,ru;
    print_sha_bounds(sha);
    vprintf MWSha: "(Searched up to height 10^%o on the 4-coverings.)\n", Ilog(10,B);
   end if;

    Done(~done,rl,ru,ranal,~MWgens,sha,~sha1,RankOnly,Sha,verbosity);
    if done then return [rl,ru], MWgens, sha1; end if; 
  end if; 

  // HEEGNER
//use_heegner := false;
  if use_heegner and #MWgens eq 0 then
    // First guess whether using covers is likely to help
    // (Note: the non-cover case is faster code!)
    // TO DO: also use HeegnerIndex for the number of cosets (though occasionally,
    // if the twisting discriminant is large, this may actually be time-dominant)
    cover := 1;
    terms1 := HeegnerPointNumberOfTerms(E);
    if terms1 gt 10^7 then 
      terms2 := (#TDCT eq 1) 
                select HeegnerPointNumberOfTerms(E : Cover:=TDCT[1])
                 else  Infinity();
      terms4 := (#FDs eq 1 and #FDs[1] eq 1) 
                select HeegnerPointNumberOfTerms(E : Cover:=FDs[1][1])
                 else  Infinity();
      if Min(terms2, terms4) lt terms1 div 100 then 
        cover := (terms2 lt terms4) select 2 else 4;
      end if;
    end if;
    vprintf MWSha: "Finding a Heegner point"; 
    if cover gt 1 then
      vprintf MWSha: " (using a %o-cover)", cover;
    end if;
    vprintf MWSha: "\n";

    case cover:
      when 1: bool, pt := HeegnerPoint(E);
      when 2: bool, pt := HeegnerPoint(E : Cover:=TDCT[1]);
      when 4: bool, pt := HeegnerPoint(E : Cover:=FDs[1][1], NaiveSearch:=10);
    end case;

    if bool then 
      vprint MWSha: "Found Heegner point of height", Height(pt);
      MWgens := [P : P in Saturation([pt], 10 : TorsionFree)]; 
    else 
      vprint MWSha: "Failed to find a Heegner point!";
    end if;
        
    Done(~done,rl,ru,ranal,~MWgens,sha,~sha1,RankOnly,Sha,verbosity);
    if done then return [rl,ru], MWgens, sha1; end if; 
  end if;
 
  // EIGHT DESCENT
  MW_rank_known := #MWgens;
  if MW_rank_known lt ru or (Sha mod 8 eq 0 and ruTDCT ne ru) then
    
    // Do all 4-coverings where we didn't already find points.
    // (Note: it's not advantageous to quotient by the known MW group;
    //  must still try all candidates because only one candidate has
    //  the point with small height.  TO DO: should mod out torsion.)
    EDs        := [ [] : i in [1..#FDs] ];
    EDmaps     := [ [] : i in [1..#FDs] ];
    pts_on_EDs := [[ [] : j in [1..#FDs[i]] ] : i in [1..#FDs] ];

    // The sequences FDs, TDCT and S2_4 match up with eachother.
    // Let S2_8 be the elements of S2 known to have at least one 8-covering
    // and S2_not8 the elements of S2 known to have no 8-coverings
    S2_not8 := {S2| }; 
    S2_8 := sub< S2 | [S2_4[i] : i in [1..#FDs]
                               | exists{j : j in [1..#FDs[i]] | not IsEmpty(pts_on_FDs[i,j])}] >;
    S2_8done := #S2_8 + #S2_not8 eq #S2_4 + 1;
    if S2_8done then
      r8 := Ilog(2, #S2_8);
      ru := Min(ru, r8); // refine rank bound
    end if;
   
    searchedB := 0;
    for b in [2 .. 20 by 2] do
      B := 10^b;
      for i in [1..#FDs], j in [1..#FDs[i]] do
        if #MWgens eq ru and (Sha mod 8 ne 0 or S2_8done) then 
          break b; 
        end if;
        if S2_4[i] in S2_not8 then
          continue i;
        elif IsEmpty(pts_on_FDs[i,j]) then
          if not IsDefined(EDs[i],j) then
//printf "EightDescent: "; time
            EDs[i,j], EDmaps[i,j] := EightDescent(FDs[i,j]);
            if not IsEmpty(EDs[i,j]) then
              S2_8 := S2_8 + sub< S2 | S2_4[i] >;
            elif forall{jj : jj in [1..#FDs[i]] | IsDefined(EDs[i],jj) and IsEmpty(EDs[i,jj])} then
              Include(~S2_not8, S2_4[i]);
            end if;
            S2_not8 := {S2| s+n : s in S2_8, n in S2_not8}; // S2_not8 = union of cosets of S2_8
            S2_8done := #S2_8 + #S2_not8 eq #S2_4 + 1;
            if S2_8done then
              r8 := Ilog(2, #S2_8);
              ru := Min(ru, r8); // refine rank bound
            end if;
          end if;

          for k := 1 to #EDs[i,j] do
            if #MWgens eq ru then 
              break k;
            end if;
            // TO DO: probably should skip when a pt has been found on *any* of the EDs[i,j]
            if not IsDefined(pts_on_EDs[i,j], k) then
              searchedB := B;
//printf "Search (B = 10^%o): ", b; time
              pts := PointSearch(EDs[i,j,k], B : OnlyOne);
              if not IsEmpty(pts) then
                pts_on_EDs[i,j,k] := pts[1];
                ptE := pts[1] @EDmaps[i,j,k] @FDmaps[i,j];
                MWgens := IndependentGenerators(MWgens cat [ptE]);
                assert #MWgens ge rl and #MWgens le ru;
                rl := #MWgens;
                if #MWgens gt MW_rank_known then
                  MW_rank_known := #MWgens;
                  vprintf MWSha: (#MWgens eq 1) 
                         select "Found a point of infinite order.\n"
                          else  "Found %o independent points.\n", #MWgens;
                end if;
              end if;
            end if;
          end for;

        end if;
      end for;
    end for;

    refine_sha_bounds(~sha, 2, ruTD-ru, ruTD-rl);
    refine_sha_bounds(~sha, 4, ruTDCT-ru, ruTDCT-rl); 
    if S2_8done then
      refine_sha_bounds(~sha, 8, r8-ru, r8-rl);
    end if;
    vprintf MWSha: "After 8-descent: \n    %o <= Rank(E) <= %o\n", rl,ru;
    print_sha_bounds(sha);
    if searchedB gt 0 then
      vprintf MWSha: "(Searched up to height 10^%o on the 8-coverings.)\n", Ilog(10,searchedB);
    end if;

    Done(~done,rl,ru,ranal,~MWgens,sha,~sha1,RankOnly,Sha,verbosity);
    if done then return [rl,ru], MWgens, sha1; end if; 
  end if;

  // THREE DESCENT (not useful for finding points: better to search on FDs)
  if rl lt ru or Sha mod 3 eq 0 then
    t3 := (#tors mod 3 eq 0) select 1 else 0;
    S3 := ThreeSelmerGroup(E);
    ruD3 := Ilog(3, #S3) - t3;  assert 3^(ruD3+t3) eq #S3;
    ru := Min(ru, ruD3);
    refine_sha_bounds(~sha, 3, ruD3-ru, ruD3-rl);
    vprintf MWSha: "The 3-Selmer group has rank %o\n", ruD3+t3;
    vprintf MWSha: "After 3-descent: \n    %o <= Rank(E) <= %o\n", rl,ru;
    print_sha_bounds(sha);
  end if;

end if; // FldRat

  // There's nothing more we can do

  assert forall{P : P in MWgens | Order(P) eq 0};
  if #MWgens lt rl then
    vprint MWSha: "Didn't find generators for the full Mordell-Weil group\n"; 
  end if;

  Done(~done,rl,ru,ranal,~MWgens,sha,~sha1,RankOnly,Sha,verbosity : final);
  return [rl,ru], MWgens, sha1;
end intrinsic;

