freeze;

/*******************************************************************
L-Series of higher genus curves

implements
  intrinsic LSeries(C:: CrvHyp: Precision:=0)

version 1: T+V Dokchitser, Aug 2010
  assumes the curve has a global regular Weierstrass model
version 2: T+V Dokchitser, Sep 2010
  L-series for arbitrary hyperelliptic curves, using RegularModel machinery
  (can't do arbitrary plane curves yet because Discriminant is
  not implemented for them, and it is needed in Ogg's formula)
*******************************************************************/


import "../Crv/RegModel/interface.m" : 
       number_of_points_on_special_fibre,
       splitting_degrees;

Z:=Integers(); Q:=Rationals(); PR:=PolynomialRing;

function HyperellGoodLocalFactor(C,p,deg) // MW 15 Oct 12, use func eq
  G:=Genus(C); if deg gt G then deg:=G; end if;
  g,f:=HyperellipticPolynomials(C); Cp:=HyperellipticCurve([PR(GF(p))|g,f]);
  // put together in a generating series for the zeta-function at p
  T:=PowerSeriesRing(Q,deg+1).1;
  if deg eq 1 then return 1-(p+1-#Cp)*T; end if;
  ser:=&+[#BaseChange(Cp,GF(p,n))/n*T^n: n in [1..deg]];
  // its numerator is the L-factor
  f:=Exp(ser)*(1-T)*(1-p*T); if deg lt G then return f; end if;
  f:=Polynomial(Coefficients(f));
  f:=f+&+[Coefficient(f,2*G-d)*p^(d-G)*Parent(f).1^d : d in [G+1..2*G]];
  return f;
end function;


function BadLocalFactor(C,p: deg:="full", R:="undefined")
  g:=Genus(C); if deg cmpeq "full" then deg:=2*g; end if;
  if deg gt 2*g then deg:=2*g; end if;
  if R cmpeq "undefined" then R:=RegularModel(C,p); end if;
  error if Degree(R`k,GF(p)) gt 1,
  "Cannot compute the local factor at "*Sprint(p)*" (need a field extension)";
  spldegs:=splitting_degrees(R); numgeocomps:=&+spldegs;
  vdisc:=Valuation(Discriminant(C),p); vcond:=vdisc-numgeocomps+1; // Ogg
  T:=PowerSeriesRing(Q,deg+1).1;
  ser:=Exp(&+[number_of_points_on_special_fibre(R,n)/n*T^n: n in [1..deg]]);
  h0term:=1-T; h2term:=&*[1-(p*T)^j: j in spldegs]; // h0 could be to power?
  h1term:=ser * h0term * h2term;
  if deg eq 2*g then return <p,vcond,PR(Q)!Eltseq(h1term),R,deg>; end if;
  return <p,vcond,h1term,R,deg>;
end function;

/* H:=HypergeometricData([3,4],[1,1,2,2]); t:=5^12;
   C:=HyperellipticCurve(H,t); // Ogg does not apply
   EulerFactor(H,t,5); // 25*x^4 + 20*x^3 + 10*x^2 + 4*x + 1
   EulerFactor(C,5); // 60*y^4 + 35*y^3 + 15*y^2 + 5*y + 1
   // difference is (1-T), need extra h0term ? */

intrinsic LSeries(C::CrvHyp[FldRat]: Precision:=0, ExcFactors:=[]) -> LSer
{L-Series of a hyperelliptic curve over Q}
  error if BaseField(C) ne Q, "The curve must be defined over the rationals";
  C:=MinimalWeierstrassModel(C); f,g:=HyperellipticPolynomials(C);
  genus:=Genus(C); vprint LSeries,1: "genus =",genus;
  D:=Discriminant(C); vprint LSeries,1: "D =",Factorisation(Z!D);
  bad:=PrimeDivisors(Z!D);
  if ExcFactors cmpeq "Ogg" then ExcFactors:=[<p,"Ogg">: p in bad]; end if;
  error if (ExcFactors cmpne []) and (ExtendedType(ExcFactors) cmpne SeqEnum[Tup]),
    "ExcFactors must be \"Ogg\", or a list of entries <p,vcond,localfactor> or <p,\"Ogg\">";
  excbad:=[e[1]: e in ExcFactors];
  badfactors:=[* e: e in ExcFactors | e[2] cmpne "Ogg" *];

  for p in bad do
    pos:=Position(excbad,p);
    ogg:=(pos ne 0) and (ExcFactors[pos][2] cmpeq "Ogg");
    if (pos eq 0) or ogg then
      vprint LSeries,1: "Local factor at",p;
      error if (genus ge 2) and (not ogg) and (Valuation(D,p) ge 12),
        "Valuation at "*Sprint(p)*" of the discriminant is >=12, cannot use "*
        "Ogg's formula unless forced by ExcFactors:=[<"*Sprint(p)*",\"Ogg\">]";
      Append(~badfactors,BadLocalFactor(C,p: deg:=1));
    end if;
  end for;

//  excbad:=[e[1]: e in badfactors]; // don't overwrite these?! MW 30 Nov12
  Cond:=&*[e[1]^e[2]: e in badfactors]; vprint LSeries,1: "Conductor = ",Cond;

  function LocalFactor(p,deg)
    if p in excbad and ExcFactors[Position(excbad,p)][2] cmpne "Ogg" then
      return badfactors[Position(excbad,p)][3];
    elif p in bad then
      assert exists(data){badfactors[j]:j in [1..#bad]| badfactors[j][1] eq p};
      _,_,loc,R,maxdeg:=Explode(data);
      if deg gt maxdeg then // recompute
        new:=BadLocalFactor(C,p: deg:=deg, R:=R);
        //! badfactors[Position(badfactors,data)]:=new;
        //  oops, magma does not allow to modify imported variables
        loc:=new[3];
      end if;
      vprint LSeries,1: "Local factor at",p,"=",loc;
      return loc;
    else return HyperellGoodLocalFactor(C,p,deg); end if;
  end function;

  gamma := [0: j in [1..genus]] cat [1: j in [1..genus]];
  return LSeries(2,gamma,Cond,LocalFactor:
		 Precision:=Precision, Name:=C,Parent:=C);
end intrinsic;

intrinsic EulerFactor(C::CrvHyp[FldRat],p::RngIntElt : Precision:=0) -> RngElt
{The Euler factor of a hyperelliptic curve at p}
 require IsPrime(p): "p must be prime"; C:=MinimalWeierstrassModel(C);
 if Precision eq 0 then deg:=Genus(C)*2; else deg:=Precision; end if;
 if Valuation(Discriminant(MinimalWeierstrassModel(C)),p) ne 0 then
  return BadLocalFactor(C,p : deg:=deg)[3]; end if;
 return HyperellGoodLocalFactor(C,p,deg); end intrinsic;

////////////////////////////////////////////////////////////////////////

function hyp_good_nf(C,P,deg) // MW 14 Nov 12
 F,m:=ResidueClassField(P); q:=#F; _,p,f:=IsPrimePower(q);
 G:=Genus(C); if deg gt G*f then deg:=G*f; end if;
 gh,fh:=HyperellipticPolynomials(C);
 gq:=Polynomial([m(x) : x in Coefficients(gh)]);
 if fh ne 0 then fq:=Polynomial([m(x) : x in Coefficients(fh)]);
 else fq:=0; end if; Cq:=HyperellipticCurve([gq,fq]);
 T:=PowerSeriesRing(Q,deg+1).1; lim:=deg div f;
 X:=[F]; for n in [2..lim] do Embed(F,GF(p,n*f)); X cat:=[GF(p,n*f)]; end for;
 ser:=&+[Parent(T)|#BaseChange(Cq,X[n])/n*T^n: n in [1..lim]]+O(T^(lim+1));
 res:=Exp(ser)*(1-T)*(1-q*T);
 if deg lt G*f then return Evaluate(res,T^f); end if;
 res:=Polynomial(Coefficients(res));
 res:=res+&+[Coefficient(res,2*G-d)*q^(d-G)*Parent(res).1^d : d in [G+1..2*G]];
 return Evaluate(res,Parent(res).1^f); end function;

intrinsic EulerFactor(C::CrvHyp[FldNum],P::RngOrdIdl : Precision:=0) -> RngElt
{The Euler factor of a hyperelliptic curve at a prime ideal P}
 require Valuation(Discriminant(C),P) eq 0: "Prime must be good";
 gh,fh:=HyperellipticPolynomials(C); CO:=Coefficients(gh) cat Coefficients(fh);
 require &and[Valuation(x,P) ge 0: x in CO]: "Model has denominators at P";
 q:=#ResidueClassField(P); _,p,f:=IsPrimePower(q);
 if Precision eq 0 then deg:=Genus(C)*2*f; else deg:=Precision; end if;
 return hyp_good_nf(C,P,deg); end intrinsic;

intrinsic EulerFactor(C::CrvHyp[FldNum],p::RngIntElt : Precision:=0) -> RngElt
{The pth Euler factor of the given elliptic curve over a number field}
 require IsPrime(p): "p must be prime";
 L:=[* EulerFactor(C,Ideal(P[1]) : Precision:=Precision) :
       P in AbsoluteDecomposition(BaseField(C),p) *];
 if &and[Type(x) eq RngUPolElt : x in L] then return &*[x : x in L]; end if;
 R:=PowerSeriesRing(Z,Precision+1); return &*[R!x : x in L]; end intrinsic;
