freeze;

//
//  Pic1desc.m
//  BrendanCreutz
//	Sept. 2011
//
//	Intrinsics for descents on hyperelliptic Jacobians
//  and on the Pic1 torsor.
//
//
//

import "../CrvG2/selmer.m" : MakeTry, CasselsKernel;
import "../Algaem/resolvent.m" : degnres;

declare verbose ConstantNeighborhoods, 2;

/***********************************************************************************
 Processing precision errors from p-adic Roots and Factorization
 ***********************************************************************************/
//Since the returned string has unpredictable linebreaks in it,
//remove all non-alphabetic characters from either string and compare those.

alphabetic_characters:={CodeToString(c): c in {StringToCode("A")..StringToCode("Z")}
                                         join {StringToCode("a")..StringToCode("z")}};

FactorizationErrorMessage:=&cat[a
   : a in Eltseq("Runtime error in'LocalFactorization': Factorization: Insufficient precision in factorization of a non-squarefree polynomial.")
   | a in alphabetic_characters];

RootFindingErrorMessage:=&cat[a
   : a in Eltseq("Runtime error in 'Roots': Insufficient precision to find roots")
   | a in alphabetic_characters];

function IsFactorizationPrecisionError(ERR)
   return FactorizationErrorMessage eq
          &cat[a: a in Eltseq(ERR`Object) | a in alphabetic_characters];
end function;

function IsRootFindingPrecisionError(ERR)
   return RootFindingErrorMessage eq
          &cat[a: a in Eltseq(ERR`Object) | a in alphabetic_characters];
end function;

/***********************************************************************************
***********************************************************************************/



SwapVar := func< a | a ne 0 select &+[ Coefficient(a,i)*Parent(a).1^(Degree(a)-i) : i in [0..Degree(a)]] else 0 >;

IntegralModel := function(f,q);
	
	// make a change of variables so that C: y^q = f is isomorphic to
	// C_2: y^q = c*f_mon(x) with f_mon(x) monic of the same degree.
	kpol := Parent(f);
	k := BaseRing(kpol);
	R<X,Y,Z> := PolynomialRing(k,3);
	d := Degree(f);
	lcf := LeadingCoefficient(f);
		if d mod q eq 0 then
			f2 := SwapVar(Evaluate(SwapVar(f),kpol.1*lcf));
			if IsIrreducible(f2) and Degree(k) eq 1 then
				//we can use this trick to get a nicer model
				O_A := Integers(ext<k|f/lcf>);
				// call optimised representation?
				f_mon := MinimalPolynomial(O_A.2);
				f2 := lcf*Polynomial([ k ! c : c in Coefficients(f_mon) ]);
			end if; //IsIrreducible(f2)
			return f2;
		else
			// homogenize differently:
			// f(x,z^q) = a_0*x^d + a_1*x^(d-1)*z^q + ... + a_d*z^(q*d)
			n := Integers()!( -(GF(q)!d)^1);
			f := Polynomial( [ k | Coefficient(f,j)*lcf^((d-j)*(q-n)-1) : j in [0..d-1] ] cat [1] );
			// x2 <- c^n*x
			// z2 <- c*z
			// y2 <- c^((dn+1)/q - d)*y
			// satisfy y2^q = f(x2)
			return f;
		end if;
			
end function;

CasselsKernel2 := function(f);
	// the imported CasselsKernel only works for genus 2
	d := Degree(f);
	if d eq 6 then
		CasGlob, resfact, subext_disc := CasselsKernel([a[1] : a in Factorization(f)]);
		return CasGlob,resfact;
	end if;
	ress := degnres(f,d div 2);
	resfact := Factorization(ress);
	CasGlob := exists{ r[1] : r in resfact | Degree(r[1]) eq 1 };
	return CasGlob,[ a[1] : a in resfact ];
end function; //CasselsKernel2


SelmerDiagram := function(f,S,q);

//==============================================================//
//                                                              //
//  Returns maps defining the following diagram                 //
//  where Prod_p is the product over p in S.                    //
//                                                              //
//  A ---------AtoASqmodk-----------> G > KerNorm + c1          //
//  |                                 |                         //
//  |                                 |                         //
// respFlds                         respGrps                    //
//  |                                 |                         //
//  |                                 |                         //
//  v                                 v                         //
// Prod_v K_v ----ApModqQs---------> G_v                        //
//                                                              //
//                                                              //
//  A = Q[x]/f                                                  //
//  G is the unramified outside S subgroup of                   //
//   either (A^* / A^*q) or (A^* / k^*A^*q) depending on        //
//   whether degree(f) is prime to q or not.                    //
//                                                              //
//  G_v is the local version (G tensored with k_v)              //
//                                                              //
//  c_1 is an element of G whose norm is = c mod k^*q           //
//  KerNorm is the kernel of the norm                           //
//                                                              //
//  AtoLocs is the map top left to bottom right                 //
//                                                              //
//==============================================================//

	c := LeadingCoefficient(f);
	kpol := Parent(f);
	k := BaseRing(kpol);

	if Degree(f) mod q eq 0 then
		modscalars := true;
	else
		modscalars := false;
	end if;	

	// Handle Q differently.
	if Degree(k) eq 1 then
		k := NumberField(kpol.1-1 : DoLinearExtension);
	end if;

	O_k := MaximalOrder(k);

	A := quo<kpol|f>; // The etale algebra corresponding to the roots of f
	vprintf Selmer, 3: "    Computing class/unit group info.\n";
	ASq,AtoASq,mB,B := pSelmerGroup(A,q,Set(S):Raw := true);
	kSq,ktokSq := pSelmerGroup(q,Set(S));
	ImageOfkSq := { AtoASq(Inverse(ktokSq)(a)) : a in Generators(kSq) };
		
	// Quotient to obtain A(S,q)/k(S,q) 
	if modscalars then
		ASqmodk, modk := quo<ASq | ImageOfkSq>;
	else
		ASqmodk := ASq;
		modk := hom< ASq -> ASq | OrderedGenerators(ASq)>;
	end if;
	// Labelling maps...
	AtoASqmodk := AtoASq * modk;
	ASqmodktoA := Inverse(AtoASqmodk);

	//ClearDenom := func<g | g*LCM([ Denominator(d) : d in Coefficients(g) ])>;
	//AtoASqQ := map< A -> ASqQ | x :-> AtoASqQ(ClearDenom(x)) >; //Hopefully this won't be neccessary.

	vprintf Selmer, 3: "    Getting representatives in A using power products.\n";
	B := Eltseq(B);
	// Representatives in A for the generators of ASqmodk (will be used later)
	repAs := [ PowerProduct(B,[c mod q : c in Eltseq(mB(g @@ modk))] ) : g in OrderedGenerators(ASqmodk) ];

	// Define the norm on ASqmodk
	NB := [Norm(b) : b in B];
	// N : A(S,q) -> Q(S,q)
	Norm1 := hom< ASq -> kSq | [ktokSq(PowerProduct(NB,[c mod q : c in Eltseq(mB(g))])) : g in OrderedGenerators(ASq)] >;
	// N2 : A(S,q)/k(S,q) -> k(S,q)
	Norm2 := hom< ASqmodk -> kSq | [ Norm1( a @@ modk ) : a in OrderedGenerators(ASqmodk) ] >;
	// The Norm condition
	bool, c1 := HasPreimage(ktokSq(c),Norm2);
	Ker := Kernel(Norm2);
	// Representation of the coset c1+Ker:
	_, modKer := quo<ASqmodk | Ker>;
	if bool then
		NormAndPreimc := <modKer, c1>;
		// The coset of KSqQ of elements satisfying the norm condition is
		// { c1 @@ quomap + x : x in Ker(quomap) }
	else
		// The norm condition is not satisfiable.
		NormAndPreimc := <modKer,{}>;
	end if;

	vprintf Selmer, 3: "    Computing absolute algebra.\n";
	// Represents A as a cartesian product of number fields.
	Aflds, AtoAflds := AbsoluteAlgebra(A);

	LocalSelmerGroup := function(K,v);
	/********************
	   input: 
		K: a number field (one of the factors of Aflds)
		v: a prime ideal of Integers(k)
		(We will work with K_v = K tensor k_v given as a Cartsesian product of completions of K)

	   output: maps
		ModqthPowers: PROD K_w --> PROD K_w^* / K_w^*q  	
		KtoKtensork_v: K --> PROD K_w
	*********************/ 

		O_K := MaximalOrder(K);
		pids := [ pid : pid in Factorization(ideal<O_K| Generators(v)>)];
		injs := [];
		ress := [];
		dim := 0;

		for pid in pids do
			K_p, inj := Completion(K,pid[1]);
			K_p`DefaultPrecision *:= 2*RamificationIndex(pid[1])*Degree(pid[1]);
			G,m := pSelmerGroup(q,K_p);
			injs cat:= [inj];
			ress cat:= [m];
			dim +:= #Generators(G);
		end for; // pid in pids

		G := AbelianGroup([ q : i in [1..dim]]);
		dima := 0;
		lis := [];
		for i in [1..#injs] do
			G_i := Codomain(ress[i]);
			lis cat:= [hom< G_i -> G | [G.(i+dima) : i in [1..#Generators(G_i)]] >];
			dima +:= #Generators(G_i);
		end for; // i 

		ModqthPowers := map< CartesianProduct([K_p : K_p in [ Domain(m) : m in ress]]) -> G | x :-> &+[ lis[i](ress[i](x[i])) : i in [1..#ress] ]>;
		KtoKtensork_v := map< K -> Domain(ModqthPowers) | x :-> < injs[i](x) : i in [1..#injs] > >;
		return ModqthPowers, KtoKtensork_v,pids;
	end function; // LocalSelmerGroup


	// The following will be sequences (parameterized by S) of maps to the various local objects
	respGrps := [];
	AtoLocs := [];
	CompletionsOfk := [];

	// Deal with Q differently
	if Degree(k) eq 1 then
		kk := Rationals();
	else
		kk := k;
	end if;

	for v in S do
		
		KstoKvqs := <>;
		pids := <>;
		
		vprintf Selmer, 3: "    Computing Local Selmer Groups at the prime v = %o.\n", v;
		for i in [1..NumberOfComponents(Aflds)] do
			m1,m2,pids_i := LocalSelmerGroup(Aflds[i],v);
			Append(~KstoKvqs, <m1,m2>);
			Append(~pids,pids_i);
		end for;// i in [1..NumberOfComponents(Aflds)]

		//Piece everything together and (if modscalars then) take the quotient by the image of k_v under diagonal embedding.
		Avq := CartesianProduct([Codomain(m[1]) : m in KstoKvqs ]);	
		Avq2 := AbelianGroup( &cat([ [q : i in [1..#Generators(Codomain(M[1]))] ]  : M in KstoKvqs ]));
		AvqtoAvq2 := map< Avq -> Avq2 | x :-> Avq2 ! &cat([ Eltseq(x[i]) : i in [1..NumberOfComponents(Avq)] ]) >;
		k_v,ktok_v := Completion(kk,v);
		if Degree(k) gt 1 then
			k_v`DefaultPrecision *:= Norm(v);
		end if;
		k_vq, k_vtok_vq := pSelmerGroup(q,k_v);
		k_vqtok_v := Inverse(k_vtok_vq);
		// The diagonal image of k_v in K_v
		if modscalars then	
			Imk_v := sub<Avq2 | [&cat[ Eltseq( M[1](M[2](Inverse(ktok_v)(k_vqtok_v(g))))) : M in KstoKvqs ] : g in OrderedGenerators(k_vq) ]>;
		else
			Imk_v := sub<Avq2 | Avq2!0>;
		end if;
		// Quotient by Imk_v to get the local (at v) version of A^* / k^*A^*q or A^* / A^*q 

		Avqmodk_v, modk_v := quo<Avq2| Imk_v>;
		AtoAvqmodk_v := map< A -> Avqmodk_v | 
			x :-> modk_v(AvqtoAvq2( Avq ! < KstoKvqs[i][1](KstoKvqs[i][2]( AtoAflds(x)[i] )) : i in [1..NumberOfComponents(Aflds)] > )) >; 
		ASqmodkToAvqmodk_v := hom< ASqmodk -> Avqmodk_v | 
			[ modk_v(Avq2 ! &cat[ Eltseq( KstoKvqs[i][1] ( KstoKvqs[i][2] ( AtoAflds(x)[i] ) ) ) : i in [1..NumberOfComponents(Aflds)] ]) : x in repAs] >;
		//store this information for later use:
		respGrps cat:= [ASqmodkToAvqmodk_v];
		AtoLocs cat:= [AtoAvqmodk_v];
		CompletionsOfk cat:= [ <ktok_v,k_vtok_vq,pids> ];

	end for; // v in S

	return NormAndPreimc,AtoASqmodk,ASqmodktoA,respGrps,AtoLocs,repAs,CompletionsOfk,AtoAflds;

end function; //SelmerDiagram


intrinsic PicnDescent(f :: RngUPolElt,q :: RngIntElt) -> Tup
{Computes fake phi-Selmer group of Jac(C) and the fake phi-Selmer set of Pic^1(C). 
Currently only implemented for q = 2}

// Doing a phi-descent on the curve y^q = f(x)

MethodIsValid := true;
Pic1kEmpty := false;
// Issues to keep in mind: 
// C not ELS ==> Pic1 descent is not proven to give the 2-Selmer set of C, though it may still prove Pic1(k) is empty.
// We currently have no good way of checking local solvability of Pic1.
// If C does not have period 1 everywhere locally, Pic^0(C) may not equal J(k), so we are in trouble.

// just initializing
kpol<x> := Parent(f);
k := BaseRing(f);
if Degree(k) gt 1 then
	k<alpha> := k;
end if;
k0 := k;
f_original := f;
// We need the q-th roots of unity.
if #Roots(CyclotomicPolynomial(q),k) eq 0 then
	k<zeta> := ext<k|CyclotomicPolynomial(q)>;
	if Degree(k0) ne 1 then
		k<nu> := AbsoluteField(k);
	end if;
	kpol<x> := PolynomialRing(k);
	f := Evaluate(f,kpol.1);
	vprintf Selmer, 1: "  Adjoining %o-th roots of unity to base field.\n", q;
end if;

O_k := Integers(k);
vprintf Selmer, 1: "Doing descents for the curve with defining equation: y^%o = %o defined over %o.\n", q, f, k;

f := IntegralModel(f,q);
// change variables so that f = c * monic
lcf := LeadingCoefficient(f);
f_mon := f/lcf;

//assert q eq 2;
//assert IsIsomorphic(HyperellipticCurve(f_original),HyperellipticCurve(f));
degree := Degree(f);
g := degree mod q eq 0 select Integers()!((degree - 2)*(q-1)/2) else Integers()!((degree-1)*(q-1)/2);

vprintf Selmer, 1: "  Model changed to y^%o = %o.\n", q, f;
vprintf Selmer, 3: "  Curve has genus %o.\n", g;

fact := [ a[1] : a in Factorization(f_mon) ];
PrimeToq := exists(primetoqf){ a : a in fact | GCD(Degree(a),q) eq 1 };
PrimeToqGlob := PrimeToq;
//assert q eq 2;
ranktorsglobal := #fact-1;
if degree mod q ne 0 then ranktorsglobal +:= 1; end if; // to account for ramification above infinity.
extratwotorsion := 0; // until we show otherwise we assume this to be the case. (see below)
flagphitorsglobal := false; // this flag will be set to true if we encounter problems computing J(k)[phi]

if PrimeToq then
	Seltofakeinj := true;
	resfact := [];
	CasGlob := true;
	ranktorsglobal -:= 1; // ranktors = #ramification points - 2.
else
	if q eq 2 then
		CasGlob, resfact := CasselsKernel2(f);
		if g mod 2 ne 0 then
			Seltofakeinj := false;
		  //We need the roots of resfact can lead to extra 2-torsion.
			roots := &cat[ [ r[1] : r in Roots(g)  ]: g in resfact ] ;
			DegreesOfFactors := [ Degree(g) : g in fact ];
			Pd := Partitions(Integers()!(degree/2));
			PartitionsOfRoots2 := 0;
			for P in Pd do
				vals := Set(P);
				ct := 1;
				for v in vals do
					v1 := #[ w : w in P | w eq v ];
					v2 := #[ w : w in DegreesOfFactors | w eq v];
					ct *:= Binomial(v2,v1);
					if ct eq 0 then break; end if;
				end for;
				PartitionsOfRoots2 +:= ct;
			end for;
			PartitionsOfRoots2 := Integers()!(PartitionsOfRoots2/2);
			//PartionsOfRoots2 is the number of distinct partitions of the roots of f into 
			//2 Galois stable subsets of equal size. These correspond to roots of the resolvent
			//which are the trivial 2-torsion point.
			extratwotorsion := Ilog(2,#roots - PartitionsOfRoots2 + 1);
			vprintf Selmer, 1: "    Factorizations over quadratic extensions ==> %o extra dimensions of 2-torsion.\n", extratwotorsion;
			ranktorsglobal +:= extratwotorsion;
		else
			Seltofakeinj := CasGlob;
			// g = 0 mod 2 no factors of odd degree
		end if; // g mod 2 ne 0

	else
		Seltofakeinj := false;
		resfact := [];
		CasGlob := false;
		flagphitorsglobal := true;
		vprintf Selmer, 2: "Computation of J(k)[phi], not implemented for this case.\n";
		
	end if; //q eq 2
	
end if; //PrimeToq

discr := O_k!Discriminant(f);
if Degree(k) eq 1 then
	S_small := Setseq( { p[1] : p in Factorization(Numerator(discr)*Denominator(discr)) | p[2] ge q } join { p[1] : p in Factorization(Numerator(lcf*q)*Denominator(lcf*q)) });
else
	S_small := Setseq({ p[1] : p in Factorization(discr*O_k) | p[2] ge q } join { p[1] : p in Factorization(lcf*q*O_k) } );
end if;
//if SSize eq "Small" then
S := S_small;

// S is the set of bad primes. This can be fed in as a set of rational primes
// in which case we take the primes of k above these, or a set of prime ideals of O_k
if Type(Random(S)) eq RngOrdIdl then
	S := [ pid[1] : pid in Factorization((&*S)) ];
end if;

/* jibber jabber for sorting primes while testing.
S_1 := [ a : a in S | Norm(a) mod q ne 0 ];
S_2 := [ a : a in S | Norm(a) mod q eq 0 ];
S := S_1 cat S_2;
*/

//Now we do all the class/unit group computations and compute all the restriction maps.
vprintf Selmer, 2: "  Constructing Local-Global Selmer diagram:\n    Discriminant(O_A) = %o.\n    Set of bad primes is S = %o.\n", discr, S;
NormAndPreimc,AtoASqmodk,ASqmodktoA,respGrps,AtoLocs,repAs,CompletionsOfk,AtoAflds := SelmerDiagram(f,S,q);
A := Domain(AtoASqmodk); // the etale algebra as RngUPolRes.
Aflds := Codomain(AtoAflds);
AfldsToA := Inverse(AtoAflds);
ASqmodk := Codomain(AtoASqmodk); // The unramified outside S subgroup mod scalars and q-th powers
NASqmodktokq := NormAndPreimc[1]; // The induced norm on ASqmodk;
if Type(NormAndPreimc[2]) eq SetEnum then
	vprintf Selmer, 1: "  Norm condition not satisfiable.\n  --> Sel(C) and Sel(Pic^1(C)) are empty.\n";
	Pic1kEmpty := true;
	alpha_c := ASqmodk ! 0;
else
	alpha_c := NormAndPreimc[2]; // an element of norm = c mod k^*2
end if;

vprintf Selmer, 2: "  Bound on fake Selmer group dim after norm condition : %o.\n", NumberOfGenerators(Kernel(NormAndPreimc[1]));


// Determine the image of some known points.
if ranktorsglobal gt 0 and #fact gt 1 then
	if not PrimeToq then
		// all factors of f define divisors of degree = 0 mod q
		Im_tors := {};
		for g in fact do
			cofactor := (lcf)*&*[ h : h in fact | h ne g];
			Include(~Im_tors,AtoASqmodk(Evaluate(g-cofactor^(q-1),A.1)));
		end for;
	else
		//primetoqf is a divisors of degree prime to q
		cofactor := (lcf)*&*[ h : h in fact | h ne primetoqf];
		imageofprimetoqf := AtoASqmodk( Evaluate((-1)^Degree(primetoqf)*primetoqf + cofactor^(q-1),A.1));
		dp_v := Degree(primetoqf);
		//imprimetoq_v is the image under Pic(C) --> ASQmodk of a divisor of degree dp_v.
		im1global := (Integers()!((GF(q)!dp_v)^-1))*imageofprimetoqf;
		//im1global is the image of a divisor of degree 1 (which is not neccesarily a point)
		
		// We compute the images of the ramifiction points
		// and save them as they will be needed for descent on the curve.
		muOmegaGlobal := {};
		for g in { g : g in fact | Degree(g) eq 1 } do
			cofactor := (lcf)*&*[ h : h in fact | h ne g];
			image := AtoASqmodk( Evaluate((-g + cofactor^(q-1)),A.1) );
			//vprintf Selmer, 3:  "       Ramification point x=%o maps to %o\n", Roots(g)[1][1], Eltseq(image);
			Include(~muOmegaGlobal,image);	
		end for;		
		vprintf Selmer, 3: "        Image of k-rational ramification points : %o.\n", {Eltseq(a) : a in muOmegaGlobal};

		// we compute the image of some phi-torsion points.
		Im_tors := {};
		for g in fact do
				cofactor := (lcf)*&*[ h : h in fact | h ne g];
				Include(~Im_tors,AtoASqmodk(Evaluate( ((-1)^Degree(g)*(g-cofactor^(q-1))),A.1)) - Degree(g)*im1global );
			end for;
	end if;
else // ranktorsglobal gt 0
	Im_tors := { ASqmodk ! 0 };
end if;

muJktors := sub< ASqmodk | Im_tors >;



LocalImages := function(v);
	// Computes the images of (x-T) on Pic^0(C) and Pic^1(C) over k_v
	// v is a prime ideal of O_k (or a rational prime if k eq Q...
	vprintf Selmer, 1: "\n\n  Computing local image at %o.\n",v;
	//Import data from SelmerDiagram:
	ind_v := Index(S,v);
	AtoLoc := AtoLocs[ind_v];
	Avqmodk_v := Codomain(AtoLoc);
	res_v := respGrps[ind_v];
	k_vdata := CompletionsOfk[ind_v];
	pids := k_vdata[3];
	ktok_v := k_vdata[1];
	k_v := Codomain(ktok_v);
	modqthpowers := k_vdata[2];
	k_vq := Codomain(modqthpowers);
	Onemodqthpowers := k_vq ! 0;
	if Degree(k) eq 1 then
		O_v := pAdicRing(v);
		O_ktoO_v := map< O_k -> O_v | x :-> O_v ! x >;
		O_vtoO_k := map< O_v -> O_k | x :-> O_k ! x >;
		relDeg := 1;
		e_v := 1;
		nv := v;
	else
		O_v,O_ktoO_v := Completion(O_k,v);
		O_vtoO_k := Inverse(O_ktoO_v);
		e_v := RamificationIndex(v);
		relDeg := Degree(v)*e_v;
		nv := Norm(v)^e_v;
	end if;
	k_v`DefaultPrecision *:= (2*e_v);
	O_v`DefaultPrecision *:= (2*e_v);	
	pi_v := UniformizingElement(O_v);
	pi := O_vtoO_k(pi_v);
	p := Factorization(nv)[1][1]; // rational prime under v
	liesaboveq := nv mod q eq 0; // is p = q?
	k_vpol := PolynomialRing(k_v);
	O_vpol := PolynomialRing(O_v); 
	F_v,modv := ResidueClassField(O_v);
	FF<xbar> := PolynomialRing(F_v);
	// functions for coercing polynomials around
	Polytok_v := func< h | h ne 0 select Polynomial([ O_ktoO_v(c) : c in Coefficients(h)]) else O_vpol!0>;
	Polytok := func< h | Polynomial([ O_vtoO_k(c) : c in Coefficients(h)])>;
	PolytoF_v := func< h | Polynomial([ modv(O_ktoO_v(c)) : c in Coefficients(h)])>;
	Polymodvtok := func< h | h ne 0 select Polynomial([ O_vtoO_k(c @@ modv)  : c in Coefficients(h)]) else kpol!0>;
	
	fmonlocal := Polytok_v(f_mon);
	flocal := O_ktoO_v(O_k!lcf)*fmonlocal;
	f_F := PolytoF_v(f);

	// valuation of a polynomial:
	dpolys := func< g,h | g-h ne 0 select Min([ Valuation(c,v) : c in Coefficients(g-h) ]) else Infinity() >;
	
	muOmega := {}; // initializing image of ramification points (and point above infinity)
	Foundim1 := false; // until we get our hands on a divisor of degree 1.
	im1 := Avqmodk_v ! 0;
	
	// We have finished setting up our 'notation'.
	
	// Now factor f locally to determine size of local image etc.
	ff := [ a[1] : a in Factorization(fmonlocal) ];
	//Sometimes there are precision problems? 
	// if this persists do try--catch--increase precision... 

	ranktors := #ff - 1;
	if degree mod q ne 0 then ranktors +:= 1; end if; // to account for the ramification point above infinity
	PrimeToq := exists(primetoq_v){ a : a in ff | Degree(a) mod q ne 0 };

	if PrimeToq then
		// Locally we have a divisor of degree prime to q. 
		// This makes things a bit easier.
		resfact_v := [];
		CasLocal := true;
		ranktors -:= 1; // so in total ranktors = #ff - 2 (counting the extra factor at infty)
		vprintf Selmer, 1: "    Local Cassels kernel is trivial.\n";
		
		Foundim1 := true;
		
		if degree mod q ne 0 then
			im1 := Codomain(AtoLoc)!0;
			Include(~muOmega,im1);
			//im1 is the image of the ramification point above infty
		else
			cofactor := (O_ktoO_v(lcf))*&*[ g : g in ff | g ne primetoq_v];
			imprimetoq_v := AtoLoc(Evaluate( Polytok((-1)^Degree(primetoq_v)*primetoq_v + cofactor^(q-1)),A.1) );
			//printf " the factor %o maps to %o.\n", primetoq_v, imprimetoq_v;
			dp_v := Degree(primetoq_v);
			//imprimetoq_v is the image under Pic(C) --> localSel of a divisor of degree dp_v.
			im1 := (Integers()!((GF(q)!dp_v)^-1))*imprimetoq_v;
			//im1 is the image of a divisor of degree 1 (which is not a point)
		end if;

		// We compute the images of the ramifiction points
		// and save them as they will be needed for descent on the curve.
		for g in { g : g in ff | Degree(g) eq 1 } do
			cofactor := (O_ktoO_v(lcf))*&*[ h : h in ff | h ne g];
			image := AtoLoc( Evaluate(Polytok(-g + cofactor^(q-1)),A.1) );
			vprintf Selmer, 3:  "       Ramification point x=%o maps to %o\n", Roots(g)[1][1], Eltseq(image);
			Include(~muOmega,image);	
		end for;		
		//vprintf Selmer, 3: "        Image of ramification points : %o.\n", {Eltseq(a) : a in muOmega};

		if ranktors gt 0 then
			// we compute the image of all phi-torsion points.
			Im_tors := {};
			for g in ff do
				cofactor := (O_ktoO_v(lcf))*&*[ h : h in ff | h ne g];
				Include(~Im_tors,AtoLoc(Evaluate(Polytok((-1)^Degree(g)*(g-cofactor^(q-1))),A.1)) - Degree(g)*im1 );
			end for;
		else
			Im_tors := { Avqmodk_v ! 0 };
		end if;
		//Im_tors is the image under (x-T) of the phi-torsion in J(k_v)
				
	else //PrimeToq
	
		// Compute the image of effective divisors of degree 1
		if IsPower(ktok_v(lcf),q) then
			Foundim1 := true;
			im1 := Codomain(AtoLoc)!0;
			// image of a point above infinity
		end if;
				
		Im_tors := {};
		if #ff gt 1 then
			for g in ff do
				cofactor := (O_ktoO_v(lcf))*&*[ h : h in ff | h ne g];
				Include(~Im_tors,AtoLoc(Evaluate(Polytok(g-cofactor^(q-1)),A.1) ));
			end for;
		else
			Im_tors := { Avqmodk_v ! 0};
		end if;
		//Im_tors is most of the image under (x-T) of the phi-torsion in J(k_v)
		//For q eq 2 and g odd there can be extra torsion coming from factorization 
		//over quadratic extensions which is not accounted for.
		//For q odd, the situation can be more complicated and is not yet satisfactorily dealt with.
	
		if q eq 2 then
			
			if g mod 2 eq 0 and CasGlob or #resfact eq 0 then
				CasLocal := true;
				// no need to consider the resolvent again.
				
			else
				
				//we need to see if the resolvent has roots locally.
				if #resfact gt 0 then
					O_v_stored_precision:= O_v`DefaultPrecision;
					while true do
						resP:=Polytok_v(&*resfact);
						try
							roots:=Roots(resP);
							no_error:=true;
						catch ERR
							if IsRootFindingPrecisionError(ERR) then
								no_error:=false;
							else
								error ERR;
							end if;
						end try;
						if no_error then
							break;
						end if;
						O_v`DefaultPrecision +:= 10;
						vprint Selmer, 3: "    Finding roots of resolvent requires: increasing precision to %o", O_v`DefaultPrecision;
					end while;
					O_v`DefaultPrecision:=O_v_stored_precision;
					vprintf Selmer, 3: "    Resolvent has %o roots locally.\n",#roots;
				else 
					roots := [];
				end if; //#resfact gt 0
				
				if g mod 2 eq 0 then
					CasLocal := #roots gt 0;
				else
					CasLocal := true;
					// g is odd.
					// We need the roots of resfact can lead to extra 2-torsion.
					DegreesOfFactors := [ Degree(g) : g in ff ];
					Pd := Partitions(Integers()!(degree/2));
					PartitionsOfRoots2 := 0;
					for P in Pd do
						vals := Set(P);
						ct := 1;
						for v in vals do
							v1 := #[ w : w in P | w eq v ];
							v2 := #[ w : w in DegreesOfFactors | w eq v];
							ct *:= Binomial(v2,v1);
							if ct eq 0 then break; end if;
						end for;
						PartitionsOfRoots2 +:= ct;
					end for;
					PartitionsOfRoots2 := Integers()!(PartitionsOfRoots2/2);
					//PartionsOfRoots2 is the number of distinct partitions of the roots of f into 
					//2 Galois stable subsets of equal size. These correspond to roots of the resolvent
					//which are the trivial 2-torsion point.
					dimn := Ilog(2,#roots - PartitionsOfRoots2 + 1);
					vprintf Selmer, 1: "    Factorizations over quadratic extensions ==> %o extra dimensions of 2-torsion.\n", dimn;
					ranktors +:= dimn;
				end if;	 // g mod 2 eq 0
				
			end if; // g mod 2 eq 0 and CasGlob or #resfact eq 0 
		else // q eq 2
		  
			if q eq 3 and Degree(f) eq 6 then
				
				// there are really only two possibilities for the factorization type:
				// irreducible ==> phi-torsion has dim le 1, Cassels kernel is non trivial
				// (3,3) ==> same as above unless the Galois group is Z/3
				if #ff eq 1 then
					ranktors := 1;
					CasLocal := false;
				else
					// f = cubic*cubic
					ranktors := 2;
					flagtorsion := true; // this lets us know we only have an upper bound.
					// in fact: ranktors = 2 iff both cubics split over the same cubic field
					// otherwise ranktors = 1.

					printf "Flagging phi-torsion computation over %o.\n", k_v;
					printf "Dimension computed is only an upper bound.\n";			
					CasLocal := false;
				end if;

			else
				error "Not yet implemented in this case q = 3, Degree(f) = 3n with n ge 3.";
			end if;
		
		end if; //q eq 2
	
	end if; //PrimeToq

	// Determine dim(J(k_v)/phiJ(k_v)) and dimension of its image under (x-T)
	dim := ranktors + (liesaboveq select Integers()!(g*relDeg/(q-1)) else 0); 
	extraval := liesaboveq select 2*e_v else 0; 
	
	if not CasLocal then dimim := dim - 1; else dimim := dim; end if;
	vprintf Selmer, 1: "    J(k_v)[phi] has dimension %o.\n", ranktors;
	vprintf Selmer, 2: "    J(k_v)/phi(J(k_v)) has dimension %o, its image under (x-T) has dimension %o.\n", dim, dimim;
	ImPic0 := sub<Avqmodk_v | Im_tors >;
	vprintf Selmer, 2: "    Obvious torsion generates a space of dimension %o.\n",Ngens(ImPic0);
	
	// At this point we know the size of the local image
	// We have computed the images of some torsion points
	// The next step is to search for divisors and take their images under (x-T) until we get
	// enough points.

	// To do this we will search for points over extensions of k_v of degree d
	// then take the images under (x-T) of their traces

	SwapVar := func< a | a ne 0 select &+[ Coefficient(a,i)*Parent(a).1^(Degree(a)-i) : i in [0..Degree(a)]] else 0 >;

	
	
	
	
	TraceOfxminusT := function(f,h,v,ktok_v);

		// Input: h in kpol of degree d defines an extension K_w of k_v of degree d.
		// Output: K_w, KtoK_w and the map C(K_w) --Tr--> Div^d(C_v) --(x-T)--> L_v*	

		kpol<T> := Parent(f);
		k := BaseRing(kpol);
		d := Degree(h);

		if d eq 1 then
			//effective divisor of degree 1 (i.e. points in C(k_v)) the map is easier:
			TraceMap := func< x |  (x @@ ktok_v) - T >;
			TraceMap2 := func< z | (z @@ ktok_v)*(1 - (z@@ktok_v)*T) >;
			f_K := f;
			K_w := Codomain(ktok_v);
			KtoK_w := ktok_v;
		else
			//need to construct the extension
			Krel := ext<k | h>;
			O_Krel := Integers(Krel);
			K := AbsoluteField(Krel);
			O_K := Integers(K);
			if Type(v) eq RngIntElt then
				vO_K := ideal< O_K | v >;
			else
				vO_K := ideal< O_K | [ O_K ! g : g in Generators(v) ] >;
			end if;
			p := Factorization(Norm(v))[1][1];
			pids := { Ideal(w[1]) : w in Decomposition(K,p) };
			wabovev := { w: w in Support(vO_K) } meet pids;
			// if the polys satisfy what I said there will be only one prime here.
			//wds := { w : w in wabovev | RamificationIndex(w)*Degree(w)/(RamificationIndex(v)*Degree(v)) eq Degree(Koverk) };
			w := Random(wabovev);
			K_w<eta>, KtoK_w := Completion(K,w);
			K_w`DefaultPrecision *:= d;
			f_K := Polynomial([ K ! c : c in Coefficients(f) ]);
			// think about if we need precision *:= d?

			TraceMap := function(x);
			
				// for (x,y) in C(K_w) computes prod_sigma (sigma(x) - T) in k[T]
				xoverk := Krel ! (x @@ KtoK_w);
				return (-1)^d*CharacteristicPolynomial(xoverk);	
	
			end function;

			TraceMap2 := function(z);
		
				// As above, but used when z = 1/x in pi*O_v
				// computes prod_sigma sigma(z)*(1 - sigma(z)*T) = prod_sigma (1/sigma(z) - T)
				zoverk := Krel ! (z @@ KtoK_w);
				return Norm(zoverk)*SwapVar((-1)^d*CharacteristicPolynomial(zoverk));
	
			end function;
			
		end if; // Degree(h) eq 1 
	
		return f_K,K_w,KtoK_w,TraceMap,TraceMap2;

	end function; // TraceOfxminusT
	
	
	TamelyRamifiedPolys := function(d);
	
		//Computes polynomials (over k) which define all (at worst) tamely 
		//ramified extensions of k_v of degree d.
		//These will be plugged into TraceOfxminusT above
		
		// currently only for k_v = Q_p (though k may be of arbitrary degree),
		// or for d prime and k_v/Q_p unramified.
		
		polys := [  ];
		p := Characteristic(F_v);
		id := Ilog(p,#F_v);
		
		error if e_v gt 1, "Computation of tamely ramified extensions of k_v not implemented in this case: k_v/Q_p is ramified.\n";
		
		if id eq 1 then
			// k_v = Q_p
			Z_p := pAdicQuotientRing(p,2);
			for m in Divisors(d) do
				n := d div m;
				if n mod p ne 0 then
					// we construct extensions with ramification degree d/m = n
					R := ext< Z_p | m >;
					fR := Polynomial([ Rationals()! c : c in Coefficients(DefiningPolynomial(R))]);
					
					
					
					F_R,modw := ResidueClassField(R);
					zeta := PrimitiveElement(F_R)@@ modw;
					
					S<T> := PolynomialRing(Rationals());
					SS<TT> := PolynomialRing(S);
					
					exps := GCD(n,p^m -1);
					for s in [1..exps] do
						a_r := (zeta^s*p);
						a_rpol := &+[ Rationals()!(Eltseq(a_r)[i])*T^(i-1) : i in [1..#Eltseq(a_r)] ];
						polys cat:= [Resultant(Evaluate(fR,TT),(S.1-TT)^n+Evaluate(a_rpol,TT))];
					end for;
				end if;
			end for;
			return polys;
			
		else
			
			error if not IsPrime(d), "Computation of tamely ramified extensions of k_v not implemented in this case: k_v/Q_p nontrivial and d composite.";
			
			// this leaves totally ramified and totally unramified cases
			unr := Polymodvtok(IrreduciblePolynomial(F_v,d));
			unr := Polynomial([ c mod v : c in Coefficients(unr) ]);
			polys cat:= [ unr ];
			// this takes care of the unramified case.
			if d mod p ne 0 then 
				//we have some tamely ramified extensions to compute
				zeta := PrimitiveElement(F_v)@@ modv;
				zeta := O_vtoO_k(zeta) mod v; // we only need it to precision 1 
				exps := GCD(d,p^d -1);
				for s in [1..exps] do
					polys cat:= [ kpol.1^d - zeta^s*p ];
				end for;
			end if;
			return polys;
			
		end if; // id eq 1
		
	end function; // TamelyRamifiedPolys
		
	
	LocalImageCk_v := function(d,f_global,k_v,ktok_v,mu_1,mu_2,q,im1,Foundim1,ImPic0);


		SpaceGenerated := ImPic0;
		// for C : y^q = f_global(x), f_global a poly over k with completion k_v
		// Computes the image of C(k_v) under mu = (x - T)
		// using mu_1 for x in O_v
		// using mu_2 for 1/x in O_v
		
		// implicit input: Foundim1, im1
		// we pass the degree of the extension under consideration as d.
		// for points on the Jacobian use mu(x) - d*im1
		
		f := Polynomial([ k_v | ktok_v(c) : c in Coefficients(f_global) ]);
		extraval := 2*Valuation(k_v!q);
		//printf " 2*val(q) = %o.\n", extraval;
		lcf := LeadingCoefficient(f);
		SetPrecision(k_v,100*(extraval +1));
		ff := { g[1] : g in Factorization( f/lcf)};
		k_vpol<T> := Parent(f);
		O_v := Integers(k_v);
		extralval := 2*Valuation(ktok_v(q));
		pi := UniformizingElement(k_v);
		F_v, modv := ResidueClassField(O_v);
		RepsForF_v := [ a@@modv : a in F_v ];
		valpoly := func< g | g eq 0 select Infinity() else Min([Valuation(c) : c in Coefficients(g)])>;

		IsqthPowerOracle := function(e);
			// returns a function which given a in k_v decides if a is a q-th power mod Valuation(a) + e
			// compare with simple call IsPower(a,q)
			// no reason this should be any faster...?

			if e eq extraval + 1 then
				return func< a | IsPower(a,q)>;
			end if;
			if e eq 1 then
				if (#F_v - 1) mod q eq 0 then 
					oracle := function(a);
						// a in k_v
						if a eq 0 then return true; end if;
						val := Valuation(a);
						if val mod q ne 0 then return false; end if;
						ua := (O_v! (a/pi^(val)));
						return IsPower(modv(ua),q);
					end function;
				else
					oracle := func< a | true >;
				end if;
			else
				//e > 1 so we need to work in a residue Ring.
				//should only be needed for v | q
				qthpowers := { (&+[ a[i]*pi^(i-1) : i in [1..e] ])^q : a in CartesianProduct([RepsForF_v : i in [1..e]]) };
				oracle := function(a);
					// a in k_v
					if a eq 0 then return true; end if;
					val := Valuation(a);
					if val mod q ne 0 then return false; end if;
					ua := a/pi^(val);
					for qp in qthpowers do
						if Valuation(qp - ua) ge e then
							return true;
						end if;
					end for;
					return false;
				end function;
			end if;
			return oracle;
		end function; // IsqthPowerOracle

		qporacles := [ IsqthPowerOracle(e) : e in [1..extraval + 1] ];

		ConditionToGoDeeper := function(fx1,v1,E1);
			// v1 = ord(fx1)
			// determines intersection of fx1 + pi^E1*O_v and (O_v)^q
			// verbosity explains the conditions.
			if E1 le v1 then 
				vprintf ConstantNeighborhoods, 2: " which contains q-th and non-qth powers.\n", E1,v1;
				return true;
			end if;
			s := Min({E1 - v1,extraval+1});
			isqthp := qporacles[s](fx1);
			if not isqthp then
				vprintf ConstantNeighborhoods, 2: " which is disjoint from O_v^q.\n";
			else
				if s eq extraval + 1 then
					vprintf ConstantNeighborhoods, 2: " subset (O_v)^q.\n";
				else
					vprintf ConstantNeighborhoods, 2: " which maps into (O_v/pi^%o)^q.\n",s;
				end if;
			end if;
			return isqthp;
		end function;
		
		// inf := false;
		// used for verbosity during testing.
		
		procedure qthPowerClasses(inf,f,x0,e,G0,c0,mu,~W,~im1,~Foundim1,~SpaceGenerated);
			
			if Ngens(SpaceGenerated) lt dimim then
				//we are still looking for points
			
				for r in RepsForF_v do
					// consider smaller nbhd of x0 + pi^(e-1)O_v
					c1 := c0;
					x1 := x0 + pi^e*r;
					fx1 := Evaluate(f,x1);
					if IsWeaklyZero(fx1) then 
						vprintf ConstantNeighborhoods, 2: "         Image is torsion.\n";
					else
						v1 := Valuation(fx1); 
						E1 := valpoly( Evaluate(f,x1+pi^(e+1)*T) - fx1);
						if not inf then
							vprintf ConstantNeighborhoods, 2: "      For x in %o + pi^%o*O_v.\n",x1,e+1;
							vprintf ConstantNeighborhoods, 2: "        ==> f(x,1) in %o + pi^%o*O_v", fx1, E1;
						else
							vprintf ConstantNeighborhoods, 2: "      For z in %o + pi^%o*O_v.\n",x1,e+1;
							vprintf ConstantNeighborhoods, 2: "        ==> f(1,z) in %o + pi^%o*O_v", fx1, E1;
						end if;
						if ConditionToGoDeeper(fx1,v1,E1) then
							G1 := { g : g in G0 | valpoly( Evaluate(g,x1+pi^(e+1)*T)-Evaluate(g,x1)) le Valuation(Evaluate(g,x1)) };
							if #G1 eq 0 or (#G1 eq 1 and #{ g : g in G1 | Degree(g) eq 1} eq 1) then
								if G1 ne G0 then
									c1 := extraval;
								else
									c1 := c0 - 1;
								end if;
								if c1 eq 0 then
									if #G1 eq 0 then
										vprintf ConstantNeighborhoods, 2: "         mu is constant.\n";
										impt := mu(x1);
										// we have found the image of this neighborhood
										// we just check if it is anyhting new:
										if d mod q ne 0 and not Foundim1 then
											
											Foundim1 := true;
											im1 := 	(Integers()!Inverse(GF(q)!(d)))*impt;
											W := W join { impt };
									
										else //
											
											if not impt in W then
												W := W join {impt};
												if not inf then
													vprintf ConstantNeighborhoods, 1: "      Found new image:\n        x in %o + pi^%o*O_v maps to\n        %o.\n",x1,e+1,impt;
												else
													vprintf ConstantNeighborhoods, 1: "      Found new image:\n        z in %o + pi^%o*O_v maps to\n        %o.\n",x1,e+1,impt;
												end if;
												image := mu(x1) - d*im1;
												if not image in SpaceGenerated then
													if not inf then
														vprintf Selmer, 2: "      Found new image:\n        x in %o + pi^%o*O_v maps to\n        %o.\n",x1,e+1,Eltseq(image);
													else
														vprintf Selmer, 2: "      Found new image:\n        z in %o + pi^%o*O_v maps to\n        %o.\n",x1,e+1,Eltseq(image);
													end if;

													SpaceGenerated := sub< Avqmodk_v | Generators(SpaceGenerated) join {image}>;
													if Ngens(SpaceGenerated) ge dimim then
														break;
													end if;
												end if;
											end if;	
																				
										end if; // d mod q ne 0 and not Foundim1
										
									else
										vprintf ConstantNeighborhoods, 2: "         Image is torsion.\n";
									end if;
								else
									vprintf ConstantNeighborhoods, 2: "        higher precision needed to ensure mu is constant: %o.\n",c1;
									qthPowerClasses(inf,f,x1,e+1,G1,c1,mu,~W,~im1,~Foundim1,~SpaceGenerated);
								end if;
							else
								vprintf ConstantNeighborhoods, 2: "        higher precision needed to ensure mu is constant.\n";
								qthPowerClasses(inf,f,x1,e+1,G1,c1,mu,~W,~im1,~Foundim1,~SpaceGenerated);
							end if; //#G1 eq 0 or (#G1 eq 1 and #{ g : g in G1 | Degree(g) eq 1} eq 1)
						end if; //ConditionToGoDeeper(fx1,v1,E1)
					end if; //IsWeaklyZero(fx1)
				end for; //r in RepsForF_v
			
			end if; // Ngens lt SpaceGenerated
			
		end procedure; //qthPowerClasses

		t0 := Cputime();
	//	vprintf ConstantNeighborhoods, 1: "    Determining constant neighborhoods for (x-T) on C(K_w) for extension %o.\n",k_v;
		W := {};
		// determines nbhds with x in O_v
		
		qthPowerClasses(false,f,0,0,{ g : g in ff },-1,mu_1,~W,~im1,~Foundim1,~SpaceGenerated);
		// we also need nbhds with 1/x in O_v
		// write d = m*q - s
		// change of vars: z = 1/x, w = y/x^m
		// y^q = f(x)  <===> w^q = z^s*SwapVar(f)(z)
		s := (-Degree(f)) mod q;
		ftilde := SwapVar(f);
		if s gt 0 then
			ftilde *:= Parent(ftilde).1^s;
		end if;
		
		// factorization of ftilde over k_v
		fftilde := { SwapVar(g) : g in ff };
		fftilde := fftilde join { Universe(fftilde).1 };
		// we also include z even if s = 0 for two reasons:
		// 1) so the points with z = 0 are ignored by the procedure qthPowerClasses
		//	(since we can just compute these directly).
		// 2) we use (x - T) = (1/z - T) = z(1-zT), and the other polys in fftilde may only ensure (1-zT)
		//	 has constant q-th power class on the neighborhoods computed.

		W_inf := {};
		qthPowerClasses(true,ftilde,0,1,fftilde,-1,mu_2,~W_inf,~im1,~Foundim1,~SpaceGenerated);
		t1 := Cputime();
	/*
		vprintf ConstantNeighborhoods, 1: " Time = %o.\n", t1-t0;		
		if GetVerbose("ConstantNeighborhoods") ge 1 then
			// print out the constant neighborhoods
			emax := Max({ w[2] : w in W } join {0});
			for e in [1..emax] do
				W_e := { w[1] : w in W | w[2] eq e };
				if #W_e gt 0 then
					printf  "      x + pi^%o*O_v for x in %o\n", e,W_e;
				end if;
			end for;
			
			emax := Max({ w[2] : w in W_inf } join {0} );
			for e in [1..emax] do
				W_einf := { w[1] : w in W_inf | w[2] eq e };
				if #W_einf gt 0 then
					printf  "      z + pi^%o*O_v for z in %o\n", e,W_einf;
				end if;
			end for;
						
		end if; //GetVerbose("ConstantNeighborhoods") ge 1
	*/
		return W join W_inf,im1,Foundim1,SpaceGenerated;

	end function; //LocalImageCk_v

		//printf "before doing anything: ImPic0 = %o, %o.\n", ImPic0, { Eltseq(a) : a in Generators(ImPic0) } ;

		procedure ImagesOfExtensions(d,polys,~im1,~Foundim1,~ImPic0);
			// f are some polys defining extensions of k_v of degree d.
			W_d := {};
			for h in polys do
			
				//vprintf Selmer, 1: "  Currently ImPic0 = %o %o.\n",ImPic0, { Eltseq(a) : a in Generators(ImPic0) };
				if d gt 1 then
					vprintf Selmer, 1: "    Looking for points over extension K_w = k_v(eta) where eta is a root of %o.\n", h;
				else
					vprintf Selmer, 1: "    Looking for points over k_v.\n";
				end if;		
				if Ngens(ImPic0) ge dimim then break; end if;
				f_K,K_w,KtoK_w,mu1,mu2 := TraceOfxminusT(f,h,v,ktok_v);
				Mu1 := func< x | AtoLoc(mu1(x)) >;
				Mu2 := func< x | AtoLoc(mu2(x)) >;		
				W_dh,im1,Foundim1,ImPic0 := LocalImageCk_v(d,f_K,K_w,KtoK_w,Mu1,Mu2,q,im1,Foundim1,ImPic0);
				W_d := W_d join W_dh;
				//vprintf Selmer, 1: "  Now ImPic0 = %o, %o.\n", ImPic0, { Eltseq(a) : a in Generators(ImPic0) };
				
			end for;
		end procedure;
		

		ImagesOfExtensions(1,[kpol.1],~im1,~Foundim1,~ImPic0);

		if not Foundim1 then
			vprintf Selmer, 1: "    Curve is not locally solvable.\n";
			MethodIsValid := false;
		else
			vprintf Selmer, 2: "    Dimension of space generated by mu(C(k_v)) and torsion: %o.\n", Ngens(ImPic0);
		end if;

		
		if q eq 2 and Degree(k) eq 1 then
			// Compute the image of effective divisors of degree 2
			Degree2ExtensionsOfQ_2 := [ x^2+x+1,x^2+2*x+2, x^2+2*x-2,x^2-2, x^2+2,x^2-6,x^2+6 ];
			Degree3ExtensionsOfQ_2 := [ x^3-2,x^3+x+1 ];
			if Ngens(ImPic0) lt dimim then
				vprintf Selmer, 1: "Searching for divisors of degree 2.\n";
				ImagesOfExtensions(2,Degree2ExtensionsOfQ_2,~im1,~Foundim1,~ImPic0);
			end if;
			if Ngens(ImPic0) lt dimim then
				vprintf Selmer, 1: "Searching for divisors of degree 3.\n";
				ImagesOfExtensions(3,Degree3ExtensionsOfQ_2,~im1,~Foundim1,~ImPic0);
			end if;
			// should also write out degree 4,5,6,...
		end if;
		
		// We should do the same for small degree extensions of small primes
		// degree 2,3,4,5, extensions of Q_3
		// degree 2,3,4,5 extensions of Q_3(zeta_3)
		// degree 2,3,4, extensions of Q_5(zeta_5)
		// degree 1,2,3
		
		

		if q eq 3 and Degree(k) eq 2 then
			// k = Q(zeta_3)
zeta_3 := Roots(CyclotomicPolynomial(q),k)[1][1];
T := kpol.1;

Degree2extensionsOfQ_3zeta_3 := [ 
    T^2 + 2*zeta_3+1,
    T^2 + zeta_3 ];

	if Ngens(ImPic0) lt dimim then
		vprintf Selmer, 1: "Searching for divisors of degree 2.\n";
		ImagesOfExtensions(2,Degree2extensionsOfQ_3zeta_3,~im1,~Foundim1,~ImPic0);
	end if;


Degree3extensionsOfQ_3zeta_3 := [
    T^3 + 2*T + 1,
    T^3 + 3*T+ 3,
    T^3 + 6*T+ 3,
    T^3 - 3*T^2+ 21,
    T^3 - 3*T^2+ 3,
    T^3 - 3*T^2+ 12,
    T^3 + 3*T^2+ 3,
    T^3 + 21,
    T^3 + 3*T - 3*zeta_3,
    T^3 + (-3*zeta_3 + 3)*T - 3,
    T^3 + (-3*zeta_3 + 3)*T - 3*zeta_3 - 3,
    T^3 - 3*T^2 - 3*zeta_3,
    T^3 - 3*T^2 - 3*zeta_3 - 3,
    T^3 - 3*T^2 + (-3*zeta_3 + 3)*T + 3,
    T^3 - 3*T^2 + (-3*zeta_3 + 3)*T - 3*zeta_3 - 3,
    T^3 + (-3*zeta_3 + 3)*T^2 - 3*zeta_3,
    T^3 + (-3*zeta_3 + 3)*T^2 + (-3*zeta_3 + 3)*T - 3,
    T^3 - 12*zeta_3,
    T^3 - 12*zeta_3 + 6,
    T^3 - 12*zeta_3 - 3,
    T^3 + 3*zeta_3 + 12,
    T^3 + 6*zeta_3,
    T^3 + 6*zeta_3 - 12,
    T^3 + 6*zeta_3 - 3,
    T^3 + 9*zeta_3 - 12,
    T^3 + 9*zeta_3 + 3,
    T^3 - 2*zeta_3 - 1,
    T^3 + 3*zeta_3*T^2 + (-3*zeta_3 - 3)*T - 4*zeta_3 - 1,
    T^3 + (-zeta_3 + 1)*T + 1,
    T^3 - zeta_3 + 1,
    T^3 - zeta_3 + 4 ];

			if Ngens(ImPic0) lt dimim then
				vprintf Selmer, 1: "Searching for divisors of degree 3.\n";
				ImagesOfExtensions(3,Degree3extensionsOfQ_3zeta_3,~im1,~Foundim1,~ImPic0);
			end if;
		end if; // q eq 3 and Degree(k) eq 2


		if not liesaboveq then
		
			for d in [2..g] do 
				// we want to compute points over tamely ramified extensions of degree d
				if relDeg eq 1 or IsPrime(d) then
					degreedextensions := TamelyRamifiedPolys(d);
					if Ngens(ImPic0) lt dimim then
						_,im1,Foundim1,ImPic0 := ImagesOfExtensions(2,Degree2ExtensionsOfQ_2,im1,Foundim1,ImPic0);
					end if;
				end if;
			end for;
			
		end if;

		
	//
	//	The next few functions are a different way of searching that
	//	don't work as well, but I am reluctant to get rid of yet.
	//
	//
	//		
	//		
	//
			
					
									
	IsDivisorAtLeveln := function(a,b,n,f);
		/* given polynomials a and b with a monic:
		
			returns false,0 if we can prove there is no k_v-divisor on C which reduces to
			{ a = 0 } mod pi^n.
			
			returns true,t if we can prove there is a k_v-divisor on C which reduces to
			{ a = 0 } mod pi^t.
		*/
		//del := Valuation(Resultant(q*kpol.1^(q-1),kpol!b),v);
		del := Valuation(q*(Resultant(a,b)^(q-1)),v);
		r := b^q - f;
		nn := dpolys(r mod a, 0);
		if not nn ge n then
			return false,0;
			// a does not divide b^q-f
		end if;
		s := r div a;
		eps := Valuation(Resultant(s,a),v);
		m:= Min({del,eps});
		//if del lt eps then
		//	printf " \n USING DEL\n del = %o, eps = %o for (a,b) = (%o,%o).\n",del, eps, a,b;
		//end if;
		if nn gt 2*m then
			if n gt nn-m then
				return false,0;
			end if;
		  return true, nn-m;
			// There is a factorization congruent to this mod pi^(nn-m)
		end if;
		if  nn le 2*m and nn lt 2*n then
			return false,0;
		end if;
		return true,0;
		// we don't know if the factorization lifts (need to go deeper to determine)
	end function; //IsDivisorAtLeveln
	
	function DivisorLifts(a,bs,n,nbhda,nbhdb,f);
		/* A recursive version of the previous function.*/
		for b in bs do
			Lifts,level := IsDivisorAtLeveln(a,b,n,f);
			if Lifts then
				if level ge n then
					return true;
				else
					NewPairs := < < a + pi^n*h, <b + pi^n*h : h in nbhdb > > : h in nbhda > ;
				  for newabs in NewPairs do
						NextLevel := DivisorLifts(newabs[1],newabs[2],n+1,nbhda,nbhdb,f);
						if NextLevel then
							return true;
						end if;
					end for;
				end if;
			end if;
		end for;
		return false;
	end function; //DivisorLifts
	
	
	SizeConstantNbhd2 := function(a,d);
		

		vals := [];
		for j in [1..NumberOfComponents(Aflds)] do
			pidsj := pids[j];
			j;

			for pid in pidsj do
				Ki, toKi := Completion(Aflds[j],pid[1] : Precision := k_v`DefaultPrecision*pid[2]);
				theta := toKi(AtoAflds(A.1)[j]);
				

				
				Ki`DefaultPrecision +:= 200;
					while true do
						a_w := Polynomial([ toKi(c) : c in Coefficients(a)]);
						try
							fa := Factorization(a_w);
		
							no_error:=true;
						catch ERR
	
							if IsFactorizationPrecisionError(ERR) then
								no_error:=false;
							else
								error ERR;
							end if;
						end try;
						if no_error then

							break;
						end if;

						Ki`DefaultPrecision *:= 2;
						vprint Selmer, 3: "    Finding roots for constant neighborhood requires: increasing precision to %o", O_v`DefaultPrecision;
						if Ki`DefaultPrecision gt 10000 then

						end if;
					end while;


				vals_pid := Max([ Valuation(Evaluate(a_i[1],theta))/Degree(a_i[1]) : a_i in fa ]);
				vals cat:= [vals_pid];
			end for;

		end for;

		if d gt 1 then
			Phikflex := d*Valuation(Discriminant(a),v)/(2*(d-1));
		else
			Phikflex := Infinity();
		end if;
		
		
		return vals, Phikflex;
	
	end function; // SizeConstantNbd2

	// Do we need to consider points with z = O(pi)?
	InfiniteHood := extraval;
	if degree mod q ne 0 then
		InfiniteHood *:= 1/q;
	end if;
	//to compute the image of points (x:*:pi*z) we need to know them to precision > InfiniteHood.
	if InfiniteHood ge 1 and Degree(f) mod q eq 0 then
		im_inf := true;
	else
		im_inf := false;
	end if;
	
	
	function SystematicCheck(f,n,Polys,ImPic0,im1,nbhda,nbhdb,d,degd,cofactors,SizeTorsionNbhds,inf);

	//	printf " now at level %o.\n", n;
		for g in Polys do
			a := Evaluate(g[1],kpol.1);
			continuewitha := true;
			if inf then
				SizeConstantNbhd1 := Max(&cat[[InfiniteHood + (Valuation(AtoAflds(Evaluate(SwapVar(a),kpol.1))[i],prime[1]))/prime[2] : 
															prime in pids[i] ] : i in [1..NumberOfComponents(Aflds)] ]);
			else
				SizeConstantNbhd1 := Max(&cat[[extraval + (Valuation(AtoAflds(Evaluate(a,A.1))[i],prime[1]))/prime[2] : 
															prime in pids[i] ] : i in [1..NumberOfComponents(Aflds)] ]);
			end if;
		//	printf " considering a = %o, b in %o. SC1 = %o.\n", a, g[2], SizeConstantNbhd1;
			if n gt SizeConstantNbhd1 then
				// Anything nearby has the same image.
				if not inf then
					im := AtoLoc((-1)^d*Evaluate(a,A.1))-d*im1;
				else 
					if Degree(f) mod q ne 0 then
						error if true, " can't compute image in this case: d ne 0 mod q and infinite patch...";
					else
						im := AtoLoc(Evaluate(SwapVar(a),A.1))-d*im1;
					end if;
				end if;
				if not im in ImPic0 then
					// we should continue (otherwise not)

					if DivisorLifts(a,g[2],n,nbhda,nbhdb,f) then
						ImPic0 := sub< Avqmodk_v | Generators(ImPic0) join {im}>;
						printf "    Found new image: mu(%o) = %o.\n", a,im;
			//			printf "    SizeNbds1 = %o.\n", SizeConstantNbhd1;
						printf "    Dim(Image) = %o.\n", #Generators(ImPic0);
						
						
						if NumberOfGenerators(ImPic0) gt dimim then
							
							printf "\n\n\n\n\n There was a problem.\n   n = %o.\n",n;
							{} eq Type("");
						end if;
						
						if Ngens(ImPic0) eq dimim then
							vprintf Selmer, 2: " Local image computed...\n"; 
							break;
						end if;
						/*
						
							that := SwapVar(a); this:=  AtoLoc(Evaluate(that,A.1));
		//					printf " that = %o.\n its image is %o.\n", that,this;
							
							for i in [1..10] do
								a2 := a + pi^n*Random(nbhda);
								if not inf then
									im := AtoLoc((-1)^d*Evaluate(a2,A.1))-d*im1;
								else 
									if Degree(f) mod q ne 0 then
										printf " can't compute image in this case: d ne 0 mod q and infinite patch...";
										Type({}) eq "";
									else
										im := AtoLoc(Evaluate(SwapVar(a2),A.1))-d*im1;
									end if;
								end if;
		//						printf "    Found new image: mu(%o) = %o.\n", a2,im;
							end for;
							
							
							
							*/
							

						
						
					else
				//	printf " pair does not lift to level n %o.\n", n;
					end if;
				else
			//		printf " mu(%o) - d*im1 = %o.\n",a, im;
				//	printf "    SizeNbds1 = %o.\n", SizeConstantNbhd1;
				end if; // not im in ImPic0
			else // n gt SizeConstantNbhd1

				// see if a is near torsion.
				neartors := false;
				if #degd gt 0 then
					Distances := [ dpolys(a,f_i) : f_i in degd ];
					neartors := Max([ Distances[j] - SizeTorsionNbhds[j] : j in [1..#degd] ]) gt 0;
				end if;
				if neartors then
					// image is torsion
		//			printf "    Image is torsion: %o.\n", a;
		//			printf "    SizeTorsionNbhds = %o\n", SizeTorsionNbhds;
		//			printf "    Distances = %o.\n", Distances;
				else
				
				
				
				/*  ConstantNbhd2 is a bad idea.
					if not inf then
						// Check the factorization over K_w to get smaller nbhds
						"here 3";
						ConstantNbhd2,Phinflex := SizeConstantNbhd2(a,d);
						ConstantNbhd2 := extraval + Max(ConstantNbhd2);
						"here 4";
						if n ge Phinflex then
							Phin := n - (d-1)/d*Phinflex;
						else
							Phin := n/d;
						end if;
					else
						Phin := 0; ConstantNbhd2 := 1;
					end if;
					
				//	printf " Second size is Phin = %o <> ConstantNbhd2 = %o .\n",Phin,ConstantNbhd2;
					if Phin gt ConstantNbhd2 then
			//			printf "Using second constant nbhd.\n";
						// Anything nearby has the same image.
						if not inf then
							im := AtoLoc((-1)^d*Evaluate(a,A.1))-d*im1;
						else 
							if Degree(f) mod q ne 0 then
								printf " can't compute image in this case: d ne 0 mod q and infinite patch...";
								Type({}) eq "";
							else
								im := AtoLoc(Evaluate(SwapVar(a),A.1))-d*im1;
							end if;
						end if;
						if not im in ImPic0 then
							// we should continue (otherwise not)
							if DivisorLifts(a,g[2],n,nbhda,nbhdb,f) then
								ImPic0 := sub< Avqmodk_v | Generators(ImPic0) join {im}>;
								
								if Ngens(ImPic0) eq dimim then
									vprintf Selmer, 2: " Local image computed...\n"; 
									break;
								end if;
								
								printf "    Found new image (2): (x-T)(%o) = %o.\n", a,im;
							else
				//				printf " pair does not lift to level n %o.\n", n;
							end if;
						end if; // not im in ImPic0
					else
					
					*/
						// We may need to go deeper...
						fa := Factorization(Polytok_v(a));
						if #fa gt 1 then
							rd := Valuation(Resultant(fa[1][1],fa[2][1]));
							if n - 2*rd gt 0 then
	//							printf " factorizes ";
							else
	//							printf " factorizes no good";
							end if;
						else
										
						NewAs := [ a + pi^n*h : h in nbhda ];
						NewPairs := [];
						for a in NewAs do
							for b in g[2] do 
								NewBs := [ b + pi^n*h : h in nbhdb ];
								bsa := [ b : b in NewBs | IsDivisorAtLeveln(a,b,n+1,f) ];
								if #bsa gt 0 then
									NewPairs cat:= [ <a,bsa> ];
								end if;
							end for;
						end for;	
				//		printf " going deeper on (%o,%o).\n", a,g[2];
						if #NewPairs gt 0 then
							ImPic0 := SystematicCheck(f,n+1,NewPairs,ImPic0,im1,nbhda,nbhdb,d,degd,cofactors,SizeTorsionNbhds,inf);
						else
			//				printf " no lifts";
						end if;
		
					end if; // n gt SizeConstantNbhd2
				end if; // Max([ dpolys(a,f_i)-s : s in SizeTorsionNbhds, f_i in degd ]) gt 0
			end if; // n gt SizeConstantNbhd1
			
		end for;
		return ImPic0;
		
	end function; // SystematicCheck
				

	function RandomDiv(f,n,Polys,ImPic0,nbhda,nbhdb,d,degd,cofactors,SizeTorsionNbhds,inf);

	//	printf " now at level %o.\n", n;
		while #Polys gt 0 do
			ind := Random(#Polys-1)+ 1;
			g := Polys[ind];
			Remove(~Polys,ind);
	//  for g in Polys do
			a := Evaluate(g[1],kpol.1);
			continuewitha := true;
			if inf then
				SizeConstantNbhd1 := Max(&cat[[InfiniteHood + (Valuation(AtoAflds(Evaluate(SwapVar(a),kpol.1))[i],prime[1]))/prime[2] : 
															prime in pids[i] ] : i in [1..NumberOfComponents(Aflds)] ]);
			else
				SizeConstantNbhd1 := Max(&cat[[extraval + (Valuation(AtoAflds(Evaluate(a,A.1))[i],prime[1]))/prime[2] : 
															prime in pids[i] ] : i in [1..NumberOfComponents(Aflds)] ]);
			end if;			
			//printf " considering a = %o, b in %o. SC1 = %o, n = %o.\n", a, g[2], SizeConstantNbhd1,n;
			
			if n gt SizeConstantNbhd1 then
				// Anything nearby has the same image.
				if DivisorLifts(a,g[2],n,nbhda,nbhdb,f) then
						if not inf then
							im := AtoLoc((-1)^d*Evaluate(a,A.1));
						else 
							if Degree(f) mod q ne 0 then
								printf " can't compute image in this case: d ne 0 mod q and infinite patch...";
								Type({}) eq "";
							else
								im := AtoLoc(Evaluate(SwapVar(a),A.1));
							end if;
						end if;
					//	printf "  Found new image !!!: mu(%o) = %o.\n", a,im;
						return im;
				end if;
			else // n gt SizeConstantNbhd1
				// see if a is near torsion.
				neartors := false;
				if #degd gt 0 then
					Distances := [ dpolys(a,f_i) : f_i in degd ];
					DminusT,ind := Max([ Distances[j] - SizeTorsionNbhds[j] : j in [1..#degd] ]);
					neartors := DminusT gt 0;
				end if;
				if neartors then
					// image is torsion
			//		printf "    Image is torsion: %o.\n", a;
			//		printf "    SizeTorsionNbhds = %o\n", SizeTorsionNbhds;
					im := AtoLoc(Evaluate(degd[ind]-cofactors[ind]^(q-1),A.1));
					return "";
				else
			
					
					/*		
					// Check the factorization over K_w to get smaller nbhds
					if not inf then
												
			
						ConstantNbhd2,Phinflex := SizeConstantNbhd2(a,d);
											
						ConstantNbhd2 := extraval + Max(ConstantNbhd2);
						if n ge Phinflex then
							Phin := n - (d-1)/d*Phinflex;
						else
							Phin := n/d;
						end if;

				
					else
						ConstantNbhd2 := 1;
						Phin := 0;
					end if;
					
		//			printf " Second size is Phin = %o <> ConstantNbhd2 = %o .\n",Phin,ConstantNbhd2;
					if Phin gt ConstantNbhd2 then
			//			printf "Using second constant nbhd.\n";
						// Anything nearby has the same image.
						if DivisorLifts(a,g[2],n,nbhda,nbhdb,f) then
							if not inf then
								im := AtoLoc((-1)^d*Evaluate(a,A.1));
							else 
								im := AtoLoc(Evaluate(SwapVar(a),A.1));
							end if;						
			//				printf "  Found new image: mu(%o) = %o.\n", a,im;
							return im;
						end if;
					else
				*/
						
								
					// We may need to go deeper...
				
						fa := Factorization(Polytok_v(a));
						if #fa gt 1 then
							rd := Valuation(Resultant(fa[1][1],fa[2][1]));
							if n - 2*rd gt 0 then
						//		printf " factorizes ";
							else
						//		printf " factorizes no good";
							end if;
						else
										
						NewAs := [ a + pi^n*h : h in nbhda ];
						NewPairs := [];
						for a in NewAs do
							for b in g[2] do 
								NewBs := [ b + pi^n*h : h in nbhdb ];
								bsa := [ b : b in NewBs | IsDivisorAtLeveln(a,b,n+1,f) ];
								if #bsa gt 0 then
									NewPairs cat:= [ <a,bsa> ];
								end if;
							end for;
						end for;	
		//			printf " going deeper on (%o,%o).\n", a,g[2];
						if #NewPairs gt 0 then
							return RandomDiv(f,n+1,NewPairs,ImPic0,nbhda,nbhdb,d,degd,cofactors,SizeTorsionNbhds,inf);

						else
			//				printf " no lifts";
						end if;
				//		end if;
		
					end if; // n gt SizeConstantNbhd2
				end if; // Max([ dpolys(a,f_i)-s : s in SizeTorsionNbhds, f_i in degd ]) gt 0
			end if; // n gt SizeConstantNbhd1
			
		end while;
		
		return "";
	end function; // RandomDiv
	
	
	//W

	//For finite patch:
	degds := [];
	cofactds := [];
	STNBhds := [];
	degls := < [ f_i : f_i in ff | Degree(f_i) eq i ] : i in [1..g+1] >;
	
	for j in [1..g+1] do
		Partj := Partitions(j);
		degj := Setseq({ &*gc : gc in CartesianProduct([ degls[i] : i in Partj[k]]) , k in [1..#Partj] | #{ gc[j] : j in [1..#gc]} eq #Partj[k] });
		cofactors := [ (O_k!lcf)*Polytok(Polytok_v(f) div g) : g in degj ];
		degj := [ Polytok(g) : g in degj ];
		SizeTorsionNbhdsj := [];
		for s in [1..#degj] do
			SizeTorsionNbhdsj cat:= [Max(&cat[[extraval + (Valuation(AtoAflds(Evaluate(degj[s]+cofactors[s]^(q-1),A.1))[i],prime[1]))/prime[2] : 
														prime in pids[i] ] : i in [1..NumberOfComponents(Aflds)]  ])];
		end for;
		
		degds cat:= [degj];
		cofactds cat:= [cofactors];
		STNBhds cat:= [SizeTorsionNbhdsj];
	end for;	
	
	
	if im_inf then
		//perhaps also problems when v is ramified in k?
		// for computing images of divisors on the patch near infinity (usually not necessary...
		f_inf := SwapVar(f);
		degds_inf := [];
		for degd in degds do
			degd_inf := [];
			for ddd in degd do
				dd := SwapVar(ddd);
				if Valuation(LeadingCoefficient(dd),v) eq 0 then
					dd := Polytok_v(dd);
					degd_inf cat:= [Polytok(dd/LeadingCoefficient(dd))];
				else
					degd_inf cat:= [dd];
				end if;
			end for;
			degds_inf cat:= [degd_inf];
		end for;
	else
		degds_inf := degds;
		f_inf := f;
	end if;

	// For small residue fields start with the good random search.
	for d in [2..g+1] do
		Tries := 0;
		if #F_v^d le 100 then			
			// Write down the polynomials of smaller degree (not so good for large v)
			Polys := [];
			polysF := [ &+[ A[i]*FF.1^(d-i) : i in [1..d] ] : A in CartesianProduct([F_v : i in [1..d] ]) ];
			for a in polysF do
				bsa := [ Polymodvtok(b) : b in polysF | (b^q - f_F) mod (FF.1^d+a) eq 0 ];
				if #bsa gt 0 then
					Polys cat:= [ < Polymodvtok(FF.1^d+a), bsa > ];
				end if;
			end for;
			
			polysk := [ Polynomial([ O_vtoO_k(c @@ modv) : c in Coefficients(h)]) : h in polysF | h ne 0] cat [kpol!0];
			tries := 0;
			
			dimfound := Ngens(ImPic0);
			
			if dimfound eq dimim then
				if #F_v^d le 27 then
				Tries := 0;
				else 
				Tries := 30;
				end if;
			end if;
			
			if Ngens(ImPic0) lt dimim or Tries lt 30/d^3 then
				vprintf Selmer, 2: "    Searching for random divisors of degree %o (first method).\n", d;
			end if;
			
			while Ngens(ImPic0) lt dimim do
				//printf " Tries = %o.\n",Tries;
				if Tries ge 100/d then
					break;
				end if;
				im := RandomDiv(f,1,Polys,ImPic0,polysk,polysk,d,degds[d],cofactds[d],STNBhds[d],false);
				if Type(im) eq MonStgElt then
				else
					ImPic0 := sub<Avqmodk_v | Generators(ImPic0) join {im - d*im1} >;
					if Ngens(ImPic0) gt dimfound then
						vprintf Selmer, 2:  "       Found independent image: %o.\n", Eltseq(im - d*im1);
						vprintf Selmer, 2: "           ----> DimensionIsNow = %o.\n", NumberOfGenerators(ImPic0);
						dimfound +:= 1;
					end if;
				end if;
				Tries +:= 1;
				
			end while;
			
			Tries := 0;
			if im_inf then // check infinite patch as well:
				f_F_inf := Polynomial([ modv(O_ktoO_v(O_k!c)) : c in Coefficients(f_inf) ]);
				Polys := [];
				for a in polysF do
					bsa := [ Polymodvtok(b) : b in polysF | (b^q - f_F_inf) mod (FF.1^d+a) eq 0 ];
					if #bsa gt 0 then
						Polys cat:= [ < Polymodvtok(FF.1^d+a), bsa > ];
					end if;
				end for;
				
				dimfound := Ngens(ImPic0);
			
				if dimfound eq dimim then
					if #F_v^d le 27 then
						Tries := 0;
					else 
						Tries := 30;
					end if;
				end if;
			
				if Ngens(ImPic0) lt dimim or Tries lt 30/d^3 then
					vprintf Selmer, 2: "    Searching for random divisors of degree %o (first method).\n", d;
				end if;

				while #Generators(ImPic0) lt dimim do
					if Tries ge 30/d then 
						break;
					end if;
					im := RandomDiv(f_inf,1,Polys,ImPic0,polysk,polysk,d,degds_inf[d],cofactds[d],STNBhds[d],true);
					if Type(im) eq MonStgElt then
					else
						ImPic0 := sub<Avqmodk_v | {im - d*im1} join Generators(ImPic0) >;
						if Ngens(ImPic0) gt dimfound then
							vprintf Selmer, 2:  "       Found independent image: %o.\n", Eltseq(im - d*im1);
							vprintf Selmer, 2: "           ----> DimensionIsNow = %o.\n", NumberOfGenerators(ImPic0);
							dimfound +:= 1;
						end if;
					end if;
					Tries +:= 1;
				end while; //#Generators(ImPic0) lt dimim
				
			end if; // im_inf
				
		else
			
			
			// residue field is maybe too big to handle this way?
			Lastd := d; 
			break;
			
		end if;
		
		if d eq g then
			Lastd := g+1;
		end if;
		
	end for;

	// If that didn't work, we should do the systematic search for small residue fields.
	if dimim gt #Generators(ImPic0) then
		
		for d in [2..g+1] do
			if #F_v^d le 100 then		
				// Write down the polynomials of smaller degree (not so good for large v)
				// this is redundant
				Polys := [];
				polysF := [ &+[ A[i]*FF.1^(d-i) : i in [1..d] ] : A in CartesianProduct([F_v : i in [1..d] ]) ];
				for a in polysF do
				
					bsa := [ Polymodvtok(b) : b in polysF | (b^q - f_F) mod (FF.1^d+a) eq 0 ];
					if #bsa gt 0 then
						Polys cat:= [ < Polymodvtok(FF.1^d+a), bsa > ];
					end if;
				end for;

				polysk := [ Polynomial([ O_vtoO_k(c @@ modv) : c in Coefficients(h)]) : h in polysF | h ne 0] cat [kpol!0];
				vprintf Selmer, 2: "     Systematically searching divisors of degree %o.\n", d;
				
				
				ImPic0 := SystematicCheck(f,1,Polys,ImPic0,im1,polysk,polysk,d,degds[d],cofactds[d],STNBhds[d],false);
				
				vprintf Selmer, 2: "      Dimension is now %o.\n", Ngens(ImPic0);
				
				if Ngens(ImPic0) ge dimim then
					break;
				end if;
								
				if im_inf then // check infinite patch as well:				
					f_F_inf := Polynomial([ modv(O_ktoO_v(O_k!c)) : c in Coefficients(f_inf) ]);
					Polys := [];
					for a in polysF do
						bsa := [ Polymodvtok(b) : b in polysF | (b^q - f_F_inf) mod (FF.1^d+a) eq 0 ];
						if #bsa gt 0 then
							Polys cat:= [ < Polymodvtok(FF.1^d+a), bsa > ];
						end if;
					end for;

					vprintf Selmer, 2: "     Systematically searching divisors of degree %o supported near infinity.\n\n", d;
					ImPic0 := SystematicCheck(f_inf,1,Polys,ImPic0,im1,polysk,polysk,d,degds_inf[d],cofactds[d],STNBhds[d],true);
				
				if Ngens(ImPic0) ge dimim then
					break;
				end if;
				
				end if; // im_inf
		
		  else
				
				// residue field is maybe too big to handle this way?
				Lastd := d; 
				break;
			
				
		  end if; // #F_v^d lt 100
		
		  if d eq g then
			  Lastd := g+1;
		  end if;	
				
		end for; // d in [1..g]
		
	end if; // dimim gt NGens(ImPic0); 
	
	
	// What remains are large residue fields.
	// we scratch together a (crappy) function for finding random divisors of degree d
		
	FFF := PolynomialRing(FF); // F(x)(y)
	to3F := hom< FF -> FFF | FFF.1 >;
	
	// Used to check if h(x) = 0 defines a divisor on C mod v.
	CheckResInResidueField := function(h,fmodv);
		resfact := Factorization(Resultant(to3F(h),FF.1^q-to3F(fmodv)));
		for ff in resfact do
			cfs := Coefficients(ff[1]);
			cs := { i mod q : i in [1..#cfs] | cfs[i] ne 0 };
			if cs eq {1} then return false; end if;
		end for;
		return true;
	end function;		
	
	RandomModv := function(d,inf);
				
		//returns a function which finds Random divisors...
		// if inf we look for divisors supported on x = 1.
		// else we look for divisors suppoted on z = 1.
		if not inf then
			degd := degds[d];
			degd_0 := degds[d];
		else
			// look in the infinite patch...
			degd := degds_inf[d];
			degd_0 := degds[d];
		end if;
		cofactors := cofactds[d];
		SizeTorsionNbhds := STNBhds[d];
		
		if inf then
			f_FF := SwapVar(f_F);
			f_k := f_inf;
		else
			f_FF := f_F;
			f_k := f;
		end if;
		
		RDIV := function();
			n := 0;
			while true do		
			if n eq 0 then
				a_0 := FF.1^d + &+[ Random(F_v)*FF.1^(d-i) : i in [1..d] ];
				while {1} ne { fa[2] : fa in Factorization(a_0) } or not CheckResInResidueField(a_0,f_FF) do
					a_0 := FF.1^d + &+[ Random(F_v)*FF.1^(d-i) : i in [1..d] ];
				end while;
				// a is square free mod v and defines a divisor.
				b_0 := &+[ Random(F_v)*FF.1^(d-i) : i in [1..d] ];
				while (f_FF-b_0^q) mod a_0 ne 0 do
					b_0 := &+[ Random(F_v)*FF.1^(d-i) : i in [1..d] ];
				end while;
				
				s_0 := (f_FF-b_0^q) div a_0;
				a := Polymodvtok(a_0);
				b := Polymodvtok(b_0);
				s := Polymodvtok(s_0);
				n +:= 1;
			end if;
			
			nn := dpolys(a*s-(f_k-b^q),0);
			eps := Valuation(Resultant(a,s),v);
			if inf then
				SizeConstantNbhd1 := Max(&cat[[InfiniteHood + (Valuation(AtoAflds(Evaluate(SwapVar(a),kpol.1))[i],prime[1]))/prime[2] : 
															prime in pids[i] ] : i in [1..NumberOfComponents(Aflds)] ]);
			else
				SizeConstantNbhd1 := Max(&cat[[extraval + (Valuation(AtoAflds(Evaluate(a,A.1))[i],prime[1]))/prime[2] : 
															prime in pids[i] ] : i in [1..NumberOfComponents(Aflds)] ]);
			end if;			DistancesToTorsion := [ dpolys(a,degd[j]) : j in [1..#degd] ];
			
	//	printf " now at a = %o (n,nn,eps,constantnbhd) = (%o,%o,%o,%o).\n",Evaluate(a,A.1), n,nn,eps,SizeConstantNbhd1;
			if SizeConstantNbhd1 gt extraval + 5*e_v then
				nn := -1;
				n := 0;
			end if;
			
			if nn gt 2*eps then
				if n gt nn - eps then // start over...
					n := 0;
				else
					n := nn - eps; // jump to level nn-eps (lift should be unique from here on?)
					//reduce a and s mod p^n
					if inf then
						SizeConstantNbhd1 := Max(&cat[[InfiniteHood + (Valuation(AtoAflds(Evaluate(SwapVar(a),kpol.1))[i],prime[1]))/prime[2] : 
																	prime in pids[i] ] : i in [1..NumberOfComponents(Aflds)] ]);
					else
						SizeConstantNbhd1 := Max(&cat[[extraval + (Valuation(AtoAflds(Evaluate(a,A.1))[i],prime[1]))/prime[2] : 
																	prime in pids[i] ] : i in [1..NumberOfComponents(Aflds)] ]);
					end if;
					if n gt SizeConstantNbhd1 then
						if not inf then
							im := AtoLoc(Evaluate((-1)^d*a,A.1));
						else
							im := AtoLoc(Evaluate(SwapVar(a),A.1));
						end if;
						printf " mu(%o) = %o.\n",a,im;
						return im;
					end if;
					
					if n gt Min(SizeTorsionNbhds) then
						DistancesToTorsion := [ dpolys(a,degd[j]) : j in [1..#degd] ];
						if exists(j){ j : j in [1..#SizeTorsionNbhds] | DistancesToTorsion[j] - SizeTorsionNbhds[j] gt 0 } then
							im := AtoLoc(Evaluate(degd_0[j]-cofactors[j]^(q-1),A.1));
							printf " image is torsion: mu(%o) = %o.\n", a, im;
							//return AtoLoc(Evaluate(degd[j]-cofactors[j]^(q-1),A.1));
							n := 0;
							nn := -1;
						end if;
					end if;
						if nn gt 0 then
					
						// this GCD needs to be computed over k?
						gcd,c1,c2 := XGCD(a_0,s_0);
						//printf " gcd was... %o", gcd;
						r := PolytoF_v(-(pi^-n)*(a*s-(f_k-b^q)));
						atilde := (r*c2 mod a_0);
						stilde := (r*c1 + s_0*(r*c2 div a_0));
						a +:= pi^n*Polymodvtok(atilde);
						s +:= pi^n*Polymodvtok(stilde);
						n +:= 1;
						
						end if;
				end if;
			else
				if nn le 2*eps and nn lt 2*n and eps ne Infinity() then //start over...
					n := 0;
				else
					// move forward randomly... (probably should try several times before giving up...)
					tries := 0;
					while tries lt p^d and dpolys(a*s-(f_k-b^q),0) lt nn+1 do
//						printf " nn is now %o.\n",nn;
						a +:= pi^n*Polymodvtok(&+[ Random(F_v)*FF.1^(d-i) : i in [1..d] ]);
						b +:= pi^n*Polymodvtok(&+[ Random(F_v)*FF.1^(d-i) : i in [1..d] ]);
						s := (f_k-b^q) mod a;
						tries +:= 1;
//						printf "%o",tries;
					end while;
					n +:= 1;
				end if;
			end if;
			
			end while;
		end function;
		
		return RDIV;
//	printf "res_0 = %o.\n", Resultant(a_0,s_0);

	end function;
	
	dimfound := Ngens(ImPic0);
	
	if Ngens(ImPic0) lt dimim and Lastd le g then
		// We have done systematic search for small residue fields
		// and divisors of degree up to Lastd-1
				
		RDIVS := [ RandomModv(j,false) : j in [Lastd..g] ];
		if im_inf then
			RDIVS_inf := [RandomModv(j,true) : j in [Lastd..g]];
		end if;
		
		if Ngens(ImPic0) lt dimim then
			vprintf Selmer, 2: "     Looking for random divisors of degree %o,...,%o.\n",Lastd,g;
		end if;
		
		tries := 0;
		while Ngens(ImPic0) lt dimim do
			
			j := Random(#RDIVS-1)+1;
			if im_inf and Random(1) eq 0 then
				im := RDIVS_inf[j]() - j*im1;
			else
				im := RDIVS[j]() - j*im1; 
			end if;
			
						ImPic0 := sub<Avqmodk_v | Generators(ImPic0) join {im}>;
						if Ngens(ImPic0) gt dimfound then
							vprintf Selmer, 2:  "       Found independent image: %o.\n", Eltseq(im);
							vprintf Selmer, 2: "           ----> DimensionIsNow = %o.\n", NumberOfGenerators(ImPic0);
							dimfound +:= 1;
						end if;

			tries +:= 1;
			error if tries gt 20*g, "Couldn't find local image. Gave up.";
		
		end while;	
		
	end if;
		
		// At this point we could do a systematic search for larger degree but that is too slow.
		/*
		//For what remains we do the systematic search.... may take forever...
		if dimim gt #Generators(ImPic0) then
		
		for d in [1..g] do
			// Write down the polynomials of smaller degree (not so good for large v)
			Polys := [];
			polysF := [ &+[ A[i]*FF.1^(d-i) : i in [1..d] ] : A in CartesianProduct([F_v : i in [1..d] ]) ];
			for a in polysF do
			
				bsa := [ Polymodvtok(b) : b in polysF | (b^q - f_F) mod (FF.1^d+a) eq 0 ];
				if #bsa gt 0 then
					Polys cat:= [ < Polymodvtok(FF.1^d+a), bsa > ];
				end if;
			end for;

			polysk := [ Polynomial([ O_vtoO_k(c @@ modv) : c in Coefficients(h)]) : h in polysF | h ne 0] cat [kpol!0];
			vprintf Selmer, 2: "     Systematically searching divisors of degree %o.\n", d;
			
			printf " I have found a space of dimension %o, but I will continue anyway....\n", Ngens(ImPic0);
			ImPic0 := SystematicCheck(f,1,Polys,ImPic0,im1,polysk,polysk,d,degds[d],cofactds[d],STNBhds[d],false);
						
			if im_inf then // check infinite patch as well:				
				f_F_inf := Polynomial([ modv(O_ktoO_v(O_k!c)) : c in Coefficients(f_inf) ]);
				Polys := [];
				for a in polysF do
					bsa := [ Polymodvtok(b) : b in polysF | (b^q - f_F_inf) mod (FF.1^d+a) eq 0 ];
					if #bsa gt 0 then
						Polys cat:= [ < Polymodvtok(FF.1^d+a), bsa > ];
					end if;
				end for;

				vprintf Selmer, 2: "     Systematic searching divisors of degree %o supported near infinity.\n\n", d;
				ImPic0 := SystematicCheck(f_inf,1,Polys,ImPic0,im1,polysk,polysk,d,degds_inf[d],cofactds[d],STNBhds[d],true);
			end if; // im_inf			
		end for;
	end if;
	
	end if;
	*/

		
	error if dimim gt Ngens(ImPic0), "Failed to compute local image.";
	error if dimim lt Ngens(ImPic0), "Found spurious points!";
	vprintf Selmer, 2: " Done with local computations at v = %o.\n\n",p;
	

	//Now we want to find the preimages of ImPic0 and ImPic1p in ASqQ.
	Avqmodk_vmodImPic0, modImPic0 := quo<Avqmodk_v|ImPic0>;
	comp := hom<ASqmodk -> Avqmodk_vmodImPic0 | [ modImPic0(res_v(g)) : g in OrderedGenerators(ASqmodk) ] >;
	
	ProperPrimeToq := true;
	d1 := im1;
	
	LC0 := Kernel(comp); // The subgroup of d in ASqmodk satisfying the local condition: res_v(d) in ImPic0
	vprintf Selmer, 3: "  Local conditions on A(S,q)/k(S,q) cut out a space of dimension %o.\n",#Generators(LC0);
	if ProperPrimeToq then	
		boo,preim := HasPreimage(modImPic0(d1-res_v(alpha_c)),comp);
		if boo then
			LC1 := < Kernel(comp),preim + alpha_c>;
			vprintf Selmer, 3: "  Pic1 coset is nonempty.\n";
		else
			LC1 := "empty";
			vprintf Selmer, 3: "  Pic1 coset is empty.\n";
		end if;
		//LC1 represents the coset of d in ASqmodk satisfying the local condition: res_v(d) in ImPic1
	else	
		LC1 := "empty";
	end if;

	return LC0,LC1;
end function; //LocalImages;


LocalImagesReals := function(pl);

    vprintf Selmer, 3 :"    Looking at %o.\n",pl;

    emb:=RealExtensions(A,pl);
		
		if # emb gt 0 then
			rts:=[e(A.1):e in emb];

			RH1:=AbelianGroup([2:i in rts]);
			AtoRH1:=map<A->RH1|
				a:->RH1![Real(cnj(a)) gt 0 select 0 else 1: cnj in emb]>;
				
			fR := Polynomial( [ emb[1](a) : a in Coefficients(f) ]);
			if #rts le 1 then
				eps := 1;
			else
				eps:=Minimum([Abs(rts[i]-rts[j]): i,j in [1..#rts] | i lt j])/1000;
			end if;
			if #rts eq 0 then
				Rimg:=sub<RH1|Identity(RH1)>;
				if Evaluate(fR,0) gt 0 then
					d1 := AtoRH1(-A.1); // Image of a real point under the descent map.
				else
					Pic1kEmpty := true;
					vprintf Selmer, 2: "     Curve has no real points.\n";
				end if;
			else
				if #rts eq 1 then
					L:=[[1]];
				else
					L:=[[ rt+eps gt r select 0 else 1 :r in rts]:rt in rts];
				end if;
				if Real(emb[1](A!lcf)) gt 0 then
					L:=[RH1!l: l in L| IsEven(#[c: c in l | c eq 1])];
				else
					L:=[RH1!l: l in L| IsOdd(#[c: c in l | c eq 1])];
				end if;
				if IsOdd(Degree(f)) then
					Append(~L,AtoRH1(lcf));
				end if;
				L:=[l-L[1]:l in L];
				if IsEven(Degree(f)) then
					Append(~L,RH1![1:i in [1..#rts]]);
				end if;
				
				
				pos := [ g + eps : g in rts | Evaluate(fR,g+eps) gt 0 ] cat [ g - eps : g in rts | Evaluate(fR,g-eps) gt 0];
				if #pos eq 0 then
					Pic1kEmpty := true;
					vprintf Selmer, 2: "     Curve has no real points.\n";
					// although there could be a real Weierstrass pt.
				else
					pre := Precision(Parent(rts[1]));
					approx := Floor(10^pre*pos[1])/10^pre; //rational approximation to rts[1]+eps
					d1 := AtoRH1(-A.1+A!approx); // Image of a real point under the descent map.
				end if;
				
				Rimg:=sub<RH1| [l:l in L]>;
		
			end if;
			
			LR2mRmodImJR,mR := quo< RH1 | Rimg >; 
			comp := hom<ASqmodk -> LR2mRmodImJR | [ mR(AtoRH1(g)) : g in repAs ] >;
			LC0R := Kernel(comp);
			
			vprintf Selmer, 3: "    Local conditions on A(S,q)/k(S,q) cut out a space of dimension %o.\n",#Generators(LC0R);
			boo,preim := HasPreimage(mR(d1) - comp(alpha_c),comp);
			if boo then
				LC1R := < Kernel(comp),preim + alpha_c>;
				vprintf Selmer, 3: "    Pic1 coset is nonempty.\n";
			else
				LC1R := "empty";
				vprintf Selmer, 3: "    Pic1 coset is empty.\n";
			end if;
			vprintf Selmer, 2: "  Done with local computations at p = infty.\n\n";
		else
			vprintf Selmer, 2: "  No real embeddings.\n";
			LC0R := ASqmodk;
			LC1R := <ASqmodk,alpha_c>;
		end if;

	return LC0R,LC1R;
end function; //LocalImagesReals

		
		
IntersectCosets := function(V,W);
	if Type(V) eq MonStgElt or Type(W) eq MonStgElt then
		return "empty";
	end if;
	V1 := V[1];
	v1 := V[2];
	V2 := W[1];
	v2 := W[2];
	// we want to find  ( V1 + v1 ) meet ( V2 + v2)
	VW,i1,i2 := DirectSum(V1,V2);
	p1 := Inverse(i1); p2 := Inverse(i2);
	dif := hom< VW -> ASqmodk | [ -p1(VW.i) : i in [1..#Generators(V1)] ] cat [ p2(VW.i) : i in [#Generators(V1)+1..#Generators(VW)] ] >;
	if HasPreimage(v1-v2, dif) then
		alpha := v1 + p1((v1-v2) @@ dif);
		return <V1 meet V2,alpha>;
	else
		return "empty";
	end if;
end function; //IntersectCosets


SelfakeJ := Kernel(NASqmodktokq);
SelfakePic1 := <SelfakeJ,alpha_c>;

for v in S do
	LC0v,LC1v := LocalImages(v);
	SelfakeJ := SelfakeJ meet LC0v;
	currentdim := #Generators(SelfakeJ);
	vprintf Selmer, 2 : "  Bound for dimension of fake Selmer group is now %o.\n", currentdim; 
	SelfakePic1 := IntersectCosets(SelfakePic1,LC1v);
	if Type(SelfakePic1) eq MonStgElt and not Pic1kEmpty then
		vprintf Selmer, 2: "  Sel(Pic1) is empty.\n";
		Pic1kEmpty := true;
	end if;
end for;

if q eq 2 then

	vprintf Selmer, 2: "\n  Computing local image at real primes.\n";
	// code taken from Algaem/selmer.m
	for pl in [p: p in InfinitePlaces(k) | Type(p) eq Infty or IsReal(p)] do
		LC0R, LC1R := LocalImagesReals(pl);
		SelfakeJ := SelfakeJ meet LC0R;
		SelfakePic1 := IntersectCosets(SelfakePic1,LC1R);
	end for;
end if;

if Type(SelfakePic1) eq MonStgElt and MethodIsValid then
	SelfakePic1 := < SelfakeJ, {} >;
end if;

if Type(SelfakePic1) eq MonStgElt and not MethodIsValid then
	SelfakePic1 := < SelfakeJ, "?">;
end if;
//Ks,tt := AbsoluteAlgebra(A : Fields := { ext<QQ | f_i[1] > : f_i in factf } );


// Interpret data:


dimSel := NumberOfGenerators(SelfakeJ) + 1;
if Seltofakeinj then dimSel -:= 1; end if;
vprintf Selmer, 1 : "\n\n%o-Selmer group has dimension %o.\n", q eq 2 select 2 else "phi", dimSel;
if not flagphitorsglobal then
	vprintf Selmer, 1: "J(k)[%o] has dimension %o.\n", q eq 2 select 2 else "phi", ranktorsglobal;
else
	vprintf Selmer, 1: "J(k)[%o] has dimension ge %o.\n", q eq 2 select 2 else "phi", ranktorsglobal;
end if;


if q eq 2 then

	// we get limited information on the 4 torsion
	vprintf Selmer, 2: "  Image of J(k)[%o] under the descent map has dimension ge %o.\n", q eq 2 select 2 else "phi", Ngens(muJktors);
	assert q eq 2;
	if ranktorsglobal - Ngens(muJktors) gt extratwotorsion then
		// there must be 4-torsion
		vprintf Selmer, 2: "    ==> J(k)[4] is nontrivial.\n";
	else
		if extratwotorsion eq 0 and ranktorsglobal eq Ngens(muJktors) then
			vprintf Selmer, 2: "    ==> 2J(k)[4] = 0.\n";
		end if;
	end if;


	if PrimeToqGlob or (g mod 2 eq 0 and CasGlob) then
		vprintf Selmer, 1: "Sel(Pic1) = Sel(J) = Sel_fake(J).\n";
		vprintf Selmer, 1: "    ==> Rank(J(k)) le %o.\n", dimSel-ranktorsglobal;
	else

		if Type(SelfakePic1[2]) eq Type({}) then
			vprintf Selmer, 1: "Sel(Pic1) is empty.\n";
			if MethodIsValid then
				//for this we need C to be ELS (which we check along the way).
				vprintf Selmer, 1: "    ==> Sha(J/k)[%o] has dimension ge 2.\n", q eq 2 select 2 else "phi";
				vprintf Selmer, 1: "    ==> Cassels kernel is trivial.\n";
				vprintf Selmer, 1: "    ==> Rank(J(k)) le %o.\n", dimSel-ranktorsglobal-2;
			end if;	
		else
			
			//Sel(Pic1) nonempty, Cassels kernel is nontrival
			vprintf Selmer, 1: "Sel(Pic1) nonempty and Cassels kernel is possibly nontrivial.\n";
			// e.g. if Pic1(k) nonempty then it is nontrivial
			vprintf Selmer, 1: "    ==> Rank(J(k)) le %o.\n", dimSel - ranktorsglobal + 1;
		end if;
	end if;
else
	
	if PrimeToqGlob or (g mod 2 eq 0 and CasGlob) then
		vprintf Selmer, 1: "Sel(Pic1) = Sel(J) = Sel_fake(J).\n";
		vprintf Selmer, 1: "    ==> Rank J(k(zeta_%o)) le %o.\n",q, (q-1)*( dimSel-ranktorsglobal);
		vprintf Selmer, 1: "    ==> Rank J(k) le %o.\n", (q-1)*( dimSel-ranktorsglobal)/(Degree(k)/Degree(k0));

	else

		
		if Type(SelfakePic1[2]) eq Type({}) then
			vprintf Selmer, 1: "Sel(Pic1) is empty.\n";
			if MethodIsValid then
				//for this we need C to be ELS (which we check along the way).
				vprintf Selmer, 1: "    ==> Sha(J/k)[%o] has dimension ge 2.\n", q eq 2 select 2 else "phi";
				vprintf Selmer, 1: "    ==> Cassels kernel is trivial.\n";
				vprintf Selmer, 1: "    ==> Rank(J(k(zeta))) le %o.\n", (q-1)*( dimSel-ranktorsglobal-2);
				vprintf Selmer, 1: "    ==> Rank(J(k)) le %o.\n", (q-1)*( dimSel-ranktorsglobal-2)/(Degree(k)/Degree(k0));
			end if;	

		else
			//Sel(Pic1) nonempty, Cassels kernel is nontrival
			vprintf Selmer, 1: "Sel(Pic1) nonempty and Cassels kernel is possibly nontrivial.\n";
			// e.g. if Pic1(k) nonempty then it is nontrivial
			vprintf Selmer, 1: "    ==> Rank(J(k(zeta))) le %o.\n", (q-1)*( dimSel - ranktorsglobal + 1);
			vprintf Selmer, 1: "    ==> Rank(J(k)) le %o.\n", (q-1)*( dimSel - ranktorsglobal + 1)/(Degree(k)/Degree(k0));
		end if;
	
	end if;	

end if;
			

return SelfakeJ, SelfakePic1, ASqmodktoA;

//end function;
end intrinsic; //PicnDescents


// For Testing:
/*
intrinsic TestPicnDesc2(N :: RngIntElt : ClassBounds := 1000, ELS := false) -> MonStgElt
{Tests PicnDescents on a random everywhere locally solvable hyperelliptic curve with coefficients le N.
Set ClassBounds to 0 to do calculations rigourously}
	
if ClassBounds gt 0 then 
SetClassGroupBounds(ClassBounds);
end if;

RandomHyp := function(N);
while true do
coeffs := [ (-1)^Random(1)*Random(N) : i in [0..6] ];
f := Polynomial(coeffs);
if #Coefficients(f) eq 7 then
ff := Factorization(f);
if 2 in { a[2] : a in ff } then
else
C := HyperellipticCurve(f);
break;
end if;
end if;
end while;
return C;
end function;


IsEverywhereLocallySolvable := function(C);
	S := Set(BadPrimes(C));
	q := 2;
	g := Genus(C);
	while -2*g*q^(1/2)+1+q le 0 do
		S := S join {q}; // add small primes up to Weil bounds.
		q := NextPrime(q);
	end while;
	vprintf Selmer, 2: "  Checking local solubility at primes in S = %o and in Reals.\n", S;
	for p in S do
		if not IsLocallySolvable(C,p) then
			vprintf Selmer, 2: "   --> Not locally solvable at %o.\n", p;
			return false;
		end if;
	end for;
	R := PolynomialRing(Rationals());
	f := -Evaluate(DefiningPolynomial(C),[R.1,0,1]);
	if #Roots(f,RealField()) eq 0 and Evaluate(f,1) lt 0 then
		vprintf Selmer, 2: "   --> No Real points.\n";
		return false;		
	end if;
	return true;
end function;

ELS := true;
Pts := false;


C := RandomHyp(N);

	while not IsEverywhereLocallySolvable(C) and #RationalPoints(C) eq 0 do
		C := RandomHyp(N);
	end while;
	ZZZ<x> := PolynomialRing(Rationals());
	f := -Evaluate(DefiningPolynomial(C),[x,0,1]);
	
//SetVerbose("Selmer",3);
t0 := Cputime();
Pic0r,Pic1r := PicnDescentsq(f,2);

RBnew := #Generators(Pic0r) + 1;
if Type(Pic1r[2]) eq Type({}) then
RBnew -:= 2;
end if;
t1 := Cputime();
//SetVerbose("Selmer",0);
J := TwoSelmerGroup(Jacobian(C));
t2 := Cputime();
RBold := #Generators(J);
if Type(Pic1r[2]) eq MonStgElt then
//D := TwoCoverDescent(C);
D := {};
else 
D := TwoCoverDescent(C);
end if;
t3 := Cputime();
Inconsistent := (Type(Pic1r[2]) eq Type({}) and #D ne 0) or #Generators(J) ne #Generators(Pic0r)+1;

printf "\n\n********************\n\n%o.\n\nNew 2-descent: %os\n  ==> 2-Selmer group has dimension %o.\n  ==> 2-Selmer set of Pic^1 is the translate by %o.\n    ==> Rank(J) + dim J(Q)[2] le %o.\n\nOld 2-descent: %os\n  ==> 2-Selmer group has dimension %o.\n    ==> Rank(J) + dim J(Q)[2] le %o.\n\n2-descent on C: %os\n  ==> Sel(C) = %o\n\nIs everything consistent? %o.\nDid we get new information? %o.\nWas it faster? %o.\n\n********************\n\n",C, t1-t0,#Generators(Pic0r)+1,Pic1r[2],RBnew,t2-t1,NumberOfGenerators(J),RBold,t3-t2,D,not Inconsistent, RBnew lt RBold, 2*t1 lt t0+t2;
return "";

end intrinsic;


*/









