/*******************************************************************
L-Series of elliptic curves twisted by Artin representations
Allows all twists except those when both E and A have a common prime
of wild ramification above 2,3 where E has potentially good reduction.

implements
  intrinsic LSeries(E:: CrvEll, A:: ArtRep: Precision:=0)

version 2: Tim and Vladimir Dokchitser, Aug 2010
*******************************************************************/


freeze;


import "../../RepThry/ArtRep/artin.m": LocalData, LocalRepresentation, CharacterCharPoly;


Z:=Integers();
Q:=Rationals();
PR:=PolynomialRing;
procedure Swap(~A,i,j) t:=A[i]; A[i]:=A[j]; A[j]:=t; end procedure;


function EFrobEigenvalues(E,K,R)
  k:=ResidueClassField(IntegerRing(K));
  p:=Characteristic(k);
  EK:=BaseChange(E,K);
  red,model:=LocalInformation(EK);
  assert red[3] eq 0;                       // good reduction over K
  ered:=EllipticCurve([k!a: a in aInvariants(model)]);
  f:=1-(p+1-#ered)*x+p*x^2 where x is R.1;
  return [r[1]^(-1): r in Roots(f)];
end function;


function RachelLocRecip(x,elts,G,e)
  // From Rachel's paper on tame local reciprocity

  i:=Valuation(x);
  K:=Parent(x);
  pi:=UniformizingElement(K);
  k:=ResidueClassField(IntegerRing(K));
  u:=x/pi^i;
  q:=#k;
  F:=Universe(elts);
  kF:=ResidueClassField(IntegerRing(F));
  rhs:=[kF| beta^(q^i-1) / ((-1)^(e-1)*pi)^(Z!((q^i-1)*vkb)) / u^(Z!((q-1)*vkb))
    where vkb is Valuation(beta)/RamificationDegree(Parent(beta),K)
    : beta in elts];
  for j:=1 to #G do
    sigma:=G[j];
    if [kF| sigma(beta)/beta: beta in elts] eq rhs then return j; end if;
  end for;
  error "Local reciprocity failed";
end function;


function LocalFactorEA(E,A,p: prec:=0)
// Local factor of an elliptic curve E/Q by an Artin representation A at p
// Assumes E has bad reduction at p and A is ramified at p

  if prec eq 0 
    then CF:=ComplexField(); prec:=Precision(CF);
    else CF:=ComplexField(prec);
  end if;
  Precision:=prec;

  _,vdisc,vcond,tam,kod,split:=Explode(LocalInformation(E,p));
  a:=Valuation(Conductor(A),p);
  aloc:=EulerFactor(A,p: R:=CF);
  atame:=Dimension(A)-Degree(aloc);
  awild:=a-atame;

  if Valuation(jInvariant(E),p) lt 0 then
  if vcond eq 1 then                            // multiplicative
    cond:=2*Dimension(A)-Degree(aloc)  +  2*awild;
    if split
      then return cond,aloc;                             // split
      else return cond,Evaluate(aloc,-Parent(aloc).1);   // non-split
    end if;
  else                                          // potentially multiplicative
    _,c6:=Explode(cInvariants(E));              // (possibly additive at 2)
    Etw:=QuadraticTwist(E,-c6);
    K:=QuadraticField(Numerator(-c6)*Denominator(-c6));
    Atw:=A * K!![1,-1];
    return LocalFactorEA(Etw,Atw,p: prec:=prec);
  end if;
  end if;

  R<x>:=PR(CF);
                                               // potentially good
  if vcond gt 2 then
    if IsWildlyRamified(A,p) then
      error "Wild potentially good E tensored with wild A (at p=2,3) is not implemented";
    end if;

    // wild E tensored with tame A
    loc:=R!1;
    cond:=Dimension(A)*vcond;

    return cond,loc;
  end if;

  kod:=Sprint(kod);
  if p eq 2 then assert kod in ["IV","IV*"]; end if;
  if p eq 3 then assert kod in ["III","III*","I0*"]; end if;

  e:=12 div GCD(vdisc,12);                     // size of inertia

  rts,D,FPerm,ramgps,DtoG,_,DtoAut:=LocalData(A`K,p);
  F:=Universe(rts);
  chA:=LocalRepresentation(A`K,Character(A),p);   // A-> local rep
  frob:=DtoG[ClassMap(D)(FPerm)];
  I:=ramgps[1][1];
  C:=CharacterTable(D);

  IsNumZero:=func<x|Abs(x) lt 10^(-Precision div 2)>;

  cond:=2*Dimension(A) +  2*awild;
  loc:=R!1;

  // type D additive reduction, i.e. p<>2,3, minimal galois extension of Q_p
  // where E has good reduction has dihedral galois group D_2e.
  // It has cyclic inertia of size e, and a unique 2-dimensional faithful
  // representation tau. The only contributions to the local factor of
  // E tensor A come from copies in A of tau twisted by unramified characters

  if p mod e ne 1 then
    for c in C do
      inn:=InnerProduct(chA,c);
      if inn eq 0 then continue; end if;       // no invariants in most cases
      if Degree(c) ne 2 then continue; end if;
      cI:=Restriction(c,I);
      if Index(I,Kernel(cI)) ne e then continue; end if;
                                               // 2-dim inertia invariants
      cond-:=2;
      cfrobpol:=R!CharacterCharPoly(c,FPerm);
      cfrobdet:=LeadingCoefficient(cfrobpol);
      assert IsNumZero(Coefficient(cfrobpol,1));
      loc*:= 1 - p*cfrobdet*x^2;

    end for;
    return cond, loc;
  end if;
                                               // type C additive
  for c in C do
    inn:=InnerProduct(chA,c);
    if inn eq 0 then continue; end if;       // no invariants in most cases
    if Degree(c) ne 1 then continue; end if;
    cI:=Restriction(c,I);
    if Index(I,Kernel(cI)) ne e then continue; end if;

    Qp:=PrimeField(Universe(rts));
    QpR<X>:=PR(Qp);

    u:=Qp!PrimitiveRoot(p);

    K1<piK1>:=ext<Qp|X^e-p>;
    r1:=EFrobEigenvalues(E,K1,R);
    K2:=ext<Qp|X^e-u*p>;
    r2:=EFrobEigenvalues(E,K2,R);
    prim:=Exp(2*Pi(CF)*CF.1/e);

    v:=[r1[1]/r2[1]/prim,r1[1]/r2[2]/prim,r1[1]/r2[1]*prim,r1[1]/r2[2]*prim];
    err,j:=Min([Abs(x-1): x in v]);
    error if not IsNumZero(err), "Wrong Frobenius eigenvalues on E";
    zeta:=j le 2 select prim else prim^(-1);
    if j in [2,4] then Swap(~r2,1,2); end if;

    // The maximal unramified extension L of ker A in F is generated by
    // a uniformiser pomega and a root of unity pmu that generates kF

    kerAI:=Kernel(c) meet I;
    piF:=UniformizingElement(F);
    pomega:=&*[DtoAut(g)(piF): g in kerAI];
    kF:=ResidueClassField(IntegerRing(F));
    pmunew:=F!PrimitiveElement(kF);
    repeat                            // Teichmuller lift via x->x^p
      pmu:=pmunew; pmunew:=pmu^#kF;
    until pmu eq pmunew;


    DElts:=[g: g in D];
    AutF:=[DtoAut(g): g in DElts];

    // Local reciprocity of u in L
    // [used to be inverse because we have opposite
    // convention for the local reciprocity than Rachel]
    thetaFu:=DElts[RachelLocRecip(Qp!u,[pomega,pmu],AutF,e)];

    // Local reciprocity of u in K1
    Ce,CetoAut:=AutomorphismGroup(K1);
    Ceelts:=[g: g in Ce];
    AutK1:=[CetoAut(g): g in Ceelts];
    thetaK1u:=Ceelts[RachelLocRecip(Qp!u,[piK1],AutK1,e)];

    // chi is the character of Gal(K1/Qp) which agrees with A on inertia
    chars:=[ch: ch in CharacterTable(Ce) | ch(thetaK1u) eq c(thetaFu)];
    assert #chars eq 1;
    chi:=chars[1];

    // the quotient chi/a is an unramified character and we compute it
    // on the image of theta(p) to determine how the differ
    // (theta is local reciprocity)

    thetaFpi:=DElts[RachelLocRecip(Qp!p,[pomega,pmu],AutF,e)];
    thetaK1pi:=Ceelts[RachelLocRecip(Qp!p,[piK1],AutK1,e)];
    xi:=c(thetaFpi)/chi(thetaK1pi);

    //"xi=",CF!xi;

    k:=ResidueClassField(IntegerRing(Qp));
    etabar:=u^((p-1) div e);
    v,m:=Max([Valuation(g(piK1)/piK1 - etabar): g in AutK1]);
    assert v gt 0;
    zetadash:=chi(Ceelts[m]);

    // Galois acts on E as Phi -> ( r1[1]  0    )   i^(-1) -> ( zeta^(-1) 0 )
    //                            (   0   r1[2] )             (    0   zeta )
    // Galois acts on A as Phi -> xi, i^(-1) -> zetadash^(-1)

    zeta:=CF!zeta;
    xi:=CF!xi;
    zetadash:=CF!zetadash;

    // find inertia invariant subspaces, one of them when e=3,4,6
    // and two when e=2

    ok:=false;          
    if IsNumZero(zeta^(-1)*zetadash^(-1) - 1) then
      cond-:=1;
      loc*:= 1 - xi*r1[1]*x;
      //"loc",1-xi*r1[1]*x;
      ok:=true;
    end if;
    if IsNumZero(zeta*zetadash^(-1) - 1) then
      cond-:=1;
      loc*:= 1 - xi*r1[2]*x;
      //"loc",1-xi*r1[2]*x;
      ok:=true;
    end if;
    error if not ok, "Could not find the Frobenius eigenvalue";

  end for;
  return cond, loc;

end function;



intrinsic LSeries(E:: CrvEll, A:: ArtRep: Precision:=0) -> LSer
{L-series L(E,A,s) of an elliptic curve E/Q twisted by an Artin representation A}
  require BaseField(E) cmpeq Q:
    "Elliptic curve must be defined over the rationals";
  ch:=Character(A);
  if IsZero(ch) then
    L:=LSeries(1);
  elif IsIrreducible(ch) then
    commonramification:=PrimeDivisors(GCD(Conductor(E),Conductor(A)));
    LE:=LSeries(E: Precision:=Precision);
    LA:=LSeries(A: Precision:=Precision);
    L:=TensorProduct(LE,LA,[<p,cond,loc>
      where cond,loc:=LocalFactorEA(E,A,p: prec:=Precision)
      : p in commonramification]);
  else
    dec:=Decomposition(A);
    dec:=&cat([[d[1]: i in [1..d[2]]]: d in dec]);
    L:=&*[LSeries(E,a: Precision:=Precision): a in dec];
  end if;
  L`parent:=<E,"tensor",A>;
  L`name:="twist of " cat Sprint(E) cat " by " cat Sprint(A);
  return L;
end intrinsic;
