freeze;

intrinsic AutomorphismGroup(G :: GrpFP) -> GrpAuto
{The automorphism group of a free group G}
 /* The first two generators permute the generators of F and generate S_n.
  * The third inverts the first generator of F.
  * The fourth maps F.1 -> F.1*F.2 and fixes the other generators of F.
  */
  local ng, genims, A;
  require #Relations(G) eq 0 : "For type GrpFP, group must be free";
  ng := Ngens(G);
  if ng eq 1 then
    genims := [[G.1^-1]];
    igenims := [[G.1^-1]];
  else
    genims := [ [G.2,G.1] cat [G.i : i in [3..ng]] ];
    igenims := [ [G.2,G.1] cat [G.i : i in [3..ng]] ];
    if ng gt 2 then
      Append( ~genims, [G.i : i in [2..ng]] cat [G.1] );
      Append( ~igenims, [G.ng] cat [G.i : i in [1..ng-1]] );
    end if;
    Append(~genims, [G.1^-1] cat [G.i : i in [2..ng]] );
    Append(~igenims, [G.1^-1] cat [G.i : i in [2..ng]] );
    Append( ~genims, [G.1*G.2] cat [G.i : i in [2..ng]] );
    Append( ~igenims, [G.1*G.2^-1] cat [G.i : i in [2..ng]] );
  end if;
  A := AutomorphismGroup( G, [G.i : i in [1..ng]], genims, igenims );
  A`Group := G;
  A`InnerGenerators := [];
  return A;
end intrinsic;

WA1 := function(F)
// F = free group - return list of Type 1 Whitehead Automorphisms 
 n := Ngens(F);
 auts := [];
 S:=CartesianPower([-1,1],n);
 for p in Sym(n) do for s in S do
   ims := [ (F.(i^p))^s[i] : i in [1..n] ];
   Append(~auts, hom< F->F | ims >);
 end for; end for;
 return auts;
end function;

WA2 := function(F)
// F = free group - return list of Type 2 Whitehead Automorphisms 
 n := Ngens(F);
 auts := [];
 for i in [1..n] do for e in [-1,1] do
   a := F.i^e;
   allone := true;
   S := CartesianPower([1..4],n-1);
   for s in S do if not forall(x){x: x in [1..n-1] | s[x] eq 1} then
     ims := []; sct := 0;
     for j in [1..n] do
       if j eq i then
         ims[j] := F.i;
       else
         sct +:= 1;
         ims[j] := s[sct] eq 1 select F.j else s[sct] eq 2 select F.j*a
             else  s[sct] eq 3 select a^-1*F.j else a^-1*F.j*a;
       end if;
     end for;
     Append(~auts,hom< F->F | ims >);
   end if; end for;
  end for; end for;
  return auts;
end function;

intrinsic WhiteheadReduction(F :: GrpFP, l :: SeqEnum) -> BoolElt, SeqEnum, GrpAutoElt
{Perform Whitehead reduction of a sequence of elements of a free group F.
Return true/false if list was part of a free basis, the Whitehead reduction,
and corresponding automorphism of F}
  /* Whitehead reduce.
   * return true if fully reduced to part of free basis, false otherwise.
   * return also reduced list and automorphism of F mapping l to reduction.
   * If fully reduced, automorphism maps l to F.1, F.2, ...
   */
  local W, red, a, na, nb, nc, r, fr;
  require #Relations(F) eq 0 : "For type GrpFP, group must be free";
  r := Ngens(F);
  W := WA2(F);
  red := true;
  a := hom< F-> F | [F.i : i in [1..Ngens(F)]] >;
  fr := forall{ x : x in l | #x eq 1 } and
                        #{ Abs( Eltseq(e)[1] ) : e in l } eq #l;
  red := not fr;
  while red do
    red := false;
    for w in W do
      if &+[#w(x) : x in l] lt &+[#x : x in l] then
        l := [w(x) : x in l];
        a := a * w;
        fr := forall{ x : x in l | #x eq 1 } and
                        #{ Abs( Eltseq(e)[1] ) : e in l } eq #l;
        red := not fr;
        if fr then break; end if;
        //continue;
      end if;
    end for;
  end while;
  if fr then //fully reduced
    na := [ Eltseq(e)[1] : e in l ];
    nb := [ Sign(na[i]) * i : i in [1..#l] ];
    na := [ Abs(i) : i in na ];
    for i in [1..r] do if Position(na, i) eq 0 then
      Append(~na, i);
    end if; end for;
    nb cat:= [#l+1 .. r];
    nc := [ nb[ Position(na,i) ] : i in [1..r] ];
    a := a * hom< F-> F | [ F.nc[i] : i in [1..r] ] >;
    return true, [ F.i : i in [1..#l] ],
     AutomorphismGroup(F, [F.i : i in [1..r]], [[a(F.i) : i in [1..r]]]).1;
  else
    return false, l,
     AutomorphismGroup(F, [F.i : i in [1..r]], [[a(F.i) : i in [1..r]]]).1;
  end if;
end intrinsic;

WB := function(F,l)
  isit, rl := WhiteheadReduction(F,l);
  if &+[#x : x in rl] ne &+[#x : x in l] then
    error "WB: input not Whitehead reduced";
  end if;
  W1 := WA1(F);
  W2 := WA2(F);
  ball := {l};
  repeat
    ov := #ball;
    for b in ball do
      for w in W2 do
        if &+[#w(x) : x in b] eq &+[#x : x in b] then
          Include(~ball, [w(x) : x in b] );
        end if;
      end for;
    end for;
    nv := #ball;
    "Ball size:",nv;
  until nv eq ov;

  //close under W1
  for b in ball do
    for w in W1 do if &+[#w(x) : x in b] eq &+[#x : x in b] then
      Include(~ball, [w(x) : x in b] );
    end if; end for;
  end for;
  nv := #ball;
  "Final ball size:",nv;
  return ball;
end function;
