/******************************************************************************
 *
 *    classical.m    Classical groups package
 *
 *    File      : $HeadURL:: https://subversion.sfac.auckland.ac.nz/svn/prj_#$:
 *    Author    : Eamonn O'Brien, Elliot Costi, Henrik Brnhielm
 *    Dev start : 2008-02-18
 *
 *    Version   : $Revision:: 2330                                           $:
 *    Date      : $Date:: 2012-08-08 09:19:02 +1200 (Wed, 08 Aug 2012)       $:
 *    Last edit : $Author:: jbaa004                                          $:
 *
 *    $Id:: classical.m 2330 2012-08-07 21:19:02Z jbaa004                  $:
 *
 *****************************************************************************/

freeze;
 
/*****************************************************************************/
/* DECLARATIONS                                                              */
/*****************************************************************************/

declare verbose ClassicalRecognition, 10;

declare attributes GrpMat : ClassicalRecognitionData;

ClassicalRecognitionInfo := recformat<
    type : Tup, // Classical type
    gens : SeqEnum, // Standard generators
    slps : SeqEnum, // SLPs of standard generators
    wordGroup : GrpSLP, // SLP group for returned SLPs
    cbmInput : GrpMatElt, // From Eamonn to input
    cbmMagma : GrpMatElt, // From input to Magma
    cbmStd : GrpMatElt, // From Magma to Eamonn
    form : AlgMatElt
    >;

import "rewriting/sp-natural.m" : SymplecticCBM;
import "rewriting/orth_plus.m" : OmegaCBM;
import "rewriting/su-natural.m" : UnitaryCBM;
import "rewriting/su-natural_odd.m" : UnitaryOddCBM;
import "recognition/orthogonal/odd/dp.m" : MyOrthogonalForm;
import "recognition/standard.m" : SLChosenElements, SpChosenElements,
    SUChosenElements, SOChosenElements,
    MinusChosenElements, PlusChosenElements;
import "recognition/basics.m" : RecognitionError;

import "../reduce/parameters.m" : SLRybaParameters, SpRybaParameters,
    SURybaParameters, OmegaMinusRybaParameters, OmegaPlusRybaParameters,
    OmegaRybaParameters;

import "../reduce/reduce.m" : Reduction;
import "../reduce/parameters.m" : MaxRybaEvenFieldDegree;
import "../recog/magma/node.m" : ERROR_RECOGNITION, ERROR_MEMBERSHIP, ERROR_RYBA;
import "../GrpMat/util/forms.m" : NormaliseQuadForm;
import "../GrpMat/util/basics.m" : MatSLP;

forward ClassicalConstructiveMembershipEngine, TransformOrthogonalForm,
    Omega2Recognition, ClassicalMembership;

MembershipTests := AssociativeArray();
MembershipTests[<"linear", true>] := func<g, form | IsOne(Determinant(g))>;
MembershipTests[<"symplectic", true>] :=
    func<g, form | g * form * Transpose(g) eq form>;
MembershipTests[<"unitary", true>] :=
    func<g, form | g * form * Transpose(FrobeniusImage(g, Degree(F) div 2)) eq
    form where F is CoefficientRing(g)>;

MembershipTests[<"orthogonalcircle", false>] := function(g, form)
    if Characteristic(CoefficientRing(g)) eq 2 then
        return NormaliseQuadForm(g * form * Transpose(g)) eq form;
    else
	return g * form * Transpose(g) eq form;
    end if;
end function;

MembershipTests[<"orthogonalcircle", true>] :=
    func<g, form | MembershipTests[<"orthogonalcircle", false>](g, form)
    and IsZero(SpinorNorm(g, form))>;

MembershipTests[<"orthogonalminus", false>] :=
    MembershipTests[<"orthogonalcircle", false>];
MembershipTests[<"orthogonalplus", false>] :=
    MembershipTests[<"orthogonalcircle", false>];
MembershipTests[<"orthogonalplus", true>] :=
    MembershipTests[<"orthogonalcircle", true>];
MembershipTests[<"orthogonalminus", true>] :=
    MembershipTests[<"orthogonalcircle", true>];

/*****************************************************************************/
/* TEST AND BENCHMARK FUNCTIONS                                              */
/*****************************************************************************/

/* 
intrinsic testClassicalRecognition(d :: RngIntElt, p :: RngIntElt,
    e :: RngIntElt) -> BoolElt
    { }

    return testClassicalRecognition(RandomConjugate(SU(d, GF(p, 2 * e))));
end intrinsic;

intrinsic testClassicalRecognition(G : Case := "unknown") -> BoolElt
    { }

    SetVerbose("ClassicalRecognition", 7);
    SetVerbose("User1", 0);
    SetVerbose("User2", 0);
    SetVerbose("User3", 0);
    SetVerbose("User4", 0);
    SetVerbose("User5", 0);
    SetVerbose("CompositionTree", 6);

    try
        flag := ClassicalConstructiveRecognition(G : Case := Case);
        assert flag;

	
	delete G`UserGenerators;
	assert Evaluate(G`ClassicalRecognitionData`slps,
	    UserGenerators(G)) eq [g^G`ClassicalRecognitionData`cbmInput :
	    g in G`ClassicalRecognitionData`gens];
	
	for i in [1 .. 10] do
	    g := Random(G`RandomProcess);
	    
	    flag, slp := ClassicalElementToWord(G, g);
	    assert flag;
	    
	    assert Evaluate(slp, G`ClassicalRecognitionData`gens) eq
		g^(G`ClassicalRecognitionData`cbmInput^(-1));
            
	    
	    slpUser := Evaluate(slp, G`ClassicalRecognitionData`slps);
	    assert Evaluate(slpUser, UserGenerators(G)) eq g;
	end for;
    catch e
	print e`Object;
        print e`Position;
    end try;
    
    return true;
end intrinsic;
*/

/*****************************************************************************/
/* MAIN INTRINSICS                                                           */
/*****************************************************************************/

function SmallFieldRecognition(G, gens, RybaParameters, order)
    d := Degree(G);
    F := CoefficientRing(G);
    q := #F;
    p := Characteristic(F);
    e := Degree(F);
    
    try
	useRyba, params := RybaParameters(G, d, q);

        if IsLinearGroup(G) then
	    cbm := Identity(G);
	elif IsOrthogonalGroup(G) and d eq 2 then
	    H := sub<Generic(G) | gens>;
	    cbm := Generic(G) ! (TransformOrthogonalForm(H) *
		TransformOrthogonalForm(G)^(-1));
	else
	    H := sub<Generic(G) | gens>;
	    cbm := Generic(G) ! (TransformForm(H) * TransformForm(G)^(-1));
	end if;
	
        if not useRyba then
	    vprint ClassicalRecognition, 4 : "RandomSchreier case";
	    if not HasAttribute(G, "FactoredOrder") and d ne 2 then
		AssertAttribute(G, "FactoredOrder", order);
	    end if;

	    RandomSchreier(G);

            // if G has large number of generators, then
            // Verify can be expensive even if |G| is small 
            // choose small # of random elements to generate subgroup S of G
            // when all generators of G are in S, then Verify (S), not G
            n := 5;
            if params[1] and #UserGenerators (G) gt 2 * n then
               vprint CompositionTree, 5: "Replace G by subgroup on fewer generators";
	       gensS := [];
               slpsS := [];
               S := sub<G | >;
               repeat
	          for i in [1 .. n] do
		    g, w := Random(G`RandomProcess);
		    if not IsIdentity (g) and not (g in gensS) then
                       Append (~gensS, g);
		       Append (~slpsS, w);
                    end if;
		  end for;
		
		  S := sub<Generic(G) | gensS>;
                  assert Ngens (S) eq #gensS;
	          RandomSchreier(S);

		  vprint ClassicalRecognition, 5 : "Verify BSGS for S";
		  Verify(S);

	          vprint ClassicalRecognition, 5 :
		    "Obtain SLPs of standard generators";
	       until forall{g : g in UserGenerators(G) | g in S};
	    
	       phi := InverseWordMap(S);
               slps := [Evaluate(phi(g^cbm), slpsS) : g in gens];

               // check that result is correct 
               // Y := Evaluate (slps, UserGenerators (G));
               // assert forall{i : i in [1..#Y] | [Y[i] eq gens[i]^cbm};

            else 
               if params[1] then
	          vprint ClassicalRecognition, 5 : "Verify BSGS for G";
		  Verify(G);
	       end if;
	    
               vprint ClassicalRecognition, 5 :
	         "Obtain SLPs of standard generators";

               phi := InverseWordMap(G);
               slps := [phi(g^cbm) : g in gens];
            end if;
	else
	    // Only small fields
	    if p eq 2 and (IsUnitaryGroup(G) select e div 2 else e) gt
		MaxRybaEvenFieldDegree then
		error ERROR_RECOGNITION;
	    end if;

	    // Bounded dimension for SL(d, 2): 
	    // construction of centraliser of involution often fails 
	    MaxRybaLinearDegree := 20;
	    if q eq 2 and d gt MaxRybaLinearDegree then 
		error ERROR_RECOGNITION;
	    end if;
	    
	    vprint ClassicalRecognition, 4 : "Ryba case";
            try 
               slps := [w where _, w is Reduction(G, g^cbm : Verify := params[1],
		CentraliserMethod := params[2], Central := params[4],
		LieRank := params[5], CentraliserCompletionCheck :=
		func<G, C, g, l | NumberOfGenerators(C) ge params[3]>) :
		g in gens];
            catch err
               error ERROR_RYBA;
            end try;
	end if;

	vprint ClassicalRecognition, 3 :
	    "Small field standard generators obtained";
	return gens, slps, cbm;
    catch err
	error ERROR_RECOGNITION;
    end try;
end function;

function SLSmallEvenRecognition(G)
    d := Degree(G);
    F := CoefficientRing(G);
    q := #F;

    try
	gens := SLChosenElements(G);
    
        if d eq 2 then
	    flag, _, _, g2slp := RecogniseSL2(G, q);
	    assert flag;
	    
	    slps := [Function(g2slp)(g) : g in gens];
	elif d eq 3 then
	    flag, _, _, g2slp := RecogniseSL3(G, F : Verify := false);
	    assert flag;
	    
	    slps := [Function(g2slp)(g) : g in gens];	    
	else
	    gens, slps := SmallFieldRecognition(G, gens,
		SLRybaParameters, FactoredClassicalGroupOrder(ClassicalType(G), d, F));
	end if;

	return gens, slps, Identity(G);
    catch err
	error ERROR_RECOGNITION;
    end try;
end function;
    
function SLConstructiveRecognition(G, type)
    F := CoefficientRing(G);
    q := #F;
    p := Characteristic(F);
    d := Degree(G);

    if p eq 2 then
        if (q eq 2 and (IsOdd (d) or d gt 8)) or (q gt 2) then 
	    gens, slps, cbm := int_SolveSL(G);
	else
	    gens, slps, cbm := SLSmallEvenRecognition(G);
        end if;
    else
	gens, slps, cbm := int_SolveSL(G);
    end if;

    if assigned G`SLPGroup then
	W := G`SLPGroup;
    else
	W := WordGroup(G);
    end if;
    
    return rec<ClassicalRecognitionInfo |
	type := <type, true>,
	cbmInput := Generic(G) ! cbm,
	cbmStd := Identity(G),
	gens := gens,
	slps := slps,
	wordGroup := W,
	form := IdentityMatrix(F, d), 
	cbmMagma := Identity(G)>;
end function;

function SUSmallEvenRecognition(G)
    d := Degree(G);
    F := CoefficientRing(G);
    assert IsEven(Degree(F));
    Fdef := sub<F | Degree(F) div 2>;
    q := #Fdef;
        
    try
	gens := SUChosenElements(G);

        H := sub<Generic(G) | gens>;
	gens, slps, cbm := SmallFieldRecognition(G, gens,
	    SURybaParameters, FactoredClassicalGroupOrder(ClassicalType(G), d, F));

	return gens, slps, cbm;
    catch err
	error ERROR_RECOGNITION;
    end try;
end function;

function SUConstructiveRecognition(G, type)
    flag, form := UnitaryForm(G);
    error if not flag, ERROR_RECOGNITION;

    F := CoefficientRing(G);
    d := Degree(G);
    q := #F;
    q0 := Isqrt (q);
    p := Characteristic (F);

    if p eq 2 then
        if  (q0 eq 2 and (d in {3,4,5,6,7,9})) or 
            (q0 eq 4 and (d in {4})) then 
	    gens, slps, cbm := SUSmallEvenRecognition(G);
	else
	    gens, slps, cbm := int_SolveSU(G);
        end if;
    else
	gens, slps, cbm := int_SolveSU(G);
    end if;

    if assigned G`SLPGroup then
	W := G`SLPGroup;
    else
	W := WordGroup(G);
    end if;
    
    return rec<ClassicalRecognitionInfo |
	type := <type, true>,
	cbmInput := Generic(G) ! cbm,
	cbmStd := IsEven(d) select UnitaryCBM(G) else UnitaryOddCBM(G),
	gens := gens,
	slps := slps,
	wordGroup := W,
	form := form,
	cbmMagma := Generic(G) ! TransformForm(G)>;	
end function;

function SpConstructiveRecognition(G, type)
    flag, form := SymplecticForm(G);
    assert flag;

    F := CoefficientRing(G);
    d := Degree(G);
    p := Characteristic (F);
    q := #F;
    
    if p eq 2 then 
        if (q eq 2 and d in {6..12 by 2}) then 
	   vprint ClassicalRecognition, 3 : "Small field standard generators";
	   gens, slps, cbm := SmallFieldRecognition(G, SpChosenElements(G),
	       SpRybaParameters, FactoredClassicalGroupOrder(type, d, F));
        else 
	   gens, slps, cbm := int_SolveSp(G);
        end if;
    else	
	gens, slps, cbm := int_SolveSp(G);
    end if;

    if assigned G`SLPGroup then
	W := G`SLPGroup;
    else
	W := WordGroup(G);
    end if;
    
    vprint ClassicalRecognition, 2 : "Sp standard generators found";
    
    return rec<ClassicalRecognitionInfo |
	type := <type, true>,
	cbmInput := Generic(G) ! cbm,
	cbmStd := SymplecticCBM(G),
	gens := gens,
	slps := slps,
	wordGroup := W,
	form := form,
	cbmMagma := Generic(G) ! TransformForm(G)>;
end function;

function OmegaCircleConstructiveRecognition(G, type)
    F := CoefficientRing(G);

    // Omega_0 in char 2 is Sp
    assert Characteristic(F) gt 2;
    
    flag, _, form := OrthogonalForm(G);
    error if not flag, ERROR_RECOGNITION;
    
    isOmega := forall{g : g in UserGenerators(G) |
	IsZero(SpinorNorm(g, form))};
    cbmMagma := TransformForm(G);
    
    if isOmega then
	gens, slps, cbm := int_SolveO(G);
    else
	gens, slps, cbm := int_SolveSO(G);
    end if;

    if assigned G`SLPGroup then
	W := G`SLPGroup;
    else
	W := WordGroup(G);
    end if;
    
    return rec<ClassicalRecognitionInfo |
	type := <type, isOmega>,
	cbmInput := Generic(G) ! cbm,
	cbmStd := Identity(G),
	gens := gens,
	slps := slps,
	wordGroup := W,
	form := form,
	cbmMagma := Generic(G) ! cbmMagma>; 
end function;

function OmegaMinusConstructiveRecognition(G, type)
    F := CoefficientRing(G);
    d := Degree(G);

    if d gt 2 then
	if Characteristic (F) eq 2 then
	    flag, form := QuadraticForm(G);
	else
	    flag, form := SymmetricBilinearForm (G);
	end if;

	error if not flag, ERROR_RECOGNITION;
	cbmMagma := TransformForm(G);
    else
	H, phi := ExtendField(Generic(G), ext<F | 2>);
	cbmMagma, form := TransformOrthogonalForm(G);
    end if;
    
    isOmega := forall{g : g in UserGenerators(G) |
	IsZero(SpinorNorm(g, form))};

    p := Characteristic (F);
    q := #F;
    if d gt 2 then
	if p eq 2 then 
	    vprint ClassicalRecognition, 3 : "Small field standard generators";
	    
	    cbmStd := Identity(G);
	    order := FactoredClassicalGroupOrder(type, d, F);
	    
	    // In char 2, SO = Omega, except in one case
	    if not isOmega and (d eq 2 or (d eq 4 and #F eq 2)) then
		order *:= 2;
	    end if;

            if (d in {8,10,12,14} and q eq 2) or (d in {8,10} and q eq 4) then
         	gens, slps, cbm := SmallFieldRecognition(G, 
                   MinusChosenElements(G), OmegaMinusRybaParameters, order);
            else 
		gens, slps, cbm := int_SolveOMinus(G);
	        cbmStd := Identity(G);
                cbm := Generic(G) ! cbm;
            end if;
	else
	    if isOmega then
		gens, slps, cbm := int_SolveOMinus(G);
	    else
		gens, slps, cbm := int_SolveSOMinus(G);
	    end if;
	    
	    cbmStd := Identity(G);
	    cbm := Generic(G) ! cbm;
	end if;
    else
	gens, slps, cbm := Omega2Recognition(G, type);
	cbmStd := Identity(H);
    end if;
    
    if assigned G`SLPGroup then
	W := G`SLPGroup;
    else
	W := WordGroup(G);
    end if;
    
    return rec<ClassicalRecognitionInfo |
	type := <type, isOmega>,
	cbmInput := cbm,
	cbmStd := cbmStd,
	gens := gens,
	slps := slps,
	wordGroup := W,
	form := form,
	cbmMagma := Generic(G) ! cbmMagma>;
end function;

function OmegaPlusConstructiveRecognition(G, type)
    F := CoefficientRing(G);
    d := Degree(G);

    if d gt 2 then
        if Characteristic (F) eq 2 then
           flag, form := QuadraticForm(G);
        else
           flag, form := SymmetricBilinearForm (G);
        end if;

	error if not flag, ERROR_RECOGNITION;
	cbmMagma := TransformForm(G);
    else
	cbmMagma, form := TransformOrthogonalForm(G);
    end if;
    
    isOmega := forall{g : g in UserGenerators(G) |
	IsZero(SpinorNorm(g, form))};

    q := #F;
    p := Characteristic (F);
    
    if d gt 2 then
	if p eq 2 then
	    
	    cbmStd := OmegaCBM(G);
	    order := FactoredClassicalGroupOrder(type, d, F);
	    
	    // In char 2, SO = Omega, except in one case
	    if not isOmega and (d eq 2 or (d eq 4 and #F eq 2)) then
		order *:= 2;
	    end if;

            if (d in {4,8,10,12,14} and q eq 2) or (d eq 4 and q eq 4) then
	       vprint ClassicalRecognition, 3 : "Small field standard generators";
	       gens, slps, cbm := SmallFieldRecognition(G, PlusChosenElements(G),
		OmegaPlusRybaParameters, order);
            else 
		gens, slps, cbm := int_SolveOPlus(G);
	        cbmStd := OmegaCBM(G);
        	cbm := Generic(G) ! cbm;
            end if;
	else
	    if isOmega then
		gens, slps, cbm := int_SolveOPlus(G);
	    else
		gens, slps, cbm := int_SolveSOPlus(G);
	    end if;
	    
	    cbmStd := OmegaCBM(G);
	    cbm := Generic(G) ! cbm;
	end if;
    else
	gens, slps, cbm := Omega2Recognition(G, type);
	cbmStd := Identity(G);
    end if;
    
    if assigned G`SLPGroup then
	W := G`SLPGroup;
    else
	W := WordGroup(G);
    end if;
    
    return rec<ClassicalRecognitionInfo |
	type := <type, isOmega>,
	cbmInput := Generic(G) ! cbm,
	cbmStd := cbmStd,
	gens := gens,
	slps := slps,
	wordGroup := W,
	form := form,
	cbmMagma := Generic(G) ! cbmMagma>;    
end function;

// Hash with individual recognition functions
RecognitionFunctions := AssociativeArray();
RecognitionFunctions["linear"]           := SLConstructiveRecognition;
RecognitionFunctions["symplectic"]       := SpConstructiveRecognition;
RecognitionFunctions["unitary"]          := SUConstructiveRecognition;
RecognitionFunctions["orthogonalminus"]  := OmegaMinusConstructiveRecognition;
RecognitionFunctions["orthogonalplus"]   := OmegaPlusConstructiveRecognition;
RecognitionFunctions["orthogonalcircle"] := OmegaCircleConstructiveRecognition;

intrinsic ClassicalConstructiveRecognition(G :: GrpMat[FldFin] : Randomiser :=
    RandomProcessWithWords(G : WordGroup := WordGroup(G), Scramble := 1),
    Case := "unknown") -> BoolElt, [], []
    { Determine if G is a quasisimple classical group in natural representation.
      If so, construct standard generators S for G; return true, S
      and SLPs for S in defining generators of G. Otherwise return false.
      Case is result of ClassicalType (G) }

    F := CoefficientRing(G);
    q := #F;
    p := Characteristic(F);
    d := Degree(G);
    
    if d lt 2 then
	return false, _, _;
    end if;

    try
	if Case eq "unknown" then
	    flag := RecogniseClassical(G);
	    if not flag then
		return false, _, _;
	    end if;
	
	    Case := ClassicalType(G);
	end if;

	if not assigned G`RandomProcess then
	    G`RandomProcess := Randomiser;
	end if;

	// Hack for SL(2, 2) = SOMinus(2, 2)
	if d eq 2 and p eq 2 and Case eq "orthogonalminus" and
	    not IsCyclic (G) then
	    Case := "linear";
	end if;
		
	vprint ClassicalRecognition, 1 : "Classical type", Case;
	flag, recogniser := IsDefined(RecognitionFunctions, Case);
	if not flag then
	    error Error(Sprintf("Unknown classical type: %o\n", Case));
	end if;

	G`ClassicalRecognitionData := recogniser(G, Case);
	vprint ClassicalRecognition, 1 : "Standard generators found";

        // additions by EOB to provide useful output for intrinsic 
        S :=  [g^G`ClassicalRecognitionData`cbmInput :
                   g in G`ClassicalRecognitionData`gens];
        slps := G`ClassicalRecognitionData`slps;
        WG := WordGroup (Randomiser);
        return true, S, Evaluate (slps, [WG.i: i in [1..Ngens (WG)]]);
    catch err
	if err`Object cmpeq ERROR_RECOGNITION then
	    return false, _, _;
	end if;
	if err`Object cmpeq ERROR_MEMBERSHIP then
	    return false, _, _;
	end if;
	
	if err`Type eq "ErrUser" and Category(err`Object) eq Rec then
	    if Format(err`Object) cmpeq RecognitionError then
		// Die gracefully
		return false, _, _;
	    end if;
	end if;

	error Error(err);
    end try;
end intrinsic;

intrinsic ClassicalElementToWord(G :: GrpMat[FldFin],
    g :: GrpMatElt[FldFin]) -> BoolElt, GrpSLPElt
    { G must be a constructively recognised classical group in 
    natural representation.
    If g in an element of G, then express g as an SLP in the 
    standard generators of G and return true and the SLP; otherwise return false. }
    
    if not assigned G`ClassicalRecognitionData then
	// error Error("G must be a recognised classical group");
        "Must first run CompositionTree on G";
        T := CompositionTree (G);
    end if;
    
    if not ClassicalMembership(G, g, G`ClassicalRecognitionData`form,
	G`ClassicalRecognitionData`type) then
	return false, _;
    end if;    
    
    W := G`ClassicalRecognitionData`wordGroup;

    // Change of basis to put g in Magma copy in correct basis
    cbm := G`ClassicalRecognitionData`cbmInput^(-1);

    vprint ClassicalRecognition, 8 : "Obtain SLP of element";

    // Obtain g as SLP in generators of G (modulo Eamonn's identity)
    flag, slp := ClassicalConstructiveMembershipEngine(G, g,
	G`ClassicalRecognitionData`type, cbm,
	G`ClassicalRecognitionData`cbmInput^(-1),
	G`ClassicalRecognitionData`gens,
	G`ClassicalRecognitionData`slps);

    vprint ClassicalRecognition, 8 : "SLP obtained";
    if flag then
	assert NumberOfGenerators(Parent(slp)) eq
	    #(G`ClassicalRecognitionData`gens);
	return true, slp;
    else
	return false, _;
    end if;
end intrinsic;

/*****************************************************************************/
/* AUXILIARY FUNCTIONS                                                       */
/*****************************************************************************/


// Find change of basis that puts group in block diagonal form
// assuming such a basis exists
function blockDiagonalise(G : OrderFactors := func<l | l>)

    V := VectorSpace(G);
    M := GModule(G);
    factors := OrderFactors(CompositionFactors(M));
    homos := [GHom(f, M) : f in factors];
    assert forall{h : h in homos | Dimension(h) eq 1};
        
    basis := &cat[[V ! (x * Morphism(X, M)) :
	x in Basis(X)] where X is Image(homo.1) : homo in homos];
    cbm := Matrix(CoefficientRing(G), Degree(G), Degree(G), basis);
    return Generic(G) ! cbm^(-1);
end function;

// Constructively recognise Omega(2,q) (+ or -)
function Omega2Recognition(G, type)
    assert NumberOfGenerators(G) le 1;

    F := CoefficientRing(G);

    W := WordGroup(G`RandomProcess);
    if type eq "orthogonalplus" and #F le 5 then
	if #G eq 2 then
	    return [G.1], [W.1], Identity(G);
	end if;
    end if;
    
    // With O- we need to extend the field to make it diagonalisable
    if type eq "orthogonalminus" then
	G_ext, inc := ExtendField(Generic(G), ext<F | 2>);
	G := sub<G_ext | inc(G.1)>;

	if #F eq 3 then
	    if #G eq 2 then
		return [G.1], [W.1], Identity(G);
	    end if;
	end if;
    end if;

    // Change-of-basis necessary for membership testing
    cbm := blockDiagonalise(G);
    
    return [G.1^cbm], [W.1], cbm^(-1);
end function;

// Constructive membership testig in Omega(2, q) (+ or -)
function Omega2Rewriting(G, g, type, cbm, gens, slps)
    F := CoefficientRing(G);
    G_ext, inc := ExtendField(Generic(G), ext<F | 2>);

    // With O- we need to extend the field to make it diagonalisable
    if type[1] eq "orthogonalminus" then
	G := sub<G_ext | inc(G.1)>;
	g := inc(g);
    end if;
    
    // Special cases for order 2
    if type[1] eq "orthogonalplus" and #F le 5 or
	type[1] eq "orthogonalminus" and #F eq 3 then
	if #G eq 2 then
	    if g eq gens[1] then
		return true, slps[1];
	    elif IsIdentity(g) then
		return true, slps[1]^0;
	    else
		return false, _;
	    end if;
	end if;
    end if;
    
    // Membership testing in cyclic groups is discrete log
    if IsDiagonal(g^cbm) then
	k := Log(gens[1][1, 1], (g^cbm)[1, 1]);

	if k ge 0 then
	    return true, slps[1]^k;
	end if;
    end if;

    return false, _;
end function;

// Transform orthogonal forms to standard forms, including dimension 2
function TransformOrthogonalForm(G)
    flag, type, form := MyOrthogonalForm(G);
    assert flag;
    
    return TransformForm(form, type), form,
	[CoefficientRing(G) | 1 : i in [1 .. NumberOfGenerators(G)]];
end function;

function ClassicalMembership(G, g, form, type)
    F := CoefficientRing(G);
    return MembershipTests[type](g, form);
end function;

TypeConversion := AssociativeArray();
TypeConversion["linear"] := "SL";
TypeConversion["symplectic"] := "Sp";
TypeConversion["unitary"] := "SU";
TypeConversion["orthogonalminus"] := "Omega-";
TypeConversion["orthogonalplus"] := "Omega+";
TypeConversion["orthogonalcircle"] := "Omega";

function ClassicalConstructiveMembershipEngine(G, g, type,
    cbm, cbmTest, gens, slps)

    d := Degree(G);

    if d eq 2 and type[1] in {"orthogonalplus", "orthogonalminus"} then
        return Omega2Rewriting(G, g, type, cbm, gens, slps);
    end if;
    
    if type[1] eq "unitary" then
	F := CoefficientRing(G);	
	Fdef := sub<F | Degree(F) div 2>;
	q := #Fdef;
    else
	F := CoefficientRing(G);
	q := #F;
    end if;
	
    typeStr := TypeConversion[type[1]];
    flag, slp := ClassicalRewrite(G^cbm, gens, typeStr, d, q, g^cbm);

    if not flag then
	return false, _, _;
    else
	assert Evaluate(slp, gens) eq g^cbmTest;
	// return true, Evaluate(slp, slps);
	return true, slp;
    end if;
end function;

