/******************************************************************************
 *
 *    classical-bb.m  Composition Tree Classical groups black box Leaf Code
 *
 *    File      : $HeadURL:: https://subversion.sfac.auckland.ac.nz/svn/prj_#$:
 *    Author    : Peter Brooksbank and Henrik Brnhielm
 *    Dev start : 2008-10-07
 *
 *    Version   : $Revision:: 2377                                           $:
 *    Date      : $Date:: 2012-10-28 07:40:29 +1300 (Sun, 28 Oct 2012)       $:
 *    Last edit : $Author:: jbaa004                                          $:
 *
 *    $Id:: classical-bb.m 2377 2012-10-27 18:40:29Z jbaa004               $:
 *
 *****************************************************************************/

freeze;
 
/*****************************************************************************/
/* DECLARATIONS                                                              */
/*****************************************************************************/

import "centre.m" : C9Centre;
import "leaf.m" : MembershipTesting, LeafError,
    C9CenterMembership, C9Membership, C9Presentation, BuiltinMembership,
    LeafFactorData, MaxBSGSVerifyPermDegree;
import "recog.m" : LeafInfo, SLPMapsInfo, CTNodeInfo, NiceInfo,
    PresentationInfo, FactorInfo;
import "node.m" : ERROR_RECOGNITION, ERROR_RECOGNITION_CATCH;
import "comp-series.m" : PullbackCompFactors;
import "classical.m" : NonSimpleClassicalNames;
import "c9.m" : PerfectCheckElements;

import "../../reduce/parameters.m" :
    SLRybaParameters, SpRybaParameters, SURybaParameters;
import "../../reduce/reduce.m" : Reduction;

import "../../classical/recognition/standard.m" :
    SpChosenElements, SLChosenElements, SUChosenElements;
import "../../GrpMat/util/basics.m" : getMapFunction;

import "../../GrpMat/modules/construct.m" :
    DeltaSquare, MyAdjointRepresentation;

forward SpRecognition, SLRecognition, SURecognition, RecogniseSLAdjoint,
    SLSLPMaps, SpSLPMaps, SUSLPMaps,
    SLShortPresentation, SpShortPresentation, SUShortPresentation,
    GeneralShortPresentation, SLNiceData, SpNiceData, SUNiceData,
    ClassicalStandardGenerators, RecogniseSpOddWrapper,
    RecogniseSLDelta, RecogniseSLAlternatingSquare, RecogniseSLSymmetricSquare,
    SL2RecognitionBlackBox, SL3RecognitionGreyBox, RecogniseSLWrapper,
    SU3RecognitionBlackBoxWrapper, SU4RecognitionBlackBoxWrapper;

StandardCopyInfo := recformat<
    Creator : Intrinsic,
    Recogniser : UserProgram,
    RankToDim : UserProgram,
    SLPMaps : UserProgram,
    NiceData : UserProgram,
    Relators : UserProgram,
    StandardGenerators : UserProgram>;

StandardCopyData := AssociativeArray();

StandardCopyData["C"]  := rec<StandardCopyInfo |
    Creator := Sp,
    SLPMaps := SpSLPMaps,
    Recogniser := SpRecognition,
    Relators := SpShortPresentation,
    NiceData := SpNiceData,
    RankToDim := func<x | 2 * x>,
    StandardGenerators := SpChosenElements>;

StandardCopyData["A"]  := rec<StandardCopyInfo |
    Creator := SL,
    Relators := SLShortPresentation,
    SLPMaps := SLSLPMaps,
    NiceData := SLNiceData,
    Recogniser := SLRecognition,
    RankToDim := func<x | x + 1>,
    StandardGenerators := SLChosenElements>;

StandardCopyData["2A"]  := rec<StandardCopyInfo |
    Creator := SU,
    Relators := SUShortPresentation,
    SLPMaps := SUSLPMaps,
    NiceData := SUNiceData,
    Recogniser := SURecognition,
    RankToDim := func<x | x + 1>,
    StandardGenerators := SUChosenElements>;

// The black box recognisers often fails, retry a number of times
RecognitionTries := 30;

declare attributes GrpMat : SURecognitionData;

SURecognitionInfo := recformat<
    iso : Map,
    inv : Map,
    g2slp : Map,
    slp2g : Map>;

/*****************************************************************************/
/* TEST FUNCTIONS                                                            */
/*****************************************************************************/

/*****************************************************************************/
/* MAIN INTRINSICS                                                           */
/*****************************************************************************/

function IsClassicalBB(node)
    case Retrieve(node`LeafData`Family):
    when "A":
	return true;
    when "C":
	return IsOdd(Retrieve(node`LeafData`DefiningField));
    when "2A":
	return node`LeafData`LieRank in {2, 3};
    else
	return false;
    end case;	
end function;

/*
 ExtendedGroup = extension of classical group, extra centre or scalar flag
 S = gold copy
 Recogniser = recognition function for classical group
 Name = Lie type

 Moves to derived group to kill centre and then executes Recogniser.
 Returns inverse isomorphisms.

 Recogniser should be of the form
 flag, iso, inv := Recogniser(G, d, q)
*/
function GeneralRecognition(ExtendedGroup, S, Recogniser, Name)
    d := Degree(S);
    F := CoefficientRing(S);
    
    vprint CompositionTree, 3 : "Decide if group is perfect";

    if not IsProbablyPerfect(ExtendedGroup :
	NumberRandom := PerfectCheckElements(ExtendedGroup)) then

        vprint CompositionTree, 3 : "Compute derived group";
	repeat
	    D, D_slps := DerivedGroupMonteCarlo(ExtendedGroup);
	    flag, name := LieType(D, Characteristic(F));
	until flag and name eq Name;
    else
	// Possibly remove duplicate generators in GrpPerm context
	if Category(ExtendedGroup) eq GrpPerm then
	    D := sub<Generic(ExtendedGroup) | {@ g : g in
		UserGenerators(ExtendedGroup) |
		g ne Identity(ExtendedGroup) @}>;
	else
	    D := ExtendedGroup;
	end if;
	
	gens := ChangeUniverse(UserGenerators(ExtendedGroup),
	    Generic(ExtendedGroup));
	    
	genMap := [Index(gens, g) : g in UserGenerators(D)];
	D_slps := UserGenerators(WordGroup(ExtendedGroup))[genMap];
    end if;
    
    vprint CompositionTree, 3 : "Call black box recogniser";
    flag, isoF, invF := Recogniser(D, d, #F);
    
    if flag then
	// Recogniser may not use correct standard copy
	SS := sub<Generic(S) | [isoF(g) : g in UserGenerators(D)]>;
	if IsLinearGroup(SS) then
	    cbm := Identity(SS);
	else
	    cbm := TransformForm(SS) * TransformForm(S)^(-1);
	end if;
	cbm := Generic(SS) ! cbm;

	isoFF := function(g)
	    h := Generic(SS) ! isoF(g);
	    //print Category(h), Degree(h), CoefficientRing(h);
	    //print Category(cbm), Degree(cbm), CoefficientRing(cbm);
	    return h^cbm;
	end function;
	
	// Create rule maps
	inv := hom<S -> D | g :-> invF(g^(cbm^(-1)))>;
	iso := hom<D -> S | g :-> isoFF(g)>;

	vprint CompositionTree, 3 : "Black box recognition finished";
	return true, iso, inv, D_slps;
    else
	return false, _, _, _;
    end if;
end function;

// Recognise a C9 symplectic group G with standard copy S
function SpRecognition(G, S, Name, verify, maxVerifyOrder)
    assert IsOdd(#CoefficientRing(S));
    vprint CompositionTree, 3 : "Entering Sp odd recognition";
    
    flag, iso, inv, D_slps :=
	GeneralRecognition(G, S, RecogniseSpOddWrapper, Name);
    if flag then
	return true, iso, inv, iso, D_slps;
    else
	return false, _, _, _, _;
    end if;
end function;

// Recognise a C9 linear group G with standard copy S
function SLRecognition(G, S, Name, verify, maxVerifyOrder)

    d := Degree(S);
    F := CoefficientRing(S);
    p := Characteristic(F);
    
    // SL2 works in black box context
    if d eq 2 then
	return SL2RecognitionBlackBox(G, S);
    end if;
    if d eq 3 then
	return SL3RecognitionGreyBox(G, S, Name);
    end if;
    if Category(G) eq GrpMat and
	IsDivisibleBy(#F, Characteristic(CoefficientRing(G))) then
	if #F eq #CoefficientRing(G) and Degree(G) eq d * (d - 1) div 2 then
	    vprint CompositionTree, 3 :
		"Entering alternating square recognition";

	    flag, iso, inv, D_slps :=
		GeneralRecognition(G, S, RecogniseSLAlternatingSquare, Name);

	    vprint CompositionTree, 3 :
		"Recognised alternating square ";

	    // No word map necessary
	    if flag then		
		return true, iso, inv, iso, D_slps;
	    else
		return false, _, _, _, _;
	    end if;
	elif #F eq #CoefficientRing(G) and not (d eq 4 and #F eq 3) 
           and Degree(G) eq d * (d + 1) div 2 then
	    vprint CompositionTree, 3 :
		"Entering symmetric square recognition";

	    flag, iso, inv, D_slps :=
		GeneralRecognition(G, S, RecogniseSLSymmetricSquare, Name);

	    vprint CompositionTree, 3 :
		"Recognised symmetric square ";

	    // No word map necessary
	    if flag then		
		return true, iso, inv, iso, D_slps;
	    else
		return false, _, _, _, _;
	    end if;
	elif #F eq #CoefficientRing(G) and Degree(G) eq d^2 then
	    vprint CompositionTree, 3 :
		"Entering tensor square recognition";

	    flag, iso, inv, D_slps :=
		GeneralRecognition(G, S, RecogniseSLDelta, Name);

	    vprint CompositionTree, 3 :
		"Recognised tensor square";

	    // No word map necessary
	    if flag then		
		return true, iso, inv, iso, D_slps;
            end if;
	elif #F eq #CoefficientRing(G) and
	    ((d mod p eq 0 and Degree(G) eq d^2 - 2) or
	     (d mod p ne 0 and Degree(G) eq d^2 - 1)) then
	    vprint CompositionTree, 3 :
		"Entering adjoint recognition";

	    flag, iso, inv, D_slps :=
		GeneralRecognition(G, S, RecogniseSLAdjoint, Name);

	    vprint CompositionTree, 3 :
		"Recognised adjoint representation";

	    // No word map necessary
	    if flag then		
		return true, iso, inv, iso, D_slps;
            end if;
	end if;
	
	vprint CompositionTree, 3 : "Entering Kantor-Seress recognition";
	
	flag, iso, inv, D_slps :=
	    GeneralRecognition(G, S, RecogniseSLWrapper, Name);
	
	// No word map necessary in grey box SL
	if flag then	    
	    return true, iso, inv, iso, D_slps;
	else
	    return false, _, _, _, _;
	end if;
    end if;
   
    vprint CompositionTree, 3 : "Entering Kantor-Seress black box";
    
    flag, iso, inv, D_slps :=
	GeneralRecognition(G, S, RecogniseSLWrapper, Name);

    if flag then
	return true, iso, inv, iso, D_slps;
    else
	return false, _, _, _, _;
    end if;
end function;

// Recognise a C9 unitary group G with standard copy S
function SURecognition(G, S, Name, verify, maxVerifyOrder)
    d := Degree(S);
    assert d in {3, 4};

    // SU3 and SU4 works in black box context
    if d eq 3 then
	flag, iso, inv, D_slps :=
	    GeneralRecognition(G, S, SU3RecognitionBlackBoxWrapper, Name);
	if flag then
	    return true, iso, inv, iso, D_slps;
	else
	    return false, _, _, _, _;
	end if;
    end if;
    if d eq 4 then
	flag, iso, inv, D_slps :=
	    GeneralRecognition(G, S, SU4RecognitionBlackBoxWrapper, Name);
	if flag then
	    return true, iso, inv, iso, D_slps;
	else
	    return false, _, _, _, _;
	end if;
    end if;
end function;

procedure BBClassicalFactorData(~node, center, genMapRev)
    F := CoefficientRing(node`LeafData`GoldCopy);
    
    // Brute force for small exceptional cases
    if node`LeafData`Name in NonSimpleClassicalNames or
	Characteristic(F) eq 2 and
	not IsLinearGroup(node`LeafData`GoldCopy) then
	LeafFactorData(~node);
	return;
    end if;

    centerPC, phi := PCGroup(sub<Generic(node`Group) | center, node`Scalar>);
    centerProj, proj := quo<centerPC | phi(node`Scalar)>;
    
    series := CompositionSeries(centerProj);
    centerFactors := PullbackCompFactors(node`NiceData`Group,
	node`SLPMaps`EltToSLPBatch, phi * proj, series, #centerPC + 1);

    W := WordGroup(node`LeafData`GoldCopy);

    // Need membership testing in gold copy
    flag := ClassicalConstructiveRecognition(node`LeafData`GoldCopy);
    assert flag;
        
    // Delete identity generator
    delete node`LeafData`GoldCopy`UserGenerators;    
    stdSlps := node`LeafData`GoldCopy`ClassicalRecognitionData`slps;

    gold2slp := hom<node`LeafData`GoldCopy -> W |
    g :-> Evaluate(Evaluate(MembershipTesting(node, func<h |
	ClassicalElementToWord(node`LeafData`GoldCopy, h)>,
	g), stdSlps), UserGenerators(W))>;
    
    if IsTrivial(centerProj) then
	if IsTrivial(centerPC) then
	    slps := [UserGenerators(node`NiceData`WordGroup)];
	else
	    slps := [Prune(UserGenerators(node`NiceData`WordGroup))];
	end if;
	
	factors := [* node`LeafData`ToGold *];
	factor2slp := [* gold2slp *];
    else
	slps := Append(centerFactors`FactorSLPs,
	    Prune(UserGenerators(node`NiceData`WordGroup)));
	factors := Append(centerFactors`ToFactor, node`LeafData`ToGold);
	factor2slp := Append(centerFactors`FactorToSLP, gold2slp);
    end if;
        
    // Setup data about composition factors
    node`FactorData := rec<FactorInfo |
	FactorSLPs := slps,
	ToFactor := factors,
	FactorToSLP := factor2slp,
	Refined := true,
	FactorLeaves := [node : i in [1 .. #centerFactors`ToFactor + 1]]>;    
end procedure;

function SLShortPresentation(G, W, d, F, genMap, genMapRev, g2slp)
    // Linear groups
    vprint CompositionTree, 7 : "Find PSL presentation";
    Q, relators := ClassicalStandardPresentation("SL", d, #F : Projective := true);
	
    vprint CompositionTree, 7 : "Found PSL presentation";
    return Evaluate(relators, [W.i : i in genMapRev]);	
end function;

function SpShortPresentation(G, W, d, F, genMap, genMapRev, g2slp)
    // Symplectic groups
    Q, relators := ClassicalStandardPresentation("Sp", d, #F : Projective := true);
    return Evaluate(relators, [W.i : i in genMapRev]);
end function;

function SUShortPresentation(G, W, d, F, genMap, genMapRev, g2slp)
    // Symplectic groups
    assert IsEven(Degree(F));
    Fdef := sub<F | Degree(F) div 2>;
    Q, relators := ClassicalStandardPresentation("SU", d, #Fdef : Projective := true);
    return Evaluate(relators, [W.i : i in genMapRev]);
end function;

function GeneralShortPresentation(G, W, d, F, genMap, genMapRev, g2slp)
    pres, pres2G := FPGroup(G);
    relators := [LHS(r) * RHS(r)^(-1) : r in Relations(pres)];
    
    // Convert words to SLPs
    wordToSLP := hom<pres -> W | g2slp(pres2G(UserGenerators(pres)))>;
    return wordToSLP(relators);
end function;

procedure BBClassicalShortPresentation(~node, genMapRev,
    genMap, center, centerToSLP)
    d := Degree(node`LeafData`GoldCopy);

    slps := StandardCopyData[Retrieve(node`LeafData`Family)]`Relators(
	node`NiceData`Group, node`NiceData`WordGroup, d,
	CoefficientRing(node`LeafData`GoldCopy),
	genMap, genMapRev, node`SLPMaps`EltToSLPBatch);

    if not IsTrivial(center) then
	// Center generator is at the end
	slps := C9Presentation(node`NiceData`Group, node`NiceData`WordGroup,
	    slps, NumberOfGenerators(node`NiceData`Group),
	    centerToSLP, node`Scalar);
    end if;

    node`PresentationData := rec<PresentationInfo |
	SLPRelators := slps>;    
end procedure;

// Assume node`NiceData has already been setup
// ExtendedGroup = C9 classical group (some extra centre or scalar flag)
// Setup projective iso from ExtendedGroup to GoldCopy and
// projective word map for ExtendedGroup
// Word map is either BSGS or Ryba
function GeneralSLPMaps(node, ExtendedGroup, iso,
    stdGens, center, genMap, RybaParameters, cbm)

    d := Degree(node`LeafData`GoldCopy);
    F := CoefficientRing(node`LeafData`GoldCopy);
    q := #F;
    
    useRyba, params := RybaParameters(ExtendedGroup, d, q);
        
    if not useRyba then
	if Type(ExtendedGroup) eq GrpMat and
	    HasAttribute(ExtendedGroup, "Base") then 
	    AssertAttribute(node`NiceData`Group, "Base", Base(ExtendedGroup));
	end if;
	
	if HasAttribute(ExtendedGroup, "Order") then
	    AssertAttribute(node`NiceData`Group, "Order",
		#ExtendedGroup div Order(node`Scalar));
	end if;
	
	vprint CompositionTree, 5 : "Find BSGS of nice group";
	RandomSchreier(node`NiceData`Group);

	if node`Verify or #node`NiceData`Group le node`MaxBSGSVerifyOrder or
	    (Category(node`NiceData`Group) eq GrpPerm and
	    Degree(node`NiceData`Group) le MaxBSGSVerifyPermDegree) then
	    Verify(node`NiceData`Group);
	end if;
	
	g2slp := func<G, g | BuiltinMembership(G, g)>;
    else
	vprint CompositionTree, 5 : "Use Ryba for nice group rewriting";
	g2slp := func<G, g | Reduction(G, g :
	    Verify := params[1], Central := params[4], LieRank := params[5],
	    CentraliserMethod := params[2])>;
    end if;
    
    if #node`NiceData`NiceSLPs gt #genMap then
	// We have some center
	gens := Append([stdGens[i]^cbm : i in genMap],
	    Identity(node`LeafData`GoldCopy));
    else
	gens := [stdGens[i]^cbm : i in genMap];
    end if;
    
    vprint CompositionTree, 5 : "Setup projective isomorphisms";
    
    if IsTrivial(sub<Generic(center) | center, node`Scalar>) then
	// ExtendedGroup = D
	// iso to std copy already ok
	EtoD := hom<ExtendedGroup -> node`NiceData`Group | g :-> g>;
    else
	if NumberOfGenerators(ExtendedGroup) gt
	    NumberOfGenerators(node`Group) then
	    // We have a non-trivial scalar flag
	    
	    // Map to remove scalar flag
	    EtoD := hom<ExtendedGroup -> node`NiceData`Group | g :->
	    Evaluate(MembershipTesting(node,
		func<h | g2slp(ExtendedGroup, h)>),
		Append(UserGenerators(node`Group),
		Identity(node`Group)))>;
	else
	    // ExtendedGroup = niceGroup
	    // No scalar flag to remove
	    EtoD := hom<ExtendedGroup -> node`NiceData`Group | g :-> g>;
	end if;
		
	// Iso to standard copy
	iso := hom<ExtendedGroup -> node`LeafData`GoldCopy | g :->
	Evaluate(MembershipTesting(node,
	    func<h | g2slp(node`NiceData`Group, h)>,
	    Function(EtoD)(g)), gens)>;
    end if;

    // Projective membership testing
    g2slpProj := hom<ExtendedGroup -> node`NiceData`WordGroup | g :->
    MembershipTesting(node, func<h | g2slp(node`NiceData`Group, h)>,
	Function(EtoD)(g))>;

    return iso, g2slpProj;
end function;

function CsabaWrapper(G, gens, type, d, q, g)
    flag, slp := ClassicalRewrite(G, gens, type, d, q, g);
    if slp cmpeq false then
	return false, _;
    else
        return true, slp;
    end if;
end function;

function SpSLPMaps(node, ExtendedGroup, iso, stdGens, center,
    genMap, genMapRev, niceInputGens, cbm)
    d := Degree(Codomain(iso));
    F := CoefficientRing(Codomain(iso));
    q := #F;

    if Category(ExtendedGroup) eq GrpMat and
	IsDivisibleBy(q, Characteristic(CoefficientRing(ExtendedGroup))) and
	IsProbablyPerfect(ExtendedGroup) then
	// Elliot works in grey box context

	vprint CompositionTree, 6 : "Sp grey box SLP maps";
	
	// projective iso to standard copy
	iso := hom<ExtendedGroup -> node`LeafData`GoldCopy | g :->
	Evaluate(MembershipTesting(node, func<h |
	    CsabaWrapper(ExtendedGroup, niceInputGens, "Sp", d, q, h)>, g), 
	    stdGens)>;
	
	// Projective membership testing
	g2slpProj := hom<ExtendedGroup -> node`NiceData`WordGroup | g :->
	Evaluate(MembershipTesting(node, func<h |
	    CsabaWrapper(ExtendedGroup, niceInputGens, "Sp", d, q, h)>, g),
	    [i gt 0 select node`NiceData`WordGroup.i else
	    Identity(node`NiceData`WordGroup) : i in genMapRev])>;
	
	return iso, g2slpProj;
    else
	vprint CompositionTree, 6 : "Sp black box SLP maps";
	return GeneralSLPMaps(node, ExtendedGroup, iso, stdGens,
	    center, genMap, SpRybaParameters, cbm);
    end if;
end function;

function SLSLPMaps(node, ExtendedGroup, iso, stdGens,
    center, genMap, genMapRev, niceInputGens, cbm)
    d := Degree(Codomain(iso));
    F := CoefficientRing(Codomain(iso));
    q := #F;
    
    if d eq 2 then
	vprint CompositionTree, 6 : "SL2 SLP maps";

	// Need to recognise standard copy to obtain word map
	flag, _, _, g2slp := RecogniseSL2(node`LeafData`GoldCopy, q);
	assert flag;
	
	if #genMap lt #node`NiceData`NiceSLPs then
	    assert not IsTrivial(center);
	    n := NumberOfGenerators(node`NiceData`Group);
	    assert node`NiceData`Group.n eq center.1;
	    
	    // We have a center generator
	    gens := Prune(UserGenerators(node`NiceData`WordGroup));
	else
	    gens := UserGenerators(node`NiceData`WordGroup);
	end if;

	// Test for membership, return word in standard gens
	GtoW := function(g)
	    flag := SL2ElementToWord(ExtendedGroup, g);
	    if flag then
		return true, g2slp(Function(iso)(g));
	    else
		return false, _;
	    end if;
	end function;
	
	g2slpProj := hom<ExtendedGroup -> node`NiceData`WordGroup |
	g :-> Evaluate(MembershipTesting(node, GtoW, g), gens)>;

    	return iso, g2slpProj;
    elif Category(ExtendedGroup) eq GrpMat and
	IsDivisibleBy(q, Characteristic(CoefficientRing(ExtendedGroup))) then
	// Elliot works in grey box context

	vprint CompositionTree, 6 : "SL grey box SLP maps";
		
	// projective iso to standard copy
	iso := hom<ExtendedGroup -> node`LeafData`GoldCopy | g :->
	Evaluate(MembershipTesting(node, func<h |
	    CsabaWrapper(ExtendedGroup, niceInputGens, "SL", d, q, h)>, g), 
	    [i gt 0 select stdGens[i] else Identity(node`LeafData`GoldCopy) :
	    i in genMapRev])>;
	
	// Projective membership testing
	g2slpProj := hom<ExtendedGroup -> node`NiceData`WordGroup | g :->
	Evaluate(MembershipTesting(node, func<h |
	    CsabaWrapper(ExtendedGroup, niceInputGens, "SL", d, q, h)>, g),
	    [i gt 0 select node`NiceData`WordGroup.i else
	    Identity(node`NiceData`WordGroup) : i in genMapRev])>;
	
	return iso, g2slpProj;
    else
	vprint CompositionTree, 6 : "SL black box SLP maps";
	return GeneralSLPMaps(node, ExtendedGroup, iso, stdGens,
	    center, genMap, SLRybaParameters, cbm);
    end if;
end function;

function SUSLPMaps(node, ExtendedGroup, iso, stdGens,
    center, genMap, genMapRev, niceInputGens, cbm)

    d := Degree(Codomain(iso));
    F := CoefficientRing(Codomain(iso));
    assert IsEven(Degree(F));
    Fdef := sub<F | Degree(F) div 2>;
    q := #Fdef;

    // EOB -- temporary, these are treated as sporadic groups and 
    // SU3ElementToWord(ExtendedGroup, g) does not work for them 
    if not (q in { 3 , 4 , 5 , 8 , 11 , 17 }) and 
	d eq 3 and IsProbablyPerfect(ExtendedGroup) then
	vprint CompositionTree, 6 : "SU3 SLP maps";

	// Need to recognise standard copy to obtain word map
	flag, _, _, g2slp := RecogniseSU3(node`LeafData`GoldCopy, q);
	assert flag;
	
	if #genMap lt #node`NiceData`NiceSLPs then
	    assert not IsTrivial(center);
	    n := NumberOfGenerators(node`NiceData`Group);
	    assert node`NiceData`Group.n eq center.1;
	    
	    // We have a center generator
	    gens := Prune(UserGenerators(node`NiceData`WordGroup));
	else
	    gens := UserGenerators(node`NiceData`WordGroup);
	end if;
	
	// Test for membership, return word in standard gens
	GtoW := function(g)
	    flag := SU3ElementToWord(ExtendedGroup, g);
	    if flag then
		return true, g2slp(Function(iso)(g));
	    else
		return false, _;
	    end if;
	end function;
	
	g2slpProj := hom<ExtendedGroup -> node`NiceData`WordGroup |
	g :-> Evaluate(MembershipTesting(node, GtoW, g), gens)>;

    	return iso, g2slpProj;
    end if;
    
    if d ge 4 and Category(ExtendedGroup) eq GrpMat and
	IsDivisibleBy(q, Characteristic(CoefficientRing(ExtendedGroup))) and
	IsProbablyPerfect(ExtendedGroup) then
	// Elliot works in grey box context

	vprint CompositionTree, 6 : "SU grey box SLP maps";
	
	// Elliot wants everything to be in F
	if CoefficientRing(ExtendedGroup) eq Fdef then
	    H, inc := ExtendField(Generic(ExtendedGroup), F);
	    E := sub<H | [inc(h) : h in UserGenerators(ExtendedGroup)]>;
	else
	    E := ExtendedGroup;
	    inc := func<x | x>;
	end if;
        
	gens := ChangeUniverse(niceInputGens, GL(Degree(ExtendedGroup), F));
	assert CoefficientRing(Universe(stdGens)) eq F;
	
	elliotCore := function(h)
	    flag, w := CsabaWrapper(ExtendedGroup, niceInputGens, "SU", d, q, h);
	    if flag then
		return flag, w;
	    else
		return false, _;
	    end if;
	end function;
    
	elliot := function(g)
	    w := MembershipTesting(node, elliotCore, g);
	    h := Evaluate(w, [stdGens[i] : i in [1 .. #stdGens]]);
	    assert CoefficientRing(h) eq F;
	    return h;
	end function;
    
	// projective iso to standard copy
	iso := hom<ExtendedGroup -> node`LeafData`GoldCopy | g :->
	elliot(g)>;
	
	g2slpProj := hom<ExtendedGroup -> node`NiceData`WordGroup | g :->
	Evaluate(MembershipTesting(node, elliotCore, g),
	    [i gt 0 select node`NiceData`WordGroup.i else
	    Identity(node`NiceData`WordGroup) : i in genMapRev])>;
	
	return iso, g2slpProj;
    else    
	vprint CompositionTree, 6 : "SU black box SLP maps";
	return GeneralSLPMaps(node, ExtendedGroup, iso, stdGens,
	    center, genMap, SURybaParameters, cbm);
    end if;
end function;

// Setup nice generators for C9 classical group
// ExtendedGroup may include scalar flag, nice group is node group without
// scalar flag
// inv : gold copy -> (derived group of) ExtendedGroup 
// g2slp is word map for (derived group of) ExtendedGroup
// natCopy is gold copy on generators of (derived group of) ExtendedGroup
//
// niceInputGens are images in input copy of stdGens
procedure GivenSLPMapNiceData(~node, ~niceInputGens, ~cbm,
    ExtendedGroup, DerivedSLPs, inv, g2slp, natCopy, stdGens, genMap)

    cbm := Identity(node`LeafData`GoldCopy);
    
    vprint CompositionTree, 5 : "Find nice generators";
    niceInputGens := [Generic(node`Group) | Function(inv)(g) : g in stdGens];
	
    niceSLPs := [Function(g2slp)(g) : g in niceInputGens];
    vprint CompositionTree, 5 : "Found SLPs of nice generators";
    
    // Remove scalar flag if necessary
    if NumberOfGenerators(ExtendedGroup) gt NumberOfGenerators(node`Group) then
	assert NumberOfGenerators(Universe(niceSLPs)) eq
	    NumberOfGenerators(node`WordGroup) + 1;
	
	niceSLPs := Evaluate(niceSLPs, Append(UserGenerators(node`WordGroup),
	    Identity(node`WordGroup)));
    else
	assert NumberOfGenerators(Universe(niceSLPs)) eq
	    NumberOfGenerators(node`WordGroup);
	
	niceSLPs := Evaluate(niceSLPs, UserGenerators(node`WordGroup));
    end if;
    
    // Remove duplicate generators in GrpPerm context
    niceGroup := sub<Generic(ExtendedGroup) | {@ g : g in niceInputGens |
	g ne Identity(node`Group) @}>;
    niceSLPs := niceSLPs[genMap];
    assert #niceSLPs eq NumberOfGenerators(niceGroup);
    
    node`NiceData := rec<NiceInfo |
	Group := niceGroup,
	WordGroup := WordGroup(niceGroup),
	NiceSLPs := niceSLPs,
	NiceToUser :=
	hom<WordGroup(niceGroup) -> node`WordGroup | niceSLPs>,
    NiceToUserBatch := func<seq | Evaluate(seq, niceSLPs)>>;

    vprint CompositionTree, 5 : "Found nice group";
end procedure;

// Combine projective word map for extended classical group with centre
// word map to obtain precise word map for ExtendedGroup
procedure ObtainPreciseSLPMaps(~node, ExtendedGroup,
    g2slpProj, center, centerMembership)
	
    // Constructive membership testing with exception handling
    if not IsTrivial(sub<Generic(center) | center, node`Scalar>) then
	g2slp := hom<ExtendedGroup -> node`NiceData`WordGroup | g :->
	C9Membership(node`NiceData`Group, g2slpProj, centerMembership, g)>;
    else
	g2slp := g2slpProj;
    end if;
    
    node`SLPMaps := rec<SLPMapsInfo |
	EltToSLP := g2slp,
	EltToSLPBatch := func<seq | [Function(g2slp)(g) : g in seq]>,
	SLPToElt := hom<node`NiceData`WordGroup -> node`NiceData`Group |
    UserGenerators(node`NiceData`Group)>,
	SLPToEltBatch := func<slps |
	Evaluate(slps, UserGenerators(node`NiceData`Group))>>;
end procedure;

// Assume node`NiceData has already been setup
// Find centre of node group and possibly add centre generators to nice group
// Also setup rewriting in centre
procedure ObtainCenter(~node, ~center, ~centerMembership)    
    // Add extra nice centre generator
    center, centerSLP := C9Centre(node`Group);
    
    if not IsTrivial(center) then
	n := NumberOfGenerators(node`NiceData`Group);
	node`NiceData`Group := sub<Generic(node`Group) |
	    node`NiceData`Group, center>;
	node`NiceData`WordGroup := WordGroup(node`NiceData`Group);
	
	if NumberOfGenerators(node`NiceData`Group) gt n then
	    Append(~node`NiceData`NiceSLPs, centerSLP);
	end if;
    end if;

    // Centre membership testing
    n := NumberOfGenerators(node`NiceData`WordGroup);
    centerMembership := hom<center -> node`NiceData`WordGroup | g :->
    Evaluate(MembershipTesting(node, func<h |
	C9CenterMembership(center, node`Scalar, h)>, g),
	[node`NiceData`WordGroup.n])>;

    niceGens := Evaluate(Prune(node`NiceData`NiceSLPs),
	UserGenerators(node`Group));
    centralElts := [niceGens[i] / node`NiceData`Group.i :
	i in [1 .. #niceGens]];
    
    // Patch nice slps so that they evaluate correctly to the nice generators
    if not IsTrivial(center) then
	patchSLPs := Evaluate([Function(centerMembership)(g) :
	    g in centralElts], node`NiceData`NiceSLPs);

	node`NiceData`NiceSLPs := Append([node`NiceData`NiceSLPs[i] *
	    patchSLPs[i]^(-1) : i in [1 .. #patchSLPs]],
	    node`NiceData`NiceSLPs[#node`NiceData`NiceSLPs]);
    else
	assert SequenceToSet(centralElts) eq {Identity(node`Group)};
    end if;

    node`NiceData`NiceToUser := hom<node`NiceData`WordGroup ->
    node`WordGroup | node`NiceData`NiceSLPs>;
    node`NiceData`NiceToUserBatch :=
	func<seq | Evaluate(seq, node`NiceData`NiceSLPs)>;
end procedure;

procedure GeneralNiceData(~node, ~niceInputGens, ~cbm, ExtendedGroup,
    DerivedSLPs, inv, natCopy, stdGens, genMap)
	
    cbm := Identity(node`LeafData`GoldCopy);

    vprint CompositionTree, 5 : "Find standard generators of natural copy";
    
    // Recognise natural copy on input gens to obtain standard gens as
    // words in input gens
    flag := ClassicalConstructiveRecognition(natCopy);
    if not flag then
	error ERROR_RECOGNITION;
    end if;

    vprint CompositionTree, 5 : "Obtain SLPs of standard generators";    
    stdSLPs := [];
    for g in stdGens do
	flag, slp := ClassicalElementToWord(natCopy, g^cbm);
	if not flag then
	    error ERROR_RECOGNITION;
	end if;

	Append(~stdSLPs, Evaluate(slp, natCopy`ClassicalRecognitionData`slps));
    end for;

    assert NumberOfGenerators(Universe(stdSLPs)) in
	{#DerivedSLPs, 1 + #DerivedSLPs};

    // Obtain standard gens as words in input group with scalar
    if NumberOfGenerators(Universe(stdSLPs)) eq #DerivedSLPs + 1 then
	niceSLPs := Evaluate(stdSLPs, [Identity(Universe(DerivedSLPs))] cat
	    DerivedSLPs);
    else
	niceSLPs := Evaluate(stdSLPs, DerivedSLPs);
    end if;
    
    assert NumberOfGenerators(Universe(niceSLPs)) eq
	NumberOfGenerators(ExtendedGroup);
    niceInputGens := Evaluate(niceSLPs, UserGenerators(ExtendedGroup));

    // Remove duplicate generators in GrpPerm context
    niceGroup := sub<Generic(ExtendedGroup) | {@ g : g in niceInputGens |
	g ne Identity(node`Group) @}>;

    //print niceGroup;
    //print SymplecticForm(niceGroup);
    
    niceSLPs := niceSLPs[genMap];
    assert #niceSLPs eq NumberOfGenerators(niceGroup);

    // Remove scalar flag if necessary
    if NumberOfGenerators(ExtendedGroup) gt NumberOfGenerators(node`Group) then
	assert NumberOfGenerators(Universe(niceSLPs)) eq
	    NumberOfGenerators(node`WordGroup) + 1;
	
	niceSLPs := Evaluate(niceSLPs, Append(UserGenerators(node`WordGroup),
	    Identity(node`WordGroup)));
    else
	assert NumberOfGenerators(Universe(niceSLPs)) eq
	    NumberOfGenerators(node`WordGroup);
	
	niceSLPs := Evaluate(niceSLPs, UserGenerators(node`WordGroup));
    end if;

    node`NiceData := rec<NiceInfo |
	Group := niceGroup,
	WordGroup := WordGroup(niceGroup),
	NiceSLPs := niceSLPs,
	NiceToUser :=
	hom<WordGroup(niceGroup) -> node`WordGroup | niceSLPs>,
    NiceToUserBatch := func<seq | Evaluate(seq, niceSLPs)>>;    

    vprint CompositionTree, 5 : "Found nice data";    
end procedure;
    
procedure SpNiceData(~node, ~niceInputGens, ~cbm,
    ExtendedGroup, DerivedSLPs, inv, g2slp, natCopy, stdGens, genMap)
    GeneralNiceData(~node, ~niceInputGens, ~cbm,
	ExtendedGroup, DerivedSLPs, inv, natCopy, stdGens, genMap);
end procedure;

procedure SLNiceData(~node, ~niceInputGens, ~cbm, ExtendedGroup, DerivedSLPs,
    inv, g2slp, natCopy, stdGens, genMap)

    // The word map g2slp for the input copy is efficient for SL2 and SL3 
    if Degree(node`LeafData`GoldCopy) lt 4 then
	GivenSLPMapNiceData(~node, ~niceInputGens, ~cbm,
	    ExtendedGroup, DerivedSLPs,
	    inv, g2slp, natCopy, stdGens, genMap);
	return;
    end if;
    
    GeneralNiceData(~node, ~niceInputGens, ~cbm, ExtendedGroup, DerivedSLPs,
	inv, natCopy, stdGens, genMap);
end procedure;

procedure SUNiceData(~node, ~niceInputGens, ~cbm, ExtendedGroup, DerivedSLPs,
    inv, g2slp, natCopy, stdGens, genMap)

    assert Degree(node`LeafData`GoldCopy) in {3, 4};

    F := CoefficientRing(node`LeafData`GoldCopy);
    assert IsEven(Degree(F));
    Fdef := sub<F | Degree(F) div 2>;
    q := #Fdef;

    // The word map g2slp for the input copy is efficient for SU3
    if Degree(node`LeafData`GoldCopy) eq 3 and
	IsProbablyPerfect(ExtendedGroup) then
	flag, _, _, slpMap := RecogniseSU3(ExtendedGroup, #Fdef);
	assert flag;
	
	GivenSLPMapNiceData(~node, ~niceInputGens, ~cbm,
	    ExtendedGroup, DerivedSLPs,
	    inv, slpMap, natCopy, stdGens, genMap);
	return;
    end if;

    if Degree(node`LeafData`GoldCopy) le 4 then
	GeneralNiceData(~node, ~niceInputGens, ~cbm,
	    ExtendedGroup, DerivedSLPs, inv, natCopy, stdGens, genMap);
	return;
    end if;
end procedure;

procedure ClassicalRecogBlackBox(~node)
    try
	vprint CompositionTree, 2 : "Entering Classical C9 recognition";
	d := StandardCopyData[Retrieve(node`LeafData`Family)]`
	    RankToDim(node`LeafData`LieRank);
	q := Retrieve(node`LeafData`DefiningField);
	goldCopy := StandardCopyData[Retrieve(node`LeafData`Family)]`
	    Creator(d, q);

	ExtendedGroup := sub<Generic(node`Group) | node`Group, node`Scalar>;
	vprint CompositionTree, 4 : "Setup standard copy";
	
	// Standard copy has no scalars	
	stdGens := ClassicalStandardGenerators(node);
	node`LeafData`GoldCopy := sub<Generic(Universe(stdGens)) | stdGens>;
	
	// Some standard generators might coincide
	genMap := [Index(stdGens, g) :
	    g in UserGenerators(node`LeafData`GoldCopy)];
	genMapRev :=
	    [Index(ChangeUniverse(UserGenerators(node`LeafData`GoldCopy),
	    Generic(node`LeafData`GoldCopy)), g) : g in stdGens];
		
	vprint CompositionTree, 4 :
	    "Entering black box classical constructive recognition";
	verify := node`Verify or
	    (Category(node`Group) eq GrpPerm and
	    Degree(node`Group) le MaxBSGSVerifyPermDegree);
	flag, iso, inv, g2slp, DerivedSLPs :=
	    StandardCopyData[Retrieve(node`LeafData`Family)]`
	    Recogniser(ExtendedGroup, node`LeafData`GoldCopy,
	    node`LeafData`Name, verify, node`MaxBSGSVerifyOrder);

	if not flag then
	    error ERROR_RECOGNITION;
	end if;

	vprint CompositionTree, 4 : "Obtain natural copy on input gens";
	natCopy := sub<Generic(node`LeafData`GoldCopy) | [Function(iso)(g) :
	    g in UserGenerators(Domain(iso))]>;

	vprint CompositionTree, 4 : "Setup nice data";
	StandardCopyData[Retrieve(node`LeafData`Family)]`
	    NiceData(~node, ~niceInputGens, ~cbm, ExtendedGroup, DerivedSLPs,
	    inv, g2slp, natCopy, stdGens, genMap);
			
	vprint CompositionTree, 4 : "Find center";
	ObtainCenter(~node, ~center, ~centerMembership);

	// We can easily write down a presentation for SL and Sp
	//node`FastVerify := Retrieve(node`LeafData`Family) in {"A", "C"} or
	//  (Retrieve(node`LeafData`Family) eq "2A" and
	//  IsOdd(node`LeafData`LieRank));
	node`FastVerify := true;
	
	vprint CompositionTree, 4 : "Setup SLP maps";
	iso, g2slpProj := StandardCopyData[Retrieve(node`LeafData`Family)]`
	    SLPMaps(node, ExtendedGroup, iso, stdGens, center,
	    genMap, genMapRev, niceInputGens, cbm);

	ObtainPreciseSLPMaps(~node, ExtendedGroup,
	    g2slpProj, center, centerMembership);
	
	vprint CompositionTree, 4 : "Setup gold copy maps";
	        
	node`LeafData`ToGold := hom<ExtendedGroup -> node`LeafData`GoldCopy |
	g :-> Generic(node`LeafData`GoldCopy) ! Function(iso)(g)>;
	node`LeafData`FromGold := hom<node`LeafData`GoldCopy -> ExtendedGroup |
	g :-> Generic(ExtendedGroup) ! Function(inv)(g)>;
	node`LeafData`FromGoldBatch :=
	    func<seq | [Function(node`LeafData`FromGold)(g) : g in seq]>;
	node`LeafData`ToGoldBatch :=
	    func<seq | [Function(node`LeafData`ToGold)(g) : g in seq]>;

        // change by EOB -- set up order and factored order 
        f := FactoredClassicalGroupOrder(ClassicalType(node`LeafData`GoldCopy),
            Degree(node`LeafData`GoldCopy),
            CoefficientRing(node`LeafData`GoldCopy));
        AssertAttribute (node`LeafData`GoldCopy, "FactoredOrder", f);
        // order automatically set by FactoredOrder assignment 
        // AssertAttribute (node`LeafData`GoldCopy, "Order", o);
	
	node`FindFactors := proc< ~node |
	    BBClassicalFactorData(~node, center, genMapRev)>;
	node`FindPresentation := proc< ~node |
	    BBClassicalShortPresentation(~node,
	    genMapRev, genMap, center, centerMembership)>;

	vprint CompositionTree, 2 : "Classical black box leaf case finished";
    catch err
	if err`Object cmpeq ERROR_RECOGNITION_CATCH then
	    error ERROR_RECOGNITION;
	end if;
	
	error Error(rec<LeafError | Description := "Classical black box", 
	    Error := err>);
    end try;
end procedure;

/*****************************************************************************/
/* AUXILIARY FUNCTIONS                                                       */
/*****************************************************************************/

function ClassicalStandardGenerators(node)
    d := StandardCopyData[Retrieve(node`LeafData`Family)]`
	RankToDim(node`LeafData`LieRank);
    q := Retrieve(node`LeafData`DefiningField);
    goldCopy := StandardCopyData[Retrieve(node`LeafData`Family)]`
	Creator(d, q);
        
    return ChangeUniverse(StandardCopyData[Retrieve(node`LeafData`Family)]`
	StandardGenerators(goldCopy), Generic(goldCopy));
end function;

function SU3RecognitionBlackBoxWrapper(G, d, q)
    F := GF(q);
    assert IsEven(Degree(F));
    Fdef := sub<F | Degree(F) div 2>;
    qq := #Fdef;
    
    vprint CompositionTree, 3 : "Entering SU3 Recognition";
    for i in [1 .. RecognitionTries] do
 	flag, isoF, invF, g2slp, slp2g := RecogniseSU3(G, qq);
	
	if flag then
	    return true, Function(isoF), Function(invF);
	end if;
    end for;
    
    return false, _, _;
end function;

function SU4RecognitionBlackBoxWrapper(G, d, q)
    qq := Isqrt(q);

    vprint CompositionTree, 3 : "Entering SU4 Recognition";
    for i in [1 .. RecognitionTries] do
	flag, isoF, invF, g2slp, slp2g := RecogniseSU4(G, qq);
	
	if flag then
	    return true, Function(isoF), Function(invF);
	end if;
    end for;
    
    return false, _, _;
end function;

function SL2RecognitionBlackBox(G, S)
    d := Degree(S);
    F := CoefficientRing(S);
    assert d eq 2;

    vprint CompositionTree, 3 : "Entering SL2 Recognition";
    flag, isoF, invF, g2slp := RecogniseSL2(G, #F);
	
    if flag then
	inv := hom<S -> G | g :-> Function(invF)(g)>;
	iso := hom<G -> S | g :-> Function(isoF)(g)>;
	
	return true, iso, inv, g2slp, UserGenerators(WordGroup(G));
    else
	return false, _, _, _, _;
    end if;
end function;

function SL3RecognitionGreyBox(G, S, Name)
    d := Degree(S);
    F := CoefficientRing(S);
    assert d eq 3;

    if not IsProbablyPerfect(G : NumberRandom := PerfectCheckElements(G)) then
	repeat
	    D, D_slps := DerivedGroupMonteCarlo(G);
	    flag, name := LieType(D, Characteristic(F));
	until flag and name eq Name;
    else
	D := G;
	D_slps := UserGenerators(WordGroup(G));
    end if;

    vprint CompositionTree, 3 : "Entering SL3 Recognition";
    flag, isoF, invF, g2slp := RecogniseSL3(D, F);
    
    if flag then
	inv := hom<S -> D | g :-> Function(invF)(g)>;
	iso := hom<D -> S | g :-> Function(isoF)(g)>;
	g2slpG := hom<G -> WordGroup(G) | g :-> Evaluate(Function(g2slp)(g),
	D_slps)>;
	return flag, iso, inv, g2slpG, D_slps;
    else
	return false, _, _, _, _;
    end if;
end function;

function RecogniseSLAlternatingSquare(G, d, q)
    flag, H := RecogniseAlternatingSquare(G);
    
    if flag then
	iso := func<g | AlternatingSquarePreimage(G, g)>;
	M := GModule(ExteriorSquare(H));
	flag, cbm := IsIsomorphic(GModule(G), M);
	assert flag;
	
	inv := func<g | Generic(G) ! (cbm * ExteriorSquare(g) * cbm^(-1))>;

	return true, iso, inv;
    else
	return false, _, _;
    end if;
end function;

function RecogniseSLSymmetricSquare(G, d, q)
    flag, H := RecogniseSymmetricSquare(G);
    
    if flag then
	MA := MatrixAlgebra(CoefficientRing(G), Degree(H));
	
	iso := func<g | SymmetricSquarePreimage(G, g)>;
	M := SymmetricSquare(GModule(H));
	flag, cbm := IsIsomorphic(GModule(G), M);
	assert flag;
	    
	inv := func<g | Generic(G) !
	    (cbm * SymmetricSquare(MA ! g) * cbm^(-1))>;

	return true, iso, inv;
    else
	return false, _, _;
    end if;
end function;

/*
DeltaSquare := function (G, e)
   if Type (G) eq GrpMat then 
      d := Degree (G);
      F := BaseRing (G);
      X := [G.i: i in [1..Ngens (G)]];
      Y := [FrobeniusImage (G.i, e): i in [1..Ngens (G)]];
      Z := [KroneckerProduct (X[i], Y[i]): i in [1..#X]];
      return sub <GL (d^2, F) | Z>;
   else 
      h := FrobeniusImage (G, e);
      return TensorProduct (G, h);
   end if;
end function;
*/

function RecogniseSLDelta(G, d, q)
    flag, H, e := RecogniseDelta(G);
    
    if flag then
	MA := MatrixAlgebra(CoefficientRing(G), Degree(H));
	
	iso := func<g | DeltaPreimage(G, g)>;
	H := sub<Generic(H) | [s where _, s is ScaleMatrix(g) :
	    g in UserGenerators(H)]>;
	M := GModule(H);
	D := TensorProduct(M, FrobeniusImage(M, H, e));
	assert NumberOfActionGenerators(D) eq
	    NumberOfActionGenerators(GModule(G));

	flag, cbm := IsIsomorphic(GModule(G), D);
	assert flag;
	    
	inv := func<g | Generic(G) !
	    (cbm * DeltaSquare(MA ! g, e) * cbm^(-1))>;

	return true, iso, inv;
    else
	return false, _, _;
    end if;
end function;

function RecogniseSLAdjoint(G, d, q)
    flag, H := RecogniseAdjoint(G);
    
    if flag then
	MA := MatrixAlgebra(CoefficientRing(G), Degree(H));
	
	iso := func<g | AdjointPreimage(G, g)>;
	M := GModule(H);
	D := TensorProduct(Dual(M), M);
	_, factors, cbm := CompositionSeries(D);
	flag := exists(k){k : k in [1 .. #factors] | Dimension(factors[k]) 
	    eq Degree(G)};
	assert flag;
	X := factors[k];
	
	preDim := k gt 1 select &+[Dimension(factors[i]) :
	    i in [1 .. k - 1]] else 0;
	
	assert NumberOfActionGenerators(X) eq
	    NumberOfActionGenerators(GModule(G));
	
	flag, cbm := IsIsomorphic(GModule(G), X);
	assert flag;
	    
	inv := func<g | Generic(G) !
	    (cbm * Submatrix(TensorProduct(Transpose(g^(-1)), g),
	    preDim + 1, preDim + 1, Dimension(X), Dimension(X)) * cbm^(-1))>;

	return true, iso, inv;
    else
	return false, _, _;
    end if;
end function;

function RecogniseSpOddWrapper(G, d, q)
    for i in [1 .. RecognitionTries] do
	flag, iso, inv := RecogniseSpOdd(G, d, q);
	if flag then
	    return true, Function(iso), Function(inv);
	end if;
    end for;

    return false, _, _;
end function;

function RecogniseSLWrapper(G, d, q)
    for i in [1 .. RecognitionTries] do
	flag, iso, inv := RecogniseSL(G, d, q);
	if flag then
	    return true, iso, inv;
	end if;
    end for;

    return false, _, _;
end function;
