freeze;

digits := {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9" };

parse_L2p := function(str)
    if #str gt 2 and
	str[1..2] eq "L2" and forall{i:i in [3..#str] | str[i] in digits}
    then
	p := StringToInteger(str[3..#str]);
	if p ge 7 and p le 10000 and IsPrime(p) then
	    return true, "L2", p;
	else
	    return false, _, _;
	end if;
    elif #str gt 3 and
	str[1..3] eq "2L2" and forall{i:i in [4..#str] | str[i] in digits}
    then
	p := StringToInteger(str[4..#str]);
	if p ge 7 and p le 10000 and IsPrime(p) then
	    return true, "2L2", p;
	else
	    return false, _, _;
	end if;
    else
	return false, _, _;
    end if;
end function;

StandardGens := function(G, gpname : Projective:=false, MaxTries:=2000 )
/* Return standard generators a,b, as defined in the online ATLAS
 * http://brauer.maths.qmul.ac.uk/Atlas/v3/
 * of the (permutation or matrix) group G known to be isomorphic to
 * the nearly simple group with name "gpname".
 * the names are as returned by the Magma command ATLASGroupNames();
 * For matrix groups, if Projective is true, then work modulo scalars.
 * It works in the same way as StandardGenerators and
 * returns true/false, [a,b], [wa,wb],
 * where wa,wb are SLPs for a,b.
 */
  local O, CO, R, S, P, gens, z, a, b, bb, A, B, BB, c, C, d, D, oa, oA,
        o, seeking, nt, ct, sct;
  nt := MaxTries;
  O := Projective select ProjectiveOrder else Order;
  CO := CentralOrder;
  S := SLPGroup(Ngens(G));
  gens := [G.i : i in [1..Ngens(G)]];
  P := RandomProcess(S);
  R := function()
       X := Random(P);
       x := Evaluate(X, gens);
       return x, X;
  end function;

  F := function(n) //try to find an element of order n
    local ct;
    ct:=0;
    repeat ct+:=1;  a,A := R(); until O(a) mod n eq 0 or ct eq nt;
    if ct eq nt then return false, _, _; end if; 
    o := O(a) div n; a:=a^o; A:=A^o;
    return true, a, A;
  end function;

  Fb := function(n) //try to find an element of order n not using powers
    local ct;
    ct:=0;
    repeat ct+:=1;  a,A := R(); until O(a) eq n or ct eq nt;
    if ct eq nt then return false, _, _; end if; 
    return true, a, A;
  end function;

  FC := function(n,m) //try to find an element of central order n, order m
    local ct, ct2;
    ct2:=0;
    repeat  ct2+:=1 ;
      ct:=0;
      repeat ct+:=1;  a,A := R();
             until O(a) mod m eq 0 and CO(a) mod n eq 0 or ct eq nt;
      if ct eq nt then return false, _, _; end if; 
      o := O(a) div m; a:=a^o; A:=A^o;
      if CO(a) eq n then
        return true, a, A;
      end if;
    until ct2 eq nt;
    if ct2 eq nt then return false, _, _; end if;
  end function;

    //deal with all of the L2p together
    fl, head, p := parse_L2p(gpname);
    if fl then
	if head eq "L2" then
	  z,a,A := F(2);
	  if not z then return false,_,_; end if;
	  z,b,B := F(3);
	  if not z then return false,_,_; end if;
	  bb := b; BB := B;
	  ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
		 until O(a*b) eq p or ct eq nt;
	  if ct eq nt then return false, _, _; end if;
	  return true, [a,b], [A,B];
	elif head eq "2L2" then
	  z,a,A := F(4);
	  if not z then return false,_,_; end if;
	  z,b,B := F(3);
	  if not z then return false,_,_; end if;
	  bb := b; BB := B;
	  ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
		 until O(a*b) eq p or ct eq nt;
	  if ct eq nt then return false, _, _; end if;
	  return true, [a,b], [A,B];
	else
	    error "This cannot happen!!";
	end if;
    end if;

  case gpname:
   when "A5": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 5 or ct eq nt;
   when "2A5": 
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 5 or ct eq nt;
   when "A6": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(4);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 5 or ct eq nt;
   when "2A6": 
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(8);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 5 and O(a*b^2) eq 5 or ct eq nt;
   when "3A6": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(4);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until CO(a*b) eq 5 or ct eq nt;
     assert O(a*b) eq 15;
   when "6A6": 
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(8);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 15 and O(a*b^2) eq 5 or ct eq nt;
   when "L28": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "A7": 
     z,a,A := F(6); //to square into Class 3A
     if not z then return false,_,_; end if;
     o := O(a) div 3; a:=a^o; A:=A^o;
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "2A7": 
     z,a,A := FC(6,12); //to square into Class 3A
     if not z then return false,_,_; end if;
     o := O(a) div 3; a:=a^o; A:=A^o;
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "3A7": 
     z,a,A := FC(6,6); //to square into Class 3A
     if not z then return false,_,_; end if;
     o := O(a) div 3; a:=a^o; A:=A^o;
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "6A7": 
     z,a,A := FC(6,12); //to square into Class 3A
     o := O(a) div 3; a:=a^o; A:=A^o;
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "L216": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 15 or ct eq nt;
   when "L33": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     //need b in Class 3B, but cannot check that directly!
     seeking:=true; ct:=0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false, _, _; end if;
       z,b,B := Fb(3);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C;
         if O(a*b) eq 13 and O(a*b*a*b^2) eq 4 then
           seeking:=false; break;
         end if;
       end for;
     end while;
   when "U33": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(6);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "L225":
//Not in ATLAS database, so I have chosen my own standard generators
//standard generators for PSL(2,25), a order 2, b order 3, order(ab)=12.
//standard generators for SL(2,25), a order 4, b order 3, order(ab)=24,
//order(abab^-1ab) = 13.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 12 or ct eq nt;
   when "2L225":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until CO(a*b) eq 12 and O(a*b) eq 24 and O(a*b*a*b^-1*a*b) eq 13
                  or ct eq nt;
   when "M11": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(4);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 11 and O(a*b*a*b^2*a*b^3) eq 5 or ct eq nt;
   when "L227": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "2L227": 
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "A8": 
     z,a,A := F(15); //to power into 3A
     if not z then return false,_,_; end if;
     o := O(a) div 3; a:=a^o; A:=A^o;
     z,b,B := F(7);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 6 and O(a*b^2) eq 15 or ct eq nt;
   when "2A8": 
     z,a,A := F(15); //to power into 3A
     if not z then return false,_,_; end if;
     o := O(a) div 3; a:=a^o; A:=A^o;
     z,b,B := F(7);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until CO(a*b) eq 6 and CO(a*b^2) eq 15 or ct eq nt;
     assert O(a*b) eq 6 and O(a*b^2) eq 30;
   when "L34": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(4);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 and O(a*b^2) eq 5 or ct eq nt;
   when "3L34": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(4);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until CO(a*b) eq 7 and CO(a*b^2) eq 5 or ct eq nt;
     assert O(a*b) eq 21 and O(a*b^2) eq 5;
   when "2L34": //this one is tricky! 
     seeking := true; ct:=0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false, _, _; end if;
       z,a,A := FC(2,2);
       if not z then return false,_,_; end if;
       z,b,B := FC(4,4);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C; 
         if CO(a*b) eq 7 and CO(a*b^2) eq 5 and CO(a*b*a*b*a*b^3) eq 5 then
            if O(a*b^2) eq 10 then a *:= (a*b^2)^5; A *:= (A*B^2)^5; end if;
            if O(a*b) eq 14 then b *:= (a*b)^7; B *:= (A*B)^7; end if;
            if O(a*b*a*b*a*b^3) eq 5 then
              assert O(a*b) eq 7 and O(a*b^2) eq 5 and O(a*b*a*b*a*b^3) eq 5;
              seeking:=false; break;
            end if;
         end if;
       end for;
     end while;
     assert O(a*b) eq 7 and O(a*b^2) eq 5 and O(a*b*a*b*a*b^3) eq 5;
   when "4aL34":
     seeking := true; ct:=0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false, _, _; end if;
       z,a,A := FC(2,2);
       if not z then return false,_,_; end if;
       z,b,B := FC(4,4);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C; 
         if CO(a*b) eq 7 and CO(a*b^2) eq 5 then
            if O(a*b^2) eq 20 then a *:= (a*b^2)^5; A *:= (A*B^2)^5; end if;
            if O(a*b) eq 28 then b *:= (a*b)^7; B *:= (A*B)^7; end if;
            if O(a*b^2) eq 10 then a *:= (a*b^2)^5; A *:= (A*B^2)^5; end if;
            if O(a*b) eq 14 then b *:= (a*b)^7; B *:= (A*B)^7; end if;
            assert O(a*b) eq 7 and O(a*b^2) eq 5 and O(a*b*a*b*a*b^3) eq 5;
            seeking:=false; break;
         end if;
         if O(a*b) eq 7 and O(a*b^2) eq 5 then
            seeking:=false; break;
         end if;
       end for;
     end while;
   when "4bL34":
     seeking := true; ct:=0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false, _, _; end if;
       z,a,A := FC(2,2);
       if not z then return false,_,_; end if;
       z,b,B := FC(4,4);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C; 
         if CO(a*b) eq 7 and CO(a*b^2) eq 5 then
            if O(a*b^2) eq 20 then a *:= (a*b^2)^5; A *:= (A*B^2)^5; end if;
            if O(a*b) eq 28 then b *:= (a*b)^7; B *:= (A*B)^7; end if;
            if O(a*b^2) eq 10 then a *:= (a*b^2)^5; A *:= (A*B^2)^5; end if;
            if O(a*b) eq 14 then b *:= (a*b)^7; B *:= (A*B)^7; end if;
            assert O(a*b) eq 7 and O(a*b^2) eq 5 and O(a*b*a*b*a*b^3) eq 10;
            seeking:=false; break;
         end if;
       end for;
     end while;
   when "6L34":
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false, _, _; end if;
       z,a,A := FC(2,2);
       if not z then return false,_,_; end if;
       z,b,B := FC(4,4);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C; 
         if CO(a*b) eq 7 and CO(a*b^2) eq 5 and CO(a*b*a*b*a*b^3) eq 5 then
            if O(a*b^2) eq 10 then a *:= (a*b^2)^5; A *:= (A*B^2)^5; end if;
            if O(a*b) eq 42 then b *:= (a*b)^21; B *:= (A*B)^21; end if;
            if O(a*b*a*b*a*b^3) eq 5 then
              assert O(a*b) eq 21 and O(a*b^2) eq 5 and O(a*b*a*b*a*b^3) eq 5;
              seeking:=false; break;
            end if;
         end if;
       end for;
     end while;
     assert O(a) eq 2 and O(b) eq 4 and O(a*b) eq 21 and
            O(a*b^2) eq 5 and O(a*b*a*b*a*b^3) eq 5;
   when "12aL34":
     seeking := true; ct:=0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false, _, _; end if;
       z,a,A := FC(2,2);
       if not z then return false,_,_; end if;
       z,b,B := FC(4,4);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C; 
         if CO(a*b) eq 7 and CO(a*b^2) eq 5 then
            if O(a*b^2) eq 20 then a *:= (a*b^2)^5; A *:= (A*B^2)^5; end if;
            if O(a*b) eq 84 then b *:= (a*b)^21; B *:= (A*B)^21; end if;
            if O(a*b^2) eq 10 then a *:= (a*b^2)^5; A *:= (A*B^2)^5; end if;
            if O(a*b) eq 42 then b *:= (a*b)^21; B *:= (A*B)^21; end if;
            if O(a*b) eq 21 and O(a*b^2) eq 5 and O(a*b*a*b*a*b^3) eq 5 then
              seeking:=false; break;
            end if;
         end if;
       end for;
     end while;
   when "12bL34":
     seeking := true; ct:=0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false, _, _; end if;
       z,a,A := FC(2,2);
       if not z then return false,_,_; end if;
       z,b,B := FC(4,4);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C;  
         if CO(a*b) eq 7 and CO(a*b^2) eq 5 then
            if O(a*b^2) eq 20 then a *:= (a*b^2)^5; A *:= (A*B^2)^5; end if;
            if O(a*b) eq 84 then b *:= (a*b)^21; B *:= (A*B)^21; end if;
            if O(a*b^2) eq 10 then a *:= (a*b^2)^5; A *:= (A*B^2)^5; end if;
            if O(a*b) eq 42 then b *:= (a*b)^21; B *:= (A*B)^21; end if;
            if O(a*b) eq 21 and O(a*b^2) eq 5 and O(a*b*a*b*a*b^3) eq 10 then
              seeking:=false; break;
            end if;
         end if;
       end for;
     end while;
   when "U42":
     z,a,A := F(12); //to power into 2A
     if not z then return false,_,_; end if;
     o := O(a) div 2; a:=a^o; A:=A^o;
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 9 or ct eq nt;
   when "2U42":
     z,a,A := FC(12,12); //to power into 2A
     if not z then return false,_,_; end if;
     o := O(a) div 2; a:=a^o; A:=A^o;
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 9 or ct eq nt;
   when "Sz8":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     seeking := true; ct:=0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false, _, _; end if;
       z,b,B := F(4);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C;
         if O(a*b) eq 5 and O(a*b^2) eq 7 and O(a*b*a*b^3*a*b^2) eq 7 then
            seeking := false; break;
         end if;
       end for;
     end while;
   when "2Sz8":
     z,a,A := FC(2,2);
     if not z then return false,_,_; end if;
     o := O(a) div 2; a:=a^o; A:=A^o;
     seeking := true; ct:=0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false, _, _; end if;
       z,b,B := FC(4,4);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C;
         if O(a*b) eq 5 and O(a*b^2) eq 7 and O(a*b*a*b^3*a*b^2) eq 7 then
            seeking := false; break;
         end if;
       end for;
     end while;
   when "L232": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 31 and O(a*b*a*b^2) eq 11 or ct eq nt;
   when "L249":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 25 and O(a*b*a*b^2) eq 24 or ct eq nt;
   when "2L249":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 25 and O(a*b*a*b^2) eq 48 or ct eq nt;
   when "U34":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 13 or ct eq nt;
   when "M12":
     z,a,A := F(4); //squares into 2B
     if not z then return false,_,_; end if;
     a:=a^2; A:=A^2;
     //elements b in Class 3B are products of two elements in Class 2A
     z,bb,BB := F(10); //fifth power in 2A
     if not z then return false,_,_; end if;
     bb:=bb^5; BB:=BB^5;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(b*bb) eq 3;
     if ct eq nt then return false,_,_; end if;
     b := b*bb; B := B*BB;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 11;
   when "2M12":
     z,a,A := FC(4,4); //squares into 2B
     if not z then return false,_,_; end if;
     a:=a^2; A:=A^2;
     //elements b in Class 3B are products of two elements in Class 2A
     z,bb,BB := FC(10,20); //fifth power in 2A
     if not z then return false,_,_; end if;
     bb:=bb^5; BB:=BB^5;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until CO(b*bb) eq 3 and O(b*bb) eq 6;
     if ct eq nt then return false,_,_; end if;
     b := b*bb; B := B*BB;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 11;
   when "U35":
     z,a,A := F(3);
     if not z then return false,_,_; end if;
     z,b,B := F(10); //to power into 5A
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "3U35":
     z,a,A := FC(3,3); 
     if not z then return false,_,_; end if;
     z,b,B := F(10); //to power into 5A
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "J1":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 and Order(a*b*a*b^2) eq 19 or ct eq nt;
   when "A9":
     z,a,A := F(12); //to power into 3A
     a:=a^4; A:=A^4;
     if not z then return false,_,_; end if;
     z,b,B := F(7);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 9 or ct eq nt;
   when "2A9":
     z,a,A := FC(12,24); //to power into 3A
     a:=a^4; A:=A^4;
     if not z then return false,_,_; end if;
     z,b,B := F(7);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until CO(a*b) eq 9 or ct eq nt;
   when "L264":
//Not in ATLAS database, so I have chosen my own standard generators
//standard generators for PSL(2,64), a order 2, b order 3, order(ab)=12,
//order [a,b]=13.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 63 and Order((a,b)) eq 13 or ct eq nt;
   when "L281":
//Not in ATLAS database, so I have chosen my own standard generators
//standard generators for PSL(2,81), a order 2, b order 3, order(ab)=20.
//standard generators for SL(2,81), a order 4, b order 3, order(ab)=40,
//order(ababab^-1) = 41.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 20 or ct eq nt;
   when "2L281":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 40 and Order(a*b*a*b*a*b^-1) eq 41 or ct eq nt;
   when "L35":
     z,a,A := F(3);
     if not z then return false,_,_; end if;
     z,b,B := F(10); //to power into 5A
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 20 or ct eq nt;
   when "M22":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(8); //to power into 4A
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 11 and O(a*b*a*b^2) eq 11 or ct eq nt;
   when "2M22":
     z,a,A := F(8); //to power into class +2A
     if not z then return false,_,_; end if;
     a:=a^4; A:=A^4;
     z,b,B := F(8); //to power into 4A
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     //we need to multiply this by the central element, which we
     //get as a power of an element of order 10,14 or 22
     ct:=0; repeat ct+:=1;  c,C := R(); until O(c) in {10,14,22} or ct eq nt;
     if ct eq nt then return false,_,_; end if;
     b := b * c^(Order(c) div 2); B := B * C^(Order(c) div 2);
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 11 and O(a*b*a*b^2) eq 11 or ct eq nt;
   when "3M22":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(8); //to power into 4A
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until CO(a*b) eq 11 and CO(a*b*a*b^2) eq 11 or ct eq nt;
   when "6M22":
     z,a,A := F(8); //to power into class +2A
     if not z then return false,_,_; end if;
     a:=a^4; A:=A^4;
     z,b,B := F(8); //to power into 4A
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     //we need to multiply this by the central element, which we
     //get as a power of an element of order 10,14,22,30,42,66
     ct:=0; repeat ct+:=1;  c,C := R();
            until O(c) in {10,14,22,30,42,66} or ct eq nt;
     if ct eq nt then return false,_,_; end if;
     b := b * c^(O(c) div 2); B := B * C^(O(c) div 2);
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until CO(a*b) eq 11 and CO(a*b*a*b^2) eq 11 or ct eq nt;
   when "4M22":
     z,a,A := FC(2,2);
     if not z then return false,_,_; end if;
     z,b,B := FC(8,8);
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     //we need to multiply this by a central element of order 4, which we
     //get as a power of an element of order 20,48,44
     ct:=0; repeat ct+:=1;  c,C := R();
            until O(c) in {20,28,44} or ct eq nt;
     if ct eq nt then return false,_,_; end if;
     b := b * c^(O(c) div 4); B := B * C^(O(c) div 4);
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 11 and O(a*b*a*b^2) eq 11 or ct eq nt;
   when "12M22":
     z,a,A := FC(2,2);
     if not z then return false,_,_; end if;
     z,b,B := FC(8,8);
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     //we need to multiply this by a central element of order 4, which we
     //get as a power of an element of order 20,28,44, 60,84,132
     ct:=0; repeat ct+:=1;  c,C := R();
            until O(c) in {20,28,44,60,84,132} or ct eq nt;
     if ct eq nt then return false,_,_; end if;
     b := b * c^(O(c) div 4); B := B * C^(O(c) div 4);
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 33 and O(a*b*a*b^2) eq 33 or ct eq nt;
   when "J2":
     //2B and 3B elements are hard to find. We use a 2A element to
     //test them
     z,c,C := F(4);
     if not z then return false,_,_; end if;
     c := c^(O(c) div 2); //c in 2A
     seeking := true; ct2 := 0;
     while seeking do
       ct2 +:= 1;
       if ct2 eq nt then return false,_,_; end if;
       ct:=0; repeat ct+:=1; a,A := R(); until O(a) in {2,6,10} or ct eq nt;
       if ct eq nt then return false,_,_; end if;
       o := O(a); a := a^(o div 2); A := A^(o div 2);
       if not O(a*c) in {1,2,3,4,5} then //a is in 2B
          seeking:=false; break;
       end if;
     end while;
     seeking := true; ct2 := 0;
     while seeking do
       ct2 +:= 1;
       if ct2 eq nt then return false,_,_; end if;
       ct:=0; repeat ct+:=1; b,B := R(); until O(b) in {3,6} or ct eq nt;
       if ct eq nt then return false,_,_; end if;
       o := O(b); b := b^(o div 3); B := B^(o div 3);
       if not O(b*c) in {6,12} then //b is in 3B
          seeking:=false; break;
       end if;
     end while;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 and O(a*b*a*b^2) eq 12 or ct eq nt;
   when "2J2":
     //projective 3B elements are hard to find. We use a 2A element to
     //test them
     z,c,C := FC(2,2); //c in 2A
     if not z then return false,_,_; end if;
     z,a,A := FC(2,4); //c in 2B
     if not z then return false,_,_; end if;
     seeking := true; ct2 := 0;
     while seeking do
       ct2 +:= 1;
       if ct2 eq nt then return false,_,_; end if;
       ct:=0; repeat ct+:=1; b,B := R(); until CO(b) in {3,6} or ct eq nt;
       if ct eq nt then return false,_,_; end if;
       o := O(b); b := b^(o div 3); B := B^(o div 3);
       if not CO(b*c) in {6,12} then //b is in 3B
          seeking:=false; break;
       end if;
     end while;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 and CO(a*b*a*b^2) eq 12 or ct eq nt;
   when "L2121":
//Not in ATLAS database, so I have chosen my own standard generators
//Standard generators of L2(121) are a and b where a has order 2, b has order 3
//ab has order 15 and abab^-1 has order 61.
//Standard generators of the double cover 2.L2(121) = SL2(121) are preimages
//A and B where B has order 3,  and AB has order 15.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 15 and O(a*b*a*b^2) eq 61 or ct eq nt;
   when "2L2121":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 15 and CO(a*b*a*b^2) eq 61 or ct eq nt;
   when "L2125":
//Not in ATLAS database, so I have chosen my own standard generators
//standard generators for PSL(2,121), a order 2, b order 3, order(ab)=63
//and Order(abab^2) = 7.
//standard generators for SL(2,121), preimages of a,b,
//a order 4, b order 3, order(ab)=63,
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 63 and O(a*b*a*b^2) eq 7 or ct eq nt;
   when "2L2125":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 63 and CO(a*b*a*b^2) eq 7 or ct eq nt;
   when "S44":
     z,a,A := F(10); //to power into 2A/B
     if not z then return false,_,_; end if;
     a := a^5; A := A^5;
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,b,B := Fb(5); //want it in Class 5E but cannot test this directly
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i in [1..20] do
          c,C := R(); b := bb^c; B := BB^C;
          if O(a*b) eq 17 and O(a*b*a*b^2) eq 15 then
             seeking := false; break;
          end if;
       end for;
     end while;
   when "S62":
     z,a,A := F(10); //to power into 2A
     if not z then return false,_,_; end if;
     a := a^5; A := A^5;
     z,b,B := F(7);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 9 or ct eq nt;
   when "2S62":
     z,a,A := FC(10,20); //to power into 2A
     if not z then return false,_,_; end if;
     a := a^5; A := A^5;
     z,b,B := F(7);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 9 or ct eq nt;
   when "A10":
     z,a,A := Fb(15); //to power into 3A
     a:=a^5; A:=A^5;
     if not z then return false,_,_; end if;
     z,b,B := F(9);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 8 and O(a*b^2) eq 12 or ct eq nt;
   when "2A10":
     z,a,A := FC(15,15); //to power into 3A
     a:=a^5; A:=A^5;
     if not z then return false,_,_; end if;
     z,b,B := F(9);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C; 
            until CO(a*b) eq 8 and CO(a*b^2) eq 12 or ct eq nt;
   when "L37":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 19 and O(a*b*a*b^2) eq 6 or ct eq nt;
   when "3L37":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := FC(3,3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 19 and CO(a*b*a*b^2) eq 6 or ct eq nt;
   when "L2128":
//Not in ATLAS database, so I have chosen my own standard generators
//standard generators for PSL(2,128), a order 2, b order 3, order(ab)=43,
//order [a,b]=129.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 43 and Order((a,b)) eq 129 or ct eq nt;
   when "L2169":
//Not in ATLAS database, so I have chosen my own standard generators
//Standard generators of L2(169) are a and b where a has order 2, b has order 3
//ab has order 85 and abab^-1 has order 5.
//Standard generators of the double cover 2.L2(169) = SL2(169) are preimages
//A and B where B has order 3,  and AB has order 85.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 85 and O(a*b*a*b^2) eq 5 or ct eq nt;
   when "2L2169":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 85 and CO(a*b*a*b^2) eq 5 or ct eq nt;
   when "U43":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := Fb(12); //to power into 6A
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 and O((a*b)^3*(b*a)^2*b^2) eq 5 or ct eq nt;
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := FC(12,12); //to power into 6A
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 and O((a*b)^3*(b*a)^2*b^2) eq 5 or ct eq nt;
   when "2U43":
     //not in ATLAS - standard generators are preimages A,B of a,b where
     //ABAB^2 has order 7 and AB^3 has order 5
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,a,A := FC(2,2);
       if not z then return false,_,_; end if;
       z,b,B := FC(12,12); //to power into 6A
       if not z then return false,_,_; end if;
       b:=b^2; B:=B^2;
       bb := b; BB := B;
       for i in [1..20] do
          c,C := R(); b := bb^c; B := BB^C;
          if CO(a*b) eq 7 and CO((a*b)^3*(b*a)^2*b^2) eq 5
            and O(a*b*a*b^2) eq 7 and O(a*b^3) eq 5 then
             seeking := false; break;
          end if;
       end for;
     end while;
   when "3aU43":
     //not in ATLAS standard generators are A,B, preimages of a and b where
     //A,B,AB,ABABABBABABB,ABBABBBAB^-1 have orders 2, 6, 7, 5 and 5
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,a,A := F(2);
       if not z then return false,_,_; end if;
       z,b,B := FC(12,12); //to power into 6A
       if not z then return false,_,_; end if;
       b:=b^2; B:=B^2;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C;
         if O(a*b) eq 7 and O((a*b)^3*(b*a)^2*b^2) eq 5
              and O(a*b^2*a*b^3*a*b^-1) eq 5 then
            seeking := false; break;
         end if;
       end for;
     end while;
   when "3bU43":
     //not in ATLAS standard generators are A,B, preimages of a and b where
     //A,B,AB,ABABABBABABB,ABABBB have orders 2, 6, 7, 5 and 7
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,a,A := F(2);
       if not z then return false,_,_; end if;
       z,b,B := FC(12,12); //to power into 6A
       if not z then return false,_,_; end if;
       b:=b^2; B:=B^2;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C;
         if O(a*b) eq 7 and O((a*b)^3*(b*a)^2*b^2) eq 5
              and O(a*b*a*b^3) eq 7 then
            seeking := false; break;
         end if;
       end for;
     end while;
   when "4U43":
     //not in ATLAS - standard generators are preimages A,B of a,b where
     //ABAB^2 has order 7 and AB^3 has order 5
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,a,A := FC(2,2);
       if not z then return false,_,_; end if;
       z,b,B := FC(12,12); //to power into 6A
       if not z then return false,_,_; end if;
       b:=b^2; B:=B^2;
       bb := b; BB := B;
       for i in [1..20] do
          c,C := R(); b := bb^c; B := BB^C;
          if CO(a*b) eq 7 and CO((a*b)^3*(b*a)^2*b^2) eq 5
            and O(a*b*a*b^2) eq 7 and O(a*b^3) eq 5 then
             seeking := false; break;
          end if;
       end for;
     end while;
   when "6aU43":
     //not in ATLAS standard generators are A,B, preimages of a and b where
     //A,B,AB,ABABABBABABB,ABBABBBAB^-1 have orders 2, 6, 7, 5 and 10
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,a,A := F(2);
       if not z then return false,_,_; end if;
       z,b,B := FC(12,12); //to power into 6A
       if not z then return false,_,_; end if;
       b:=b^2; B:=B^2;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C;
         if O(a*b) eq 7 and O((a*b)^3*(b*a)^2*b^2) eq 5
              and O(a*b^2*a*b^3*a*b^-1) eq 10 then
            seeking := false; break;
         end if;
       end for;
     end while;
   when "6bU43":
     //not in ATLAS standard generators are A,B, preimages of a and b where
     //A,B,AB,ABABABBABABB,ABABBB have orders 2, 6, 14, 10 and 14
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,a,A := F(2);
       if not z then return false,_,_; end if;
       z,b,B := FC(12,12); //to power into 6A
       if not z then return false,_,_; end if;
       b:=b^2; B:=B^2;
       bb := b; BB := B;
       for i in [1..20] do
         c,C := R(); b := bb^c; B := BB^C;
         if O(a*b) eq 14 and O((a*b)^3*(b*a)^2*b^2) eq 10
              and O(a*b*a*b^3) eq 14 then
            seeking := false; break;
         end if;
       end for;
     end while;
   when "12aU43":
     //not in ATLAS standard generators are A,B, preimages of a and b where
     //A,B,ABABB,ABABABBABABB,ABBABBBAB^-1 have orders 2, 6, 7, 20 and 5
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,a,A := F(2);
       if not z then return false,_,_; end if;
       z,b,B := FC(12,12); //to power into 6A
       if not z then return false,_,_; end if;
       b:=b^2; B:=B^2;
       bb := b; BB := B;
       for i in [1..30] do
         c,C := R(); b := bb^c; B := BB^C;
         if  O(a*b*a*b^2) eq 7 and O((a*b)^3*(b*a)^2*b^2) eq 20
              and O(a*b^2*a*b^3*a*b^-1) eq 5 then
            seeking := false; break;
         end if;
       end for;
     end while;
   when "G23":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := Fb(9); //to power into 3C
     if not z then return false,_,_; end if;
     b := b^3; B := B^3;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 13 or ct eq nt;
   when "3G23":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := FC(9,9); //to power into 3C
     if not z then return false,_,_; end if;
     b := b^3; B := B^3;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 13 or ct eq nt;
   when "L2243":
//Not in ATLAS database, so I have chosen my own standard generators
//Standard generators of L2(243) are a and b where a has order 2, b has order 3
//ab has order 61 and abab^-1 has order 11.
//Standard generators of the double cover 2.L2(243) = SL2(243) are preimages
//A and B where B has order 3,  and AB has order 61.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 61 and O(a*b*a*b^2) eq 11 or ct eq nt;
   when "2L2243":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 61 and CO(a*b*a*b^2) eq 11 or ct eq nt;
   when "S45":
     //first get a in Class 2B
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,a,A := F(2);
       if not z then return false,_,_; end if;
       z,b := F(3);
       if not z then return false,_,_; end if;
       for i in [1..20] do
         c := R();
         if O(a*b^c) in {3,5,13,15,20} then
            seeking := false; break; //a in Class 2B
         end if;
       end for;
     end while;
     //now get b in Class 3B
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,b,B := F(3);
       if not z then return false,_,_; end if;
       for i  in [1..20] do
         c := R();
         if O(a^c*b*a^c*b^2) in {1,30} then
            seeking := false; break; //b in Class 3B
         end if;
       end for;
     end while;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 13 and O(a*b*a*b^2) eq 12  or ct eq nt;
   when "2S45":
     z,a,A := FC(2,4); //inverse image of class 2B
     if not z then return false,_,_; end if;
     //now get b in Class 3B
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,b,B := FC(3,3);
       if not z then return false,_,_; end if;
       for i  in [1..20] do
         c := R();
         if CO(a^c*b*a^c*b^2) in {1,30} then
            seeking := false; break; //b in Class 3B
         end if;
       end for;
     end while;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 13 and CO(a*b*a*b^2) eq 12  or ct eq nt;
   when "U38":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,b,B := F(3); //must get in in 3C
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i  in [1..20] do
         c,C := R(); b := bb^c; B := BB^C;
         if CO(a*b) eq 19 then
            seeking := false; break; //b in Class 3B
         end if;
       end for;
     end while;
   when "3U38":
     z,a,A := FC(2,2);
     if not z then return false,_,_; end if;
     seeking := true; ct := 0;
     while seeking do
       ct +:= 1;
       if ct eq nt then return false,_,_; end if;
       z,b,B := FC(3,3); //must get in in 3C
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       for i  in [1..20] do
         c,C := R(); b := bb^c; B := BB^C;
         if O(a*b) eq 19 then
            seeking := false; break; //b in Class 3B
         end if;
       end for;
     end while;
   when "U37": 
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 43 and O(a*b*a*b^2) eq 4 or ct eq nt;
   when "L43":
     z,a,A := F(10); //to power into 2A
     if not z then return false,_,_; end if;
     o:=O(a); a:=a^(o div 2); A:=A^(o div 2);
     z,b,B := Fb(8); //to power into 4B
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 13 and O(a*b*a*b^2) eq 13 or ct eq nt;
   when "2L43":
     z,a,A := FC(10,20); //to power into 2A
     if not z then return false,_,_; end if;
     o:=CO(a); a:=a^(o div 2); A:=A^(o div 2);
     z,b,B := FC(8,8); //to power into 4B
     if not z then return false,_,_; end if;
     b:=b^2; B:=B^2;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 13 and O(a*b*a*b^2) eq 13 or ct eq nt;
   when "L52":
     z,a,A := F(8); //to power into 2A
     if not z then return false,_,_; end if;
     o:=O(a); a:=a^(o div 2); A:=A^(o div 2);
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 5 or ct eq nt;
   when "M23":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(4);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 23 and O(a*b*a*b*a*b*a*b*b*a*b*a*b*b*a*b*b) eq 8
                                                              or ct eq nt;
   when "L2289":
//Not in ATLAS database, so I have chosen my own standard generators
//Standard generators of L2(289) are a and b where a has order 2, b has order 3
//ab has order 29 and abab^-1 has order 72.
//Standard generators of the double cover 2.L2(289) = SL2(289) are preimages
//A and B where B has order 3,  and AB has order 29.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 29 and O(a*b*a*b^2) eq 72 or ct eq nt;
   when "2L2289":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 29 and CO(a*b*a*b^2) eq 72 or ct eq nt;
   when "L38":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 21 or ct eq nt;
   when "U52":
     z,a,A := F(8); //to power into 2A
     if not z then return false,_,_; end if;
     o:=O(a); a:=a^(o div 2); A:=A^(o div 2);
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 11 or ct eq nt;
   when "L2256":
//Not in ATLAS database, so I have chosen my own standard generators
//standard generators for PSL(2,256), a order 2, b order 3, order(ab)=51,
//order [a,b]=85.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 51 and Order((a,b)) eq 85 or ct eq nt;
   when "TF42":
     z,a,A := F(10); //to power into 2A
     if not z then return false,_,_; end if;
     o:=O(a); a:=a^(o div 2); A:=A^(o div 2);
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 13 or ct eq nt;
   when "A11":
     z,a,A := F(21); //to power into 3A
     if not z then return false,_,_; end if;
     o:=O(a); a:=a^(o div 3); A:=A^(o div 3);
     z,b,B := F(9);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 11 or ct eq nt;
   when "2A11":
     z,a,A := FC(21,21); //to power into 3A
     a:=a^7; A:=A^7;
     if not z then return false,_,_; end if;
     z,b,B := F(9);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C; 
            until CO(a*b) eq 11 or ct eq nt;
   when "L2343":
//Not in ATLAS database, so I have chosen my own standard generators
//Standard generators of L2(343) are a and b where a has order 2, b has order 3
//ab has order 19 and abab^-1 has order 171.
//Standard generators of the double cover 2.L2(343) = SL2(343) are preimages
//A and B where B has order 3,  and AB has order 19.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 19 and O(a*b*a*b^2) eq 171 or ct eq nt;
   when "2L2343":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 19 and CO(a*b*a*b^2) eq 72 or ct eq nt;
   when "Sz32":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     //not all elements of order 4 will work!
     ct:=0;
     repeat ct+:=1;
       z,b,B := F(4);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       sct:=0;
       repeat sct+:=1;
          if sct eq 101 then break; end if;
          c,C := R(); b := bb^c; B := BB^C;
       until O(a*b) eq 5 and O(a*b^2) eq 25 and O(a*b*a*b^2*a*b^3) eq 25;
     until sct le 100 or ct eq nt;
   when "L2361":
//Not in ATLAS database, so I have chosen my own standard generators
//Standard generators of L2(361) are a and b where a has order 2, b has order 3
//ab has order 45 and abab^-1 has order 36.
//Standard generators of the double cover 2.L2(361) = SL2(361) are preimages
//A and B where B has order 3,  and AB has order 45.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 45 and O(a*b*a*b^2) eq 36 or ct eq nt;
   when "2L2361":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 45 and CO(a*b*a*b^2) eq 36 or ct eq nt;
   when "L39": //not in ATLAS, so must find standard generators!
     //a of order2, b order 5, ab order 3 works fine.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 3 or ct eq nt;
   when "U39":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(6);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 15 or ct eq nt;
   when "HS":
     z,a,A := F(4); //to power into 2A
     if not z then return false,_,_; end if;
     o:=O(a); a:=a^(o div 2); A:=A^(o div 2);
     z,b,B := F(20); //to power into 5A
     o:=O(b); b:=b^(o div 5); B:=B^(o div 5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 11 or ct eq nt;
   when "2HS":
     z,a,A := FC(4,4); //to power into 2A
     if not z then return false,_,_; end if;
     a:=a^2; A:=A^2;
     z,b,B := FC(20,20); //to power into 5A
     if not z then return false,_,_; end if;
     b:=b^4; B:=B^4;
     z,d,D := FC(3,6); //just to get the central element
     if not z then return false,_,_; end if;
     b:=b*d^3; B:=B*D^3;
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0;
     oa:=a; oA:=A;
     repeat
       a:=oa; A:=oA; sct:=0;
       while sct le 20 do sct+:=1;
           c,C := R(); b := bb^c; B := BB^C; 
           if  O(a*b) eq 11 then break; end if;
       end while;
       if sct gt 20 then
          a := a*d^3; A:=A*D^3; sct:=0;
          while sct le 20 do sct+:=1;
            c,C := R(); b := bb^c; B := BB^C; 
            if  O(a*b) eq 11 then break; end if;
          end while;
       end if;
     until sct le 20 or ct eq nt;
   when "J3":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(6); //to power into 3A
     if not z then return false,_,_; end if;
     o:=O(b); b:=b^(o div 3); B:=B^(o div 3);
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 19 and O(a*b*a*b^2) eq 9 or ct eq nt;
   when "3J3":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     ct := 0;
     repeat ct+:=1;
       z,b,B := FC(6,6);
       if not z then return false,_,_; end if;
       o:=O(b); b:=b^(o div 3); B:=B^(o div 3);
       bb := b; BB := B;
       sct:=0;
       while sct lt 20 do
         c,C := R(); b := bb^c; B := BB^C; 
         if CO(a*b) eq 19 and CO(a*b*a*b^2) eq 9 and O(a*b*a*b*a*b^2) eq 17 then
            break;
         end if;
         sct := sct+1;
       end while;
     until sct lt 20 or ct eq nt;
   when "U311":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 37 and O(a*b*a*b^2) eq 4 or ct eq nt;
   when "3U311":
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := FC(3,3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 37 and CO(a*b*a*b^2) eq 4 or ct eq nt;
   when "L2529":
//Not in ATLAS database, so I have chosen my own standard generators
//Standard generators of L2(529) are a and b where a has order 2, b has order 3
//ab has order 53 and abab^-1 has order 264.
//Standard generators of the double cover 2.L2(529) = SL2(529) are preimages
//A and B where B has order 3,  and AB has order 53.
     z,a,A := F(2);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 53 and O(a*b*a*b^2) eq 264 or ct eq nt;
   when "2L2529":
     z,a,A := F(4);
     if not z then return false,_,_; end if;
     z,b,B := F(3);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 53 and CO(a*b*a*b^2) eq 264 or ct eq nt;
   when "S47":
     z,a,A := F(42); //to power into 2A
     if not z then return false,_,_; end if;
     o:=O(a); a:=a^(o div 2); A:=A^(o div 2);
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until O(a*b) eq 7 or ct eq nt;
   when "2S47":
     z,a,A := FC(42,42); //to power into 3A
     if not z then return false,_,_; end if;
     o:=O(a); a:=a^(o div 2); A:=A^(o div 2);
     z,b,B := F(5);
     if not z then return false,_,_; end if;
     bb := b; BB := B;
     ct:=0; repeat ct+:=1;  c,C := R(); b := bb^c; B := BB^C; 
            until O(a*b) eq 7 or ct eq nt;
   when "O8p2":
     ct:=0;
     repeat ct+:=1;
       z,a,A := Fb(6);
       if not z then return false,_,_; end if;
       a := a^3; A := A^3;
       z,b,B := F(5);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       sct:=0; repeat sct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until (O(a*b) eq 12 and O(a*b^2) eq 15 and
                   O(a*b*a*b*a*b*a*b*b*a*b*a*b*b*a*b*b) eq 8) or sct eq 120;
     until sct lt 50 or ct eq nt;
   when "2O8p2":
     ct:=0;
     repeat ct+:=1;
       repeat z,a,A := Fb(6); until CO(a) eq 6;
       a := a^3; A := A^3;
       z,b,B := F(5);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       sct:=0; repeat sct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until (CO(a*b) eq 12 and O(a*b^2) eq 15 and O(a*b*a*b^2) eq 15 and
                   CO(a*b*a*b*a*b*a*b*b*a*b*a*b*b*a*b*b) eq 8) or sct eq 120;
     until sct lt 50 or ct eq nt;
   when "2^2O8p2":
     ct:=0;
     repeat ct+:=1;
       repeat z,a,A := Fb(6); until CO(a) eq 6;
       a := a^3; A := A^3;
       z,b,B := F(5);
       if not z then return false,_,_; end if;
       bb := b; BB := B;
       sct:=0; repeat sct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until (CO(a*b) eq 12 and O(a*b^2) eq 15 and O(a*b*a*b^2) eq 30 and
                   CO(a*b*a*b*a*b*a*b*b*a*b*a*b*b*a*b*b) eq 8) or sct eq 120;
     until sct lt 50 or ct eq nt;
   when "O8m2":
     ct:=0;
     repeat ct+:=1;
       z,a,A := Fb(6);
       if not z then return false,_,_; end if;
       a := a^3; A := A^3;
       z,b,B := F(9); //to cube into 3C
       if not z then return false,_,_; end if;
       b := b^3; B := B^3;
       bb := b; BB := B;
       sct:=0; repeat sct+:=1;  c,C := R(); b := bb^c; B := BB^C;
            until (O(a*b) eq 17 and O(a*b*a*b^2) eq 17 and
                   O(a*b*a*b*a*b*b) eq 30) or sct eq 120;
     until sct lt 50 or ct eq nt;
   else gpname,
     "is not a valid identifier for a small quasisimple or sporadic group";
   return false,_,_;
  end case;
  if ct eq nt then return false, _, _; end if;
  return true, [a,b], [A,B];
end function;

CheckSG := function(G,a,b,aa,bb)
//for checking standard generators applied to G.
//Do a quick check to see if the map a->aa, b->bb appears to be a homomorphism
  F:=FreeGroup(2);
  h1:=hom<F->G|a,b>; h2:=hom<F->G|aa,bb>;
  for i in [1..100] do
    g := Random(F,30,50); 
    if Order(h1(g)) ne Order(h2(g)) then
      "NOT A HOMOMORPHISM!";
      return false;
    end if;
  end for;
  return true;
end function;

CheckSGGrp := function(s : G:=Sym(1))
//Check standard generators code of ATLAS group s.
  if Type(G) eq GrpPerm and Degree(G) eq 1  then
    if #PermRepKeys(ATLASGroup(s)) gt 0 then
      G:=X^Random(Sym(Degree(X))) where X:=PermutationGroup(s,1);
    elif #MatRepKeys(ATLASGroup(s)) gt 0 then
      G:=X^Random(GL(Dimension(X),BaseRing(X))) where X:=MatrixGroup(s,1);
    else "No ATLAS representation of",s,"available";
      return false;
    end if;
  end if;
  gens := [G.i : i in [1..Ngens(G)]];
  isit := false;
  repeat  isit, l1, l2 :=StandardGens(G,s); if not isit then "ERROR";end if;until isit;
  a:=l1[1]; b:=l1[2];
  assert Evaluate(l2[1],gens) eq a and Evaluate(l2[2],gens) eq b;
  for i in [1..50] do
i;
    repeat  isit, l1, l2 := StandardGens(G,s);if not isit then "ERROR";end if; until isit;
    if isit and not CheckSG(G,a,b,l1[1],l1[2]) then
      "FATAL ERROR!";
      return false;
    end if;
  end for;
  return true;
end function;

StandardCopy := function(gpname)
//The standard copy of gpname + flag to indicate whether projective or not

    //deal with all of the L2p together
    fl, head, p := parse_L2p(gpname);
    if fl then
	if head eq "L2" then
	  return SL(2,p), true;
	elif head eq "2L2" then
	  return SL(2,p), false;
	end if;
    end if;

  case gpname:
   when "A5": 
     return Alt(5), false;
   when "2A5": 
     return SL(2,5), false;
   when "A6": 
     return Alt(6), false;
   when "2A6": 
     return SL(2,9), false;
   when "3A6": 
     return PermutationGroup("3A6",1), false;
   when "6A6": 
     return MatrixGroup("6A6",1), false;
   when "L28": 
     return SL(2,8), false;
   when "A7": 
     return Alt(7), false;
   when "2A7": 
     return MatrixGroup("2A7",1), false;
   when "3A7": 
     return PermutationGroup("3A7",1), false;
   when "6A7": 
     return MatrixGroup("6A7",1), false;
   when "L216": 
     return SL(2,16), false;
   when "U33": 
     return SU(3,3), false;
   when "L225":
     return SL(2,25), true;
   when "2L225":
     return SL(2,25), false;
   when "M11": 
     return PermutationGroup("M11",1), false;
   when "L227": 
     return SL(2,27), true;
   when "2L227": 
     return SL(2,27), false;
   when "A8": 
     return Alt(8), false;
   when "2A8": 
     return MatrixGroup("2A8",1), false;
   when "L34": 
     return SL(3,4), true;
   when "3L34": 
     return SL(3,4), false;
   when "2L34":
     return PermutationGroup("2L34",1), false;
   when "4aL34":
     return PermutationGroup("4aL34",1), false;
   when "4bL34":
     return PermutationGroup("4bL34",1), false;
   when "6L34":
     return PermutationGroup("6L34",1), false;
   when "12aL34":
     return PermutationGroup("12aL34",1), false;
   when "12bL34":
     return PermutationGroup("12bL34",1), false;
   when "U42":
     return SU(4,2), false;
   when "2U42":
     return Sp(4,3), false;
   when "Sz8":
     return Sz(8), false;
   when "2Sz8":
     return PermutationGroup("2Sz8",1), false;
   when "L232": 
     return SL(2,32), false;
   when "L249":
     return SL(2,49), true;
   when "2L249":
     return SL(2,49), false;
   when "U34":
     return SU(3,4), false;
   when "M12":
     return PermutationGroup("M12",1), false;
   when "2M12":
     return PermutationGroup("2M12",1), false;
   when "U35":
     return SU(3,5), true;
   when "3U35":
     return SU(3,5), false;
   when "J1":
     return PermutationGroup("J1",1), false;
   when "A9":
     return Alt(9), false;
   when "2A9":
     return MatrixGroup("2A9",1), false;
   when "L264":
     return SL(2,64), false;
   when "L281":
     return SL(2,81), true;
   when "2L281":
     return SL(2,81), false;
   when "L35":
     return SL(3,5), false;
   when "M22":
     return PermutationGroup("M22",1), false;
   when "2M22":
     return PermutationGroup("2M22",1), false;
   when "3M22":
     return PermutationGroup("3M22",1), false;
   when "6M22":
     return PermutationGroup("6M22",1), false;
   when "4M22":
     return PermutationGroup("4M22",1), false;
   when "12M22":
     return PermutationGroup("12M22",1), false;
   when "J2":
     return PermutationGroup("J2",1), false;
   when "2J2":
     return PermutationGroup("2J2",1), false;
   when "L2121":
     return SL(2,121), true;
   when "2L2121":
     return SL(2,121), false;
   when "L2125":
     return SL(2,125), true;
   when "2L2125":
     return SL(2,125), false;
   when "S44":
     return Sp(4,4), false;
   when "S62":
     return Sp(6,2), false;
   when "2S62":
     return PermutationGroup("2S62",1), false;
   when "A10":
     return Alt(10), false;
   when "2A10":
     return MatrixGroup("2A10",1);
   when "L37":
     return SL(3,7), true;
   when "3L37":
     return SL(3,7), false;
   when "L2128":
     return SL(2,128), false;
   when "L2169":
     return SL(2,169), true;
   when "2L2169":
     return SL(2,169), false;
   when "U43":
     return SU(4,3), true;
   when "2U43":
     return SU(4,3), true;
   when "4U43":
     return SU(4,3), false;
   when "G23":
     return G2(3), false;
   when "3G23":
     return PermutationGroup("2G23",1), false;
   when "L2243":
     return SL(2,243), true;
   when "2L2243":
     return SL(2,243), false;
   when "S45":
     return Sp(4,5), true;
   when "2S45":
     return Sp(4,5), false;
   when "U38":
     return SU(3,8), true;
   when "3U38":
     return SU(3,8), false;
   when "U37": 
     return SU(3,7), false;
   when "L43":
     return SL(4,3), true;
   when "2L43":
     return SL(4,3), false;
   when "L52":
     return SL(5,2), false;
   when "M23":
     return PermutationGroup("M23",1), false;
   when "L2289":
     return SL(2,289), true;
   when "2L2289":
     return SL(2,289), false;
   when "L38":
     return SL(3,8), false;
   when "U52":
     return SU(5,2), false;
   when "L2256":
     return SL(2,256), false;
   when "TF42":
     return MatrixGroup("TF42",1), false;
   when "A11":
     return Alt(11), false;
   when "2A11":
     return MatrixGroup("2A11",1);
   when "L2343":
     return SL(2,343), true;
   when "2L2343":
     return SL(2,343), false;
   when "Sz32":
     return Sz(32), false;
   when "L2361":
     return SL(2,361), true;
   when "2L2361":
     return SL(2,361), false;
   when "L39":
     return SL(3,9), false;
   when "U39":
     return SU(3,9), false;
   when "HS":
     return PermutationGroup("HS",1), false;
   when "2HS":
     return PermutationGroup("2HS",1), false;
   when "J3":
     return PermutationGroup("J3",1), false;
   when "3J3":
     return MatrixGroup("J3",1), false;
   when "U311":
     return SU(3,11), true;
   when "3U311":
     return SU(3,11), false;
   when "L2529":
     return SL(2,529), true;
   when "2L2529":
     return SL(2,529), false;
   when "S47":
     return Sp(4,7), true;
   when "2S47":
     return Sp(4,7), false;
   when "O8p2":
     return OmegaPlus(8,2), false;
   when "O8m2":
     return OmegaMinus(8,2), false;
   when "M24":
     return  PermutationGroup("M24",1), false;
   when "McL":
     return  PermutationGroup("McL",1), false;
   when "He":
     return  PermutationGroup("He",1), false;
   when "Ru":
     return  PermutationGroup("Ru",1), false;
   when "Suz":
     return  PermutationGroup("Suz",1), false;
   when "ON":
     return  MatrixGroup("ON",1), false;
   when "Co3":
     return  PermutationGroup("Co3",1), false;
   when "Co2":
     return  PermutationGroup("Co2",1), false;
   when "F22":
     return  PermutationGroup("F22",1), false;
   when "Fi22":
     return  PermutationGroup("F22",1), false;
   else
     error gpname, "is not a valid identifier, or no standard copy available";
   end case;
end function;

intrinsic IsomorphismToStandardCopy (G:: Grp, str :: MonStgElt :
          AutomorphismGroup := false,
          Projective := false,
          MaxTries := 2000) -> BoolElt, Map
{Use standard generators to find a (possibly projective) isomorphism from
sporadic or small quasisimple group G to its standard copy}
    local flag, A, GG, proj, AA;
    flag, A := StandardGenerators(G, str :
                        AutomorphismGroup := AutomorphismGroup,
                        Projective := Projective,
                        MaxTries := MaxTries);
    if not flag then return false, _; end if;
    GG, proj := StandardCopy(str);
    flag, AA := StandardGenerators(GG, str :
                        Projective := proj,
                        MaxTries := MaxTries);
    if not flag then return false, _; end if;
    return true, hom< sub<G|A[1],A[2]> -> GG | AA[1], AA[2] >;
end intrinsic;

