freeze;

RingIsZ := func<G | BaseRing(G) cmpeq IntegerRing()>;
RingIsQ := func<G | BaseRing(G) cmpeq RationalField()>;



intrinsic HermiteNumber(L::Lat : Proof := true, Prune := [1.0: i in [1..Dimension(L)]]) -> FldReElt
{The Hermite invariant of L, i.e., N(L) / det(L)^(1/Dimension(L))}
   require Rank(L) ge 1: "Argument 1 must be non-zero";
   return Minimum(L: Proof:=Proof, Prune:=Prune) / Determinant(L) ^ (RealField()!1/Dimension(L));
end intrinsic;


intrinsic PackingRadius(L::Lat : Proof := true, Prune := [1.0: i in [1..Dimension(L)]]) -> FldReElt
{The packing radius of L (half the minimal distance between lattice points)}
    require Rank(L) ge 1: "Argument 1 must be non-zero";    
    return Sqrt(Minimum(L: Proof:=Proof, Prune:=Prune)) / 2;
end intrinsic;






intrinsic EnumerationCostArray(L::Lat: Prune := [1.0:i in [1..Dimension(L)]]) -> ModTupFldElt
{Heuristic evaluations of the sizes of all the layers of the tree to be visited during the lattice shortest vector enumeration}
local G, d, total, det, pi, length, minko, R, test, tmp;

  require Rank(L) ge 1: "Argument 1 must be non-zero";
  if Type(Prune) ne Infty then 
    require #Prune eq Rank(L): "The length of Prune must be the rank of argument 1";
    require forall{i: i in [1..Rank(L)] | Prune[i] ge 0.0 and Prune[i] le 1.0}:       "The elemements of Prune must be in [0.0, 1.0]"; 
  end if;  


  G:=GramMatrix (L);
  d:=NumberOfRows (G);
  R:=RealField(Ceiling(53+1.6*d):Bits:=true);
  RL := BaseRing(L);
  if Type(RL) ne FldRe then
    G := ChangeRing(G,R);
  else
    // Symmetrize G (*sigh*)                                                                              
    G := (G+Transpose(G))/2;
  end if;
  pi:=Pi(R);
  length := Sqrt(RealField()!(G[1][1]));
  T:=Cholesky(G);

  test := true;
  for j:=1 to d do test := test and (Prune[j] eq 1.0); end for;
  test := not(test);

  det := 1.0; for i:=1 to d do det := det * T[i][i]; end for;
  minko := (HermiteConstant(d))^(1/(2*d)) * det^(1/d);
  if (minko le length) then length:=minko; end if;

  total := VectorSpace(RealField(3), d)!0;
  det := 1.0; 
  for j:=d to 1 by -1 do
    det:=det*(length/T[j][j])^2; 
    if (test) 
      then tmp:= TruncatedHyperball ([Prune[j+i]: i in [0..d-j]]); 
      else tmp := 1.0;
    end if;
    total[j]:=tmp*Sqrt(det)*pi^((d-j+1)/2.0)/Gamma(1.0+(d-j+1)/2.0);
  end for;

  delete T; delete pi; delete G; delete R;
  return total;
end intrinsic;



intrinsic EnumerationCost(L::Lat: Prune := [1.0:i in [1..Dimension(L)]]) -> FldReElt
{A heuristic evaluation of the size of the tree to be visited during the lattice shortest vector enumeration}
local G, length, d, R, pi, T, det, minko, total, test, tmp;

  require Rank(L) ge 1: "Argument 1 must be non-zero";
  if Type(Prune) ne Infty then 
    require #Prune eq Rank(L): "The length of Prune must be the rank of argument 1";
    require forall{i: i in [1..Rank(L)] | Prune[i] ge 0.0 and Prune[i] le 1.0}:       "The elemements of Prune must be in [0.0, 1.0]"; 
  end if;  

  G:=GramMatrix (L);
  d:=NumberOfRows(G);
  length:=RealField()!G[1][1];
  R:=RealField(Ceiling(53+1.6*d):Bits:=true);
  RL := BaseRing(L);
  if Type(RL) ne FldRe then
    G := ChangeRing(G,R);
   else
     // Symmetrize G (*sigh*) 
   G := (G+Transpose(G))/2;
  end if;
  pi:=Pi(R);
  T:=Cholesky(G);


  test := true;
  for j:=1 to d do test := test and (Prune[j] eq 1.0); end for;
  test := not(test);

  det := 1.0; for i:=1 to d do det := det * T[i][i]; end for;
  minko := (HermiteConstant(d))^(1/(2*d)) * det^(1/d);
  if (minko^2 le length) 
     then length:=minko; 
     else length := Sqrt(length); 
  end if;

  total := 0.0;
  det := 1.0; 
  for j:=d to 1 by -1 do
    det:=det*(length/T[j][j])^2; 
    if (test) 
      then tmp:= TruncatedHyperball ([Prune[j+i]: i in [0..d-j]]); 
      else tmp := 1.0;
    end if;
    total:=total + tmp*Sqrt(det)*pi^((d-j+1)/2.0)/Gamma(1.0+(d-j+1)/2.0);
  end for;

  delete T;
  delete pi;
  delete (G);
  total := RealField(3)!total;
  delete R;
return total;
end intrinsic;



intrinsic EnumerationCost(L::Lat, u::FldReElt: Prune := [1.0:i in [1..Dimension(L)]]) -> FldReElt
{A heuristic evaluation of the size of the tree to be visited during the enumeration of vectors within the prescribed norm}
local G, d, total, det, pi, length, minko, R, test, tmp;

  require Rank(L) ge 1: "Argument 1 must be non-zero";
  require u ge 0.0: "Argument 2 must be non-negative";

  if Type(Prune) ne Infty then 
    require #Prune eq Rank(L): "The length of Prune must be the rank of argument 1";
    require forall{i: i in [1..Rank(L)] | Prune[i] ge 0.0 and Prune[i] le 1.0}:       "The elemements of Prune must be in [0.0, 1.0]"; 
  end if;  



  G:=GramMatrix (L);
  d:=NumberOfRows (G);
  R:=RealField(Ceiling(53+1.6*d):Bits:=true);
  RL := BaseRing(L);
  if Type(RL) ne FldRe then
    G := ChangeRing(G,R);
  else
    // Symmetrize G (*sigh*)
    G := (G+Transpose(G))/2;
  end if;
  pi:=Pi(R);
  length := Sqrt(u);
  T:=Cholesky(G);

  test := true;
  for j:=1 to d do test := test and (Prune[j] eq 1.0); end for;
  test := not(test);

  total := 0.0;
  det := 1.0; 
  for j:=d to 1 by -1 do
    det:=det*(length/T[j][j])^2;     
    if (test) 
      then tmp:= TruncatedHyperball ([Prune[j+i]: i in [0..d-j]]); 
      else tmp := 1.0;
    end if;
    total:=total + tmp*Sqrt(det)*pi^((d-j+1)/2.0)/Gamma(1.0+(d-j+1)/2.0);
  end for;

  total := RealField(3)!total;

  delete T; delete pi; delete G; delete R;
return total;
end intrinsic;





intrinsic EnumerationCostArray(L::Lat, u::FldReElt: Prune := [1.0:i in [1..Dimension(L)]]) -> ModTupFldElt
{Heuristic evaluations of the sizes of all the layers of the tree to be visited during the enumeration of vectors within the prescribed norm}
local G, d, total, det, pi, length, minko, R, test, tmp;

  require Rank(L) ge 1: "Argument 1 must be non-zero";
  require u ge 0.0: "Argument 2 must be non-negative";

  if Type(Prune) ne Infty then 
    require #Prune eq Rank(L): "The length of Prune must be the rank of argument 1";
    require forall{i: i in [1..Rank(L)] | Prune[i] ge 0.0 and Prune[i] le 1.0}:       "The elemements of Prune must be in [0.0, 1.0]"; 
  end if;  
  
  G:=GramMatrix (L);
  d:=NumberOfRows (G);
  R:=RealField(Ceiling(53+1.6*d):Bits:=true);
  RL := BaseRing(L);
  if Type(RL) ne FldRe then
    G := ChangeRing(G,R);
  else
     // Symmetrize G (*sigh*)
    G := (G+Transpose(G))/2;
  end if;

  pi:=Pi(R);
  length := Sqrt(u);
  T:=Cholesky(G);

  test := true;
  for j:=1 to d do test := test and (Prune[j] eq 1.0); end for;
  test := not(test);

  total := VectorSpace(RealField(3), d)!0;
  det := 1.0; 
  for j:=d to 1 by -1 do
    det:=det*(length/T[j][j])^2; 
    if (test) 
      then tmp:= TruncatedHyperball ([Prune[j+i]: i in [0..d-j]]); 
      else tmp := 1.0;
    end if;
    total[j]:=tmp*Sqrt(det)*pi^((d-j+1)/2.0)/Gamma(1.0+(d-j+1)/2.0);
  end for;

  delete T; delete pi; delete G; delete R;
return total;
end intrinsic;



intrinsic ReconstructLatticeBasis(C::ModMatRngElt, B::ModMatRngElt) -> ModMatRngElt
{Reconstruct a basis from a full-dimensional set of linearly independent lattice vectors}
local QQ, ZZ, U, T, tmp, gB, d, n;  

  require RingIsZ(C): "Base ring of argument 1 must be Z";
  require RingIsZ(B): "Base ring of argument 2 must be Z";
  require Dimension(Lattice(C)) eq Dimension(Lattice(B)): "Arguments 1 and 2 must span the same vector space";
  require  forall{i: i in [1..NumberOfRows(C)] | C[i] in Lattice(B)}: "Each row of Argument 1 must belong to the lattice spanned by Argument 2";


  QQ:=Rationals(); ZZ:=Integers();

  U:=ChangeRing(ChangeRing(C, QQ)*ChangeRing(B, QQ)^(-1), ZZ);
// C eq U * B;

  _, T := HermiteForm(Transpose(U));
// Transpose(H) eq U*Transpose(T);

  gB := ChangeRing(ChangeRing(Transpose(T), QQ)^(-1) * ChangeRing(B, QQ), ZZ);
// C eq Transpose(H) * gB;

  d:=NumberOfRows(gB); n:=NumberOfColumns(gB);

// size-reduction step
  for i:=2 to d do
    tmp := RMatrixSpace(ZZ, i, n+1)!0;
    for j:=1 to i do for k:=1 to n do tmp[j][k]:=gB[j][k]; end for; end for;
    tmp[i][n+1] := 2*Max([Norm(gB[j]): j in [1..i-1]]);
    tmp := LLL(tmp);
    for j:=1 to n do gB[i][j]:=tmp[i][j]; end for;
  end for;  

  delete (QQ); delete (ZZ); delete (U); delete (T); delete (tmp); 
  return gB;
end intrinsic;




intrinsic GaussReduce(B::MtrxSpcElt) -> MtrxSpcElt, AlgMatElt
{A Gauss-reduced basis of the lattice spanned by the rows of B};
  require Rank(B) eq 2: "Argument 1 must have rank 2";
  C, T:=HKZ(B);
  return C, T;
end intrinsic;



intrinsic GaussReduceGram(B::MtrxSpcElt) -> MtrxSpcElt, AlgMatElt
{A Gauss-reduced form equivalent to the Gram matrix B};
  require Rank(B) eq 2: "Argument 1 must have rank 2";
  C, T:=HKZGram(B);
  return C, T;
end intrinsic;


intrinsic GaussReduce(L::Lat) -> Lat, AlgMatElt
{Apply Gauss-reduction to L, returning a new lattice and the 
    transformation matrix T};
  require Rank(L) eq 2: "Argument 1 must have rank 2";
  F := GramMatrix(L);
  F, T := HKZGram(F);
  return LatticeWithBasis(
	MatrixRing(BaseRing(L),Rank(L)) !
           T*BasisMatrix(L), InnerProductMatrix(L) ), T;
end intrinsic;


intrinsic HKZ(L::Lat : Prune := [[1.0:j in [1..i]]:i in [1..Dimension(L)]], Proof:=true) -> Lat, AlgMatElt
{A lattice equal to L but with an HKZ-reduced basis, together the
corresponding transformation matrix}
  M, U := HKZGram(GramMatrix(L): Prune:=Prune, Proof:=Proof );
  test := true;
  for i:=1 to Dimension(L) do 
    for j:=1 to i do 
  test:= test and (Prune[i][j] eq 1.0); 
    end for; 
  end for;
  L2:= LatticeWithBasis(U*BasisMatrix(L), InnerProductMatrix(L));
  if (Proof and test) then L`Minimum:=M[1][1]; L2`Minimum:=M[1][1]; end if;
  return L2, U;
end intrinsic;

