freeze;

/*******************************************************************
Artin representations and Frobenius elements in Galois groups

implements ArtRep type and all related functionality

version 1: Tim Dokchitser, Sep 2008
  compute full splitting field: works well for Galois groups of size <70,
  possibly up to <150 or so.
version 2: Tim and Vladimir Dokchitser, Sep 2010
  implemented identification by cycle type, Serre's trick for A_n and
  general machinery from `Identifying conjugacy classes in Galois groups'
  to handle arbitrary Galois groups plus improved p-adic computations
  can usually handle Galois groups of size <10000 (acting on a small number
  of points) easily, and much larger special groups such as A_n and S_n.
*******************************************************************/

// pass precomputed local data at bad primes?
// Remove `//!' in the text that refer to bugs
// Use Method:=Artin in Dedekind zeta functions more often
//  (always? for all extensions except Galois ones?)
// Eventually replace ZptoZ by standard coercion (Claus fixed this)



declare verbose ArtRep,3;     // Verbose printing

// Attributes

declare attributes FldNum: artinrepdata;

declare type ArtRep;
declare attributes ArtRep: K, character, conductor;

artinrec:=recformat<K,f,G,Frob,p0,r,C,CC,ramified,badprimes,localdata,
                    cycs,disc,disc_unfactored,Inv,complexconjugation,
                    galoisdata>;

/*
local data at a prime p
in the format <rts, D, Frob, ramgps, classmap, prec>
  rts  = roots of f in its splitting field F over Q_p
  D    = local Galois group as a subgroup of S_n (via action on rts)
  Frob = Frobenius, an element of D
  ramgps = data for higher ramification groups, in the form of a list
              [<H1,i1>,<H2,i2>,...],
           where for H,i in each entry, H is the subgroup of DSn consisting
           of inertia automorphisms for which v=valuation(sigma(pi)-pi) >= i,
           and i runs over all v's that can occur like that;
           ramgps[1][1] is the inertia group
  classmap = indexing map ConjugacyClasses(DSn)->ConjugacyClasses(G)
           from an embedding D->G as a decomposition group
  prec = precision of Q_p, prime ring of F
*/


// Default settings for factoring the discriminant of the defining
// polynomial of K

DefaultDiscFactor:=<10000,65535,10,0,false>;


// mmylib functions

Z:=Integers();
Q:=Rationals();
PR:=PolynomialRing;
Count:=func<S,x|#[z: z in S | z eq x]>;   // count occurences of x in a list
CycleType:=func<g|Sort([#x: x in CycleDecomposition(g)])>;     // incr order
DelCRs:=func<s|&cat([x: x in Eltseq(Sprint(s)) | x ne "\n"])>; // delete \n's
DelSpaces:=func<s|&cat([x: x in Eltseq(Sprint(s)) | (x ne " ") and (x ne "\n")])>;

// hackobj creation functions


intrinsic Print(A::ArtRep, level::MonStgElt)
{Print an Artin representation}
  if level eq "Magma" then
    printf "%o !! %o",Sprint(A`K,"Magma"), DelCRs(Eltseq(Character(A)));
  else
    if not IsCharacter(Character(A)) or (not assigned A`K`artinrepdata`ramified)
      then condstr:="";
      else condstr:=" and conductor " cat Sprint(Conductor(A)) cat
        (A`K`artinrepdata`disc_unfactored cmpeq 1 select "" else "(?)");
    end if;
    printf "Artin representation of " cat Sprint(A`K,"Minimal") cat " with character "
      cat Sprint(A`character,"Minimal") cat condstr;
  end if;
end intrinsic;

intrinsic IsCoercible(x::ArtRep, y::.) -> BoolElt, .
{Coerce an Artin representation}
  return false, _;
end intrinsic;

intrinsic 'in'(x::ArtRep, y::.) -> BoolElt
{"in" function for an Artin representation}
  return false;
end intrinsic;

intrinsic ArtRepCreate() -> ArtRep
{Create an Artin representation (hackobj function)}
  return New(ArtRep);
end intrinsic;


// Degrees of the irreducible factors of f defined over F_p

function CycleTypeFp(f)

  //! Bug in DistinctDegreeFactorization for p=2
  if Characteristic(BaseRing(Parent(f))) eq 2 then
    fct:=[u[1]: u in Factorisation(f)];
    degs:=Sort([Degree(u): u in fct]);
    sdegs:=Sort(SetToSequence(Set(degs)));
    fct:=[<d,&*[u: u in fct | Degree(u) eq d]> : d in degs];
    return degs,fct;
  end if;

  fct:=DistinctDegreeFactorization(f);
  return Sort(&cat[ [d[1]: i in [1..Degree(d[2])/d[1]]]: d in fct]),fct;
end function;


// Find an unramified prime with smallest residue degree >=mindeg
// in a given number of attempts, and starting with a given p

function GoodGaloisPrime(f: p:=100, attempts:=200, mindeg:=1)
  dsc:=Z!Abs(Discriminant(f));
  assert dsc ne 0;
  success:=0;
  minextdeg:=0;
  minp:=0;
  repeat
    p:=NextPrime(p);
    if dsc mod p eq 0 then continue; end if;
    success+:=1;
    extdeg:=LCM(CycleTypeFp(PR(GF(p))!f));
    if (extdeg ge mindeg) and ((minp eq 0) or (extdeg lt minextdeg)) then
      minextdeg:=extdeg; minp:=p;
    end if;
  until success gt attempts;
  return minp,minextdeg;
end function;


// Galois group of a polynomial f in Q[x]; returns p, G, roots in Z_p
//   parameter p is either a fixed prime (e.g. 17)
//     or "galois" (default in GaloisGroup)
//     or [startp,attempts,mindeg] (use GoodGaloisPrime)

function ComputeGaloisGroup(f,p)
  if Type(p) cmpeq RngIntElt then
    G,r,data:=GaloisGroup(f: Prime:=p); return p,G,r,data;
  end if;
  if p cmpeq "galois" then
    G,r,data:=GaloisGroup(f);
    return Prime(Universe(r)),G,r,data;
  end if;
  startp,attempts,mindeg:=Explode(p);
  p,deg:=GoodGaloisPrime(f:
    p:=startp, attempts:=attempts, mindeg:=Min(Degree(f),mindeg));
  if p eq 0 then
    vprint ArtRep,1: "Failed to find a good Galois prime, data:",Sprint([startp,attempts,mindeg]);
  else
    try
      vprint ArtRep,1: Sprintf("Attempting Galois group with p=%o deg=%o",p,deg);
      G,r,data:=GaloisGroup(f: Prime:=p);
      return p,G,r,data;
    catch e;
    end try;
  end if;
  "Galois group failed, choosing default prime";
  G,r,data:=GaloisGroup(f);
  return Prime(Universe(r)),G,r,data;
end function;


// Compute splitting field of a polynomial, by repeatedly increasing
// precision until SplittingField(f,Zp) works

// very slow for some unknown reason

function ComputeSplittingField1(f,p : prec:=20)
  for tries:=1 to 10 do
    try
      vprint ArtRep, 2: "spl> Splitting field, precision =",prec;
      Zp:=pAdicRing(p,prec);
      R:=SplittingField(f,Zp);
      rts:=Roots(f,R);
      assert #rts eq Degree(f);
      return R,Zp,[r[1]: r in rts];
    catch e; 
      prec:=2*prec;
      vprint ArtRep,2: "spl> Failed (",DelCRs(e`Object),"); increasing precision to",prec;
    end try;
  end for;
  return false,_,_;
end function;


function AbsoluteTotallyRamifiedExtension(R)
//! does not construct embeddings

  f := AbsoluteInertiaDegree(R);
  e := AbsoluteRamificationDegree(R);
  B := f eq 1 select PrimeRing(R) else UnramifiedExtension(PrimeRing(R), f);
  if e eq 1 then return B; end if;
  resext := PolynomialRing(Integers())!
    MinimalPolynomial(PrimitiveElement(ResidueClassField(B)));
  rB := Roots(resext,B)[1,1];
  rR := Roots(resext,R)[1,1];
  piR := UniformizingElement(R);

  l := Matrix([Flat(piR^i*rR^j) : j in [0..f-1], i in [0..e-1] ]
       cat [Flat(piR^e)]);
  n := NullspaceMatrix(l);

  assert Nrows(n) eq 1;
  assert IsUnit(n[1][e*f+1]);

  n := n/n[1][e*f+1];
  M := Matrix(e,f,Prune(Eltseq(n)));
  f := [Evaluate(Polynomial(Eltseq(M[i])),rB): i in [1..e]] cat [B|1];
  S := TotallyRamifiedExtension(B, Polynomial(f));

  return S;
end function;


function ComputeSplittingField(f,p: prec:=20)
  prec0 := prec;

  for attempt:=1 to 10 do
  try
    Qp:=pAdicField(p,prec);
    K:=Qp;
    F:=f;
    rts:=[];
    for i:=1 to Degree(f) do
      R<x>:=PR(K);
      F:=R!F;
      vprint ArtRep,2: "loc> Finding roots (prec",prec,")";
      rtsF:=[r[1]: r in Roots(F)];
      rts:=[K!r: r in rts] cat rtsF;
      if #rts eq Degree(f) then
        OK:=IntegerRing(K);
        vprint ArtRep,2: "loc> Converting to an absolute extension";
        R:=AbsoluteTotallyRamifiedExtension(OK);
        rts:=[r[1]: r in Roots(f,FieldOfFractions(R))];
        error if #rts ne Degree(f), "Found only",#rts,"out of",Degree(f),"roots in the splitting field";
        return R,IntegerRing(Qp),rts;
      end if;
      F:=f div &*[R| x-r: r in rts];
      _,_,e:=Factorisation(F: Extensions:=true);
      _,indx:=Max([AbsoluteDegree(d`Extension): d in e]);
      K:=e[indx]`Extension;
      vprint ArtRep,2: "loc> Adjoining root -> [K:Q_p] =",AbsoluteDegree(K);
    end for;
    error "Should exit before";
  catch e;
    prec:=2*prec;
    vprint ArtRep,2: "Failed (",DelCRs(e`Object),"); increasing precision to",prec;
  end try;
  end for;
  
  // [SRD, Dec 2011]
  vprint ArtRep,2: "loc> Plan B: call (slow?) SplittingField";
  OK:=ComputeSplittingField1(f, p : prec:=prec0);
  if R cmpne false then
    vprint ArtRep,2: "loc> Converting to an absolute extension";
    R:=AbsoluteTotallyRamifiedExtension(OK);
    rts:=[r[1]: r in Roots(f,FieldOfFractions(R))];
    error if #rts ne Degree(f), "Found only",#rts,"out of",Degree(f),"roots in the splitting field";
    Zp:=BaseRing(BaseRing(R));
    return R,Zp,rts;
  end if;

  error "Failed to compute the splitting field for f = ",f," at p =",p;
end function;


// Make roots of f in Z_p accurate to at least a given precision

procedure Henselize(f,~rts,prec);
  if Precision(rts[1]) ge prec then return; end if;
  vprint ArtRep,2: "Henselizing roots to precision "*Sprint(prec);
  Qp:=ChangePrecision(Universe(rts),prec);
  f:=PR(Qp)!f;
  assert LeadingCoefficient(f) eq 1;    //! Otherwise HenselLift fails
  rts:=[HenselLift(f,Qp!r): r in rts];
  vprint ArtRep,2: "Henselized successfully";
end procedure;


// Which powers of c in G are conjugate to c?

ConjugatePowers:=func<G,c|[j: j in [1..Max(1,o-1)] |
  (GCD(j,o) eq 1) and IsConjugate(G,c,c^j)] where o is Order(c)>;


// Convert a Galois automorphism of a p-adic field (map field->field)
// to a permutation in S_n

function GaloisToPermutation(F,r)
  perm:=[];
  for x in r do
    v,m:=Max([Valuation(F(x)-R): R in r]);
    assert v gt 0;         // F must permute the roots, just checking
    Append(~perm,m);
  end for;
  return Sym(#r)!perm;
end function;


// Our main ST polynomial: roots are sums
//    evalpol(r[i]) * (r[g[i]]+r[g^u[i]]+...), u in pow
// indexed by elements g in a conjugacy class (modulo pow),
// and evalpol is x^2 or some other polynomial (if x^2 fails)

// Note: STInvariant only works for monic f's


function ZpToZ(n)
  a:=Z!n;
  b:=-(Z!(-n));
  if Abs(a) lt Abs(b) then return a; else return b; end if;
end function;


// Set of representatives of the conjugacy classes of an element classrep
// in G modulo a specified list of powers (#(classrep^G)/#pow of them)

function ClassRepsModPowers(G,classrep,pow)
  c:=[c: c in classrep^G];
  O:=[];
  while not IsEmpty(c) do
    x:=c[1];
    for j in pow do Remove(~c,Position(c,x^j)); end for;
    Append(~O,x);
  end while;
  return O;
end function;


/*
function STInvariantPoly(G,rts,classrep,pow,evalpol);
  C:=ClassRepsModPowers(G,classrep,pow);
  Rp<z>:=PR(Universe(rts));
  erts:=[Evaluate(evalpol,r): r in rts];
  f:=&*[z - &+[ erts[i]* &+[rts[i^(g^j)]: j in pow]
    : i in [1..#rts]]: g in C];
  return PR(Q)![ZpToZ(c): c in Eltseq(f)];
end function;
*/


function STInvariantPoly(G,rts,classrep,pow,evalpol);
  C:=ClassRepsModPowers(G,classrep,pow);
  Rp<z>:=PR(Universe(rts));
  erts:=[Evaluate(evalpol,r): r in rts];

  K:=Universe(rts);
  Zp:=PrimeRing(K);
  RZp:=PR(Zp);
  k:=ResidueClassField(K);
  p:=Characteristic(k);
  Fp:=PrimeField(k);
  f:=Degree(k,Fp);
  n:=#rts;
  // Find Frobenius in G
  frob:=[[j: j in [1..n] | Valuation(r-rts[j]) gt 0] where r is rts[i]^p:
          i in [1..n]];
  frob:=Sym(n)!Flat(frob);
  assert frob in G;
  ftots:=[RZp|];
  c:=[c: c in classrep^G];   // split the conjugacy class into groups
                             // both by powers in pow and by Frobenius action
                             // to minimize the computation and make
                             // intermediate polys (stored in ftots)
                             // defined over Zp rather than K
                             // (big speed improvement)
  while not IsEmpty(c) do
    x:=c[1];
    f:=1;
    repeat
      for j in pow do Remove(~c,Position(c,x^j)); end for;
      froot:=&+[ erts[i] * &+[rts[i^(x^j)]: j in pow]: i in [1..#rts]];
      f*:=(z-froot);
      x:=x^frob;
    until not (x in c);
    f:=RZp!f;
    Append(~ftots,f);
  end while;
  return PR(Q)![ZpToZ(c): c in Eltseq(&*ftots)];
end function;


function STInvariantGen(rts,F,data)
  pow,evalpol:=Explode(data);
  return &+[Evaluate(evalpol,rts[i]) * &+[rts[i^(F^j)]: j in pow]: i in [1..#rts]];
end function;


function STInvariantFp(f,fct,data)
  pow,evalpol:=Explode(data);
  R:=Parent(f);
  p:=Characteristic(R);
  Q<x>:=quo<R|R!f>;
  u:=Evaluate(evalpol,x) * &+[x^(p^j): j in pow];
  // compute its trace
  trace:=0;
  deg:=Degree(f)-1;
  for j:=0 to deg do
    trace+:=Coefficient(u,j);
    if j ne deg then u*:=x; end if;
  end for;
  return trace;
end function;


// For an irreducible f in F_p[t], compute the `alternating' invariant
// (Serre's trick), which is one of the square roots of the discriminant of f

function SqrtDiscIrrPoly(f)
  d:=Degree(f);
  R:=Parent(f);
  p:=Characteristic(R);
  Q<x>:=quo<R|f>;
  rts:=[];
  r:=x;
  for i:=1 to d do
    Append(~rts,r);
    if i ne d then r:=r^p; end if;
  end for;
  return GF(p)!&*[rts[i]-rts[j]: i,j in [1..#rts] | i gt j];
end function;

// Same, but for a general f in F_p[t], using resultants between the
// irreducible factors and SqrtDiscIrrPoly for the irreducible ones

function SqrtDiscInvariantFp(f,fct,dontcare)
  k:=BaseRing(Parent(f));
  degs:=[f[1]: f in fct];
  Sort(~degs,~perm);
  fct:=PermuteSequence([f[2]: f in fct],perm);
  return &*[k|Resultant(fct[i],fct[j]): i,j in [1..#fct] | i gt j] *
    &*[k|SqrtDiscIrrPoly(u): u in fct | Degree(u) gt 1];
end function;

// Same, but over Q_p

function SqrtDiscInvariantGen(r,F,can)
  _,c:=IsConjugate(Sym(#r),F,can);
  sgn:=IsOdd(c) select -1 else +1;
  return sgn * &*[r[i]-r[j]: i,j in [1..#r] | i gt j];
end function;


function IdentifyFrobClassFp(data,p)
  R:=PR(GF(p));
  f:=data`f;
  cyc,fct:=CycleTypeFp(R!f);
  inv:=data`Inv[Position(data`cycs,cyc)];
  if Type(inv) eq RngIntElt then return inv; end if;   // by cycle type
  funFp,funQp,par,list:=Explode(inv);
  sinv:=funFp(R!f,fct,par);
  vals:=[Evaluate(R!m[1],sinv): m in list];
  if Count(vals,0) eq 1 then         // by an invariant computation over F_p
    frob:=list[Position(vals,0),2];
    return frob;
  else                               // several invariants are roots ->
    return 0;                        // failed, need to work over Q_p
  end if;
end function;


function IdentifyConjugacyClassQp(data,F,r)
  cyc:=CycleType(F);
  inv:=data`Inv[Position(data`cycs,cyc)];
  if Type(inv) eq RngIntElt then return inv; end if;  // by cycle type
  funFp,funQp,par,list:=Explode(inv);                 // using an invariant

  sinv:=Universe(r)!funQp(r,F,par);
  vals:=[Valuation(Evaluate(m[1],sinv)): m in list];
  v,max:=Max(vals);

  assert v gt 0;               // at least one zero must occur

  // and the largest one only once, otherwise failed
  if Count(vals,v) ne 1 then return 0; end if;

  frob:=list[max][2];
  return frob;
end function;


function LocalData(K,p: prec:=20)
  A:=K`artinrepdata;
  badpos:=Position(A`badprimes,p);
  if (badpos ne 0) and (A`localdata[badpos][6] ge prec)
		   then return Explode(A`localdata[badpos]);
  end if;
  f:=A`f;

  try
    vprint ArtRep,1: Sprintf("Local data at p=%o; vdisc=%o; prec=%o",
      p,Valuation(A`disc,p),prec);
    R,Zp,rts:=ComputeSplittingField(f,p: prec:=prec);
    prec:=Precision(Zp);
    //rts:=[r[1]: r in Roots(f,R)]; assert #rts eq Degree(f);
    n:=#rts;
    assert IsIdentical(BaseRing(Zp), Zp); // Zp is a pAdicRing
    e:=AbsoluteRamificationDegree(R);
    k:=ResidueClassField(R);
    Rz:=R!k.1;
    Fp:=PrimeField(k);
    f:=Degree(k,Fp);
    assert e*f eq AbsoluteDegree(R);
    pi:=UniformizingElement(R);

    vprint ArtRep,2: "loc> Automorphisms";
    DA<[DAGens]>,DAtoAut:=AutomorphismGroup(FieldOfFractions(R));
    AutGens:=[DAtoAut(g): g in DAGens];

    vprint ArtRep,2: "loc> Subgroup of S_n";
    DGens:=[GaloisToPermutation(g,rts): g in AutGens];
    DSn:=sub<Sym(n)|DGens>;
    DAtoDSn:=hom<DA->DSn|DGens>;
    error if #DSn ne e*f, "Decomposition group failed";

    vprint ArtRep,2: "loc> Frobenius (order",f,")";
    ok:= exists(FrobDA){h: h in DA | Valuation((DAtoAut(h))(Rz)-Rz^p) gt 0};
    error if not ok, "Frobenius automorphism failed";
    Frob:=DAtoDSn(FrobDA);

    vprint ArtRep,2: "loc> Inertia (order",e,")";
    I:=[h: h in DA | Valuation((DAtoAut(h))(Rz)-Rz) gt 0];
    error if #I ne e, "Inertia failed";

    vprint ArtRep,2: "loc> Ramification groups";
    ivals:=[Valuation((DAtoAut(h))(pi)-pi): h in I];
    ramind:=Sort(SetToSequence(Set(ivals)));
    ramgps:=[<sub<DSn|[ DAtoDSn(I[j]) : j in [1..#I] | ivals[j] ge i]>,i>:
		   i in ramind | i ne ramind[#ramind]];
    if IsEmpty(ramgps) then ramgps:=[<sub<DSn|>,prec>]; end if;

    vprint ArtRep,2: "loc> Class map D->G";
    CCD:=ConjugacyClasses(DSn);
    CD:=CharacterTable(DSn);
    CC:=A`CC;

    classmap:=[IdentifyConjugacyClassQp(A,c[3],rts): c in CCD];

    error if (0 in classmap) or 
      not forall{rho: rho in A`C | IsCharacter( 
      Universe(CD)![rho(CC[classmap[i]]): i in [1..#ConjugacyClasses(DSn)]])},
      "not enough precision to construct the class map";

    vprint ArtRep,1:
      Sprintf("/Q_%o: D=%o I=%o G_i=%o classmap=%o",
	      p,#DSn,#I,DelSpaces([Sprintf("%o:%o",#H[1],H[2]): H in ramgps]),
	      DelSpaces(classmap));

  catch e
    vprint ArtRep,2: e`Object;
    prec:=2*prec;
    if prec gt 1280 then
      error "Local data failed at "*Sprint(p)*"; not enough precision";
    end if;
    return LocalData(K,p: prec:=prec);
  end try;     

  data:=<rts, DSn, Frob, ramgps, classmap, Precision(Zp),
         Inverse(DAtoDSn)*DAtoAut>;
  if badpos eq 0 then
    Append(~K`artinrepdata`badprimes,p);
    Append(~K`artinrepdata`localdata,data);
  else
    K`artinrepdata`localdata[badpos]:=data;
  end if;

  return Explode(data);

end function;


// Pull back a character rho of G to a local decomposition group at p

function LocalRepresentation(K,rho,p)
  r,DSn,FPerm,ramgps,DtoG,prec:=LocalData(K,p);
  a:=K`artinrepdata;
  CD:=CharacterTable(DSn);
  CC:=a`CC;
  chi:=Universe(CD)![rho(CC[DtoG[i]]): i in [1..#ConjugacyClasses(DSn)]];
  //assert IsCharacter(chi); // added MW, moved TD  17 Apr 2012
  return chi;
end function;


function RamifiedPrimes(K)
  if assigned K`artinrepdata`ramified then
    return K`artinrepdata`ramified;
  end if;
  vprint ArtRep, 2: "Factoring the discriminant of f";
  disc:=K`artinrepdata`disc;
  fct:=K`artinrepdata`disc_unfactored;
  if fct cmpeq true then
    vprint ArtRep, 1: "Factoring the discriminant";
    P:=PrimeDivisors(disc);
  else
    if fct cmpeq false then fct:=DefaultDiscFactor; end if;
    TL,PL,EL,ML,PF:=Explode(fct);
    P:={f[1]: f in Factorisation(disc: TrialDivisionLimit:=TL,
      PollardRhoLimit:=PL, MPQSLimit:=ML, ECMLimit:=EL, Proof:=PF)};
    P:=Sort([p: p in P]);
    U:=disc div &*[Z|p^Valuation(disc,p): p in P];
    if U ne 1 then
      vprint ArtRep,1: "Unfactored part",U;
    end if;
    K`artinrepdata`disc_unfactored:=U;
  end if;
  vprint ArtRep, 1: "Primes dividing the discriminant",DelSpaces(P);
  vprint ArtRep, 2: "Computing ramified primes";
  K`artinrepdata`ramified:=[p: p in P | #ramgps[1][1] ne 1
    where _,_,_,ramgps,_,_ is LocalData(K,p)];
  return K`artinrepdata`ramified;
end function;


intrinsic Conductor(A::ArtRep) -> RngIntElt
{Conductor of an Artin representation}
  if assigned A`conductor then return A`conductor; end if;
  require IsCharacter(A`character):
   "Conductor is not defined for virtual representations";
  cond:=1;
  for p in RamifiedPrimes(A`K) do
    _,_,_,ramgps,_,_:=LocalData(A`K,p);
    ch:=LocalRepresentation(A`K,A`character,p); // pullback of A to
                                                // the local Galois group
    first:=true;
    delta:=0;                                   // conductor exponent
    lastindex:=0;
    g0:=#ramgps[1][1];
    for I in ramgps do
      coinvdim:=Degree(ch)-InnerProduct(Restriction(ch,I[1]),
                PrincipalCharacter(I[1]));
      if coinvdim eq 0 then break; end if;
      delta +:= coinvdim * (I[2]-lastindex) * #(I[1]) / g0;
      lastindex:=I[2];
    end for;
    cond*:=p^(Z!delta);

  end for;

  A`conductor:=cond;
  return cond;
end intrinsic;


// Conjugacy class of Frobenius at infinity (complex conjugation)

function ComplexConjugation(K)

  A:=K`artinrepdata;
  if assigned A`complexconjugation then return A`complexconjugation; end if;

  vprint ArtRep,2: "Class of complex conjugation";

  r1:=#Roots(A`f,RealField());       // was r1,r2:=Signature(K);
  r2:=(Degree(A`f)-r1) div 2;        // but this should feel f instead

  cyc:=[1: i in [1..r1]] cat [2: i in [1..r2]];
  inv:=A`Inv[Position(A`cycs,cyc)];
  if Type(inv) eq RngIntElt then           // by cycle type
    K`artinrepdata`complexconjugation:=inv;
    return inv;
  end if;
  funFp,funQp,par,list:=Explode(inv);     // using an invariant

  maxdeg:=Max([Degree(u[1]): u in list]);
  maxcf:=Max([Max([Abs(c): c in Coefficients(u[1])]): u in list])+1;
  prec:=2*Round(Log(10,maxcf))+30;   //!

  vprint ArtRep,1: "Identifying complex conjugation, complex precision", prec;
  C:=ComplexField(prec);

  r:=[z[1]: z in Roots(A`f,C)];    // real roots followed by complex roots
  r:=[x: x in r | IsReal(x)] cat [x: x in r | not IsReal(x)];
  F:=&*[Sym(r1+2*r2)|(r1+2*i-1,r1+2*i): i in [1..r2]];      // permutation
  sinv:=C!funQp(r,F,par);

  vals:=[Abs(Evaluate(m[1],sinv)): m in list];
  v,min:=Min(vals);

  assert v lt 1E-15;                         // at least one zero must occur
  assert #[x: x in vals | x lt 10^4*v ] le 1;  // and only once

  frob:=list[min][2];
  K`artinrepdata`complexconjugation:=frob;
  return frob;
end function;


function Frobenius(K,p)
  if p cmpeq Infinity() then return ComplexConjugation(K); end if;

  data:=K`artinrepdata;
  if data`disc mod p ne 0 then
    frob:=IdentifyFrobClassFp(data,p);
    if frob ne 0 then return frob; end if;
  end if;

  prec:=20;
  repeat
    r,_,FPerm,_,_,prec:=LocalData(K,p: prec:=prec);
    frob:=IdentifyConjugacyClassQp(data,FPerm,r);
    if frob ne 0 then return frob; end if;
    prec:=2*prec;
    vprint ArtRep,2: "Frobenius failed, increasing local precision to",prec;
  until prec gt 1280;
  error "Failed to compute local data at "*Sprint(p);

end function;



// Ordered by cycles in increasing size, e.g. for t=[1,2,3] gives (2,3)(4,5,6)
function CanonicalElementOfCycleType(t)
  l:=[n+1: n in [1..&+t]];
  n:=0;
  for c in t do
    l[n+c]:=n+1; n+:=c;
  end for;
  return Sym(&+t)!l;
end function;


// Ramification: false, true
// FactorDiscriminant: false, true, <Trlim, Plim, Elim, Mlim, Proof>
// p0: p0:=17, p0:="galois", p0:=[minp, attempts, mindeg]
// f: monic minimal integral poly of some primitive element of K/Q


procedure ComputeArtinRepresentations(~K:
  Ramification:=false, FactorDiscriminant:=false, p0:="default", f:="default");

  if assigned K`artinrepdata then
    if Ramification then P:=RamifiedPrimes(K); end if;
    return;
  end if;

  vprint ArtRep,1: "Artin representations of",K;

  if f cmpeq "default" then
    f:=MinimalPolynomial(PrimitiveElement(AbsoluteField(K)));
    P:=&join {Set(PrimeDivisors(Denominator(a))): a in Eltseq(f)};
    for p in P do
      repeat                                    // make it monic integral
        f:=p^Degree(f)*Evaluate(f,Parent(f).1/p);
      until forall{a: a in Eltseq(f) | Valuation(a,p) ge 0};
    end for;
  end if;
  f:=PR(Z)!f;

  vprint ArtRep,1: "f =",DelSpaces(f);

  n:=Degree(f);

  // Upper bound on the size of coefficients of the ST_invariants m_c
  // if they have degree d and are symmetrized over j powers

  crts:=[r[1]: r in Roots(f,ComplexField())];
  rabs:=Reverse(Sort([Abs(r): r in crts]));
  mbnd:=func<evp|n*Max([Abs(x): x in Eltseq(evp)])*rabs[1]^Degree(evp)*rabs[2]>;
  mcfbound:=func<d,evp,j|Max([Binomial(Z!d,i)*(mbnd(evp)*j)^i: i in [0..d]])>;

  if p0 cmpeq "default" then
    p0:="galois";  //!
    //p0:=[100,200,2+(n div 10)];
  end if;

  p0,G,r,galdata:=ComputeGaloisGroup(f,p0);
  vprint ArtRep,1: Sprintf("#G=%o over Z_%o^%o",#G,p0,AbsoluteDegree(Universe(r)));

  CC:=[c: c in ConjugacyClasses(G)];
  order:=[c[1]: c in CC];
  csizes:=[c[2]: c in CC];
  cc:=[c[3]: c in CC];
  cyc:=[CycleType(c): c in cc];

  cycs:=Sort(SetToSequence(Set(cyc)));
  disc:=Z!Discriminant(f);
  sqrtdsc:=0;
  R<x>:=PR(Z);
  StdEvalPolyList:=[x^2, x^2-3*x+1, x^3-7*x^2+11*x-5, x^4-7*x^3-2*x^2+5*x+2,
                    x^2, &+[(n-i)*x^i: i in [0..n-1]] ];

  Inv:=[* "?": i in [1..#cycs] *];
  for j:=1 to #cycs do
    c:=cycs[j];
    classes:=[i: i in [1..#cc] | cyc[i] eq c ];

    // Distinguishable by cycle type
    if #classes eq 1 then
      Inv[j]:=classes[1];
      vprint ArtRep,2: DelSpaces(c),"->",Inv[j];
      continue;
    end if;

    // Sqrt(discriminant): cycle types with distinct odd cycles for G in A_n
    if (#Set(c) eq #c) and forall{x: x in c | IsOdd(x)} and (#classes eq 2)
       and (G subset Alt(n)) then

       can:=CanonicalElementOfCycleType(c);
       // can:=cc[classes[1]];  // does not always work

      _,c1:=IsConjugate(Sym(n),cc[classes[1]],can);
      _,c2:=IsConjugate(Sym(n),cc[classes[2]],can);
      if IsOdd(c1) ne IsOdd(c2) then                  // Distinct in A_n
        vprint ArtRep,1: DelSpaces(c),"is alternating";
        if sqrtdsc cmpeq 0 then
          sqrtZ:=Sqrt(Discriminant(f));
          prec:=Round(Log(p0,sqrtZ))+2;
          Henselize(f,~r,prec);
          sqrtdsc:=ZpToZ(&*[r[i]-r[j]: i,j in [1..n] | i gt j]);
          vprint ArtRep,2: "Sqrt discriminant =",sqrtdsc;
        end if;
        Inv[j]:=<SqrtDiscInvariantFp,SqrtDiscInvariantGen,can,
                [<x-sqrtdsc,IsEven(c1) select classes[1] else classes[2]>,
                 <x+sqrtdsc,IsEven(c1) select classes[2] else classes[1]>]>;
        continue;
      end if;
    end if;

    pow:=Sort(SetToSequence(&meet{Set(ConjugatePowers(G,cc[c])): c in classes}));
    // powers of elts in all our conjugacy classes that lie in the same class
    invdeg:=Max(csizes[classes])/#pow;
    vprint ArtRep,1: DelSpaces(c),"deg",invdeg,"->",classes;

    for m:=1 to 30 do              // MaxInvAttempts:=30 once and for all
      if m le #StdEvalPolyList
        then evalpol:=StdEvalPolyList[m];
             if m eq #StdEvalPolyList-1 then
               pow:=[1]; invdeg:=Max(csizes[classes]);
             end if;
        else evalpol:=R![Random(-m,m): i in [0..Random(1,n-1)]];
      end if;

      // Compute the degree of the ST invariant and ensure sufficient p0-adic
      // precision for the roots of f to compute the polynomial itself
      prec:=Round(Log(p0,mcfbound(invdeg,evalpol,#pow)))+2;
      Henselize(f,~r,prec);

      // ST invariant

      //! Check they are powers of irr polys and replace by constituents?
      ST:=[STInvariantPoly(G,r,cc[i],pow,evalpol): i in classes];

      // Check they are coprime
      if not exists{i: i in [2..#ST], j in [1..#ST-1] |
        (i gt j) and (Degree(GCD(ST[i],ST[j])) ne 0)}
      then    // found a good invariant
        Inv[j]:=<STInvariantFp,STInvariantGen,<pow,evalpol>,
          [<ST[i],classes[i]>: i in [1..#classes]]>;
        break;
      end if;
      vprint ArtRep,2: DelSpaces(evalpol), "ST failed";
    end for;
    error if Inv[j] cmpeq "?", "Could not find distinct ST polynomials";
  end for;

  K`artinrepdata:=rec<artinrec|
    f:=f,K:=K,G:=G,
    Frob:=func<p|Frobenius(K,p)>,
    p0:=p0,r:=r,C:=CharacterTable(G),CC:=cc,galoisdata:=galdata,
    badprimes:=[],localdata:=[* *],cycs:=cycs,disc:=Abs(disc),
    disc_unfactored:=FactorDiscriminant,
    Inv:=Inv
    // ramified, complexconjugation initially undefined
  >;

  if Ramification then P:=RamifiedPrimes(K); end if;

end procedure;


intrinsic DefiningPolynomial(A:: ArtRep) -> RngUPolElt
{The polynomial whose roots Group(A) permutes}
  return Field(A)`artinrepdata`f;
end intrinsic;


intrinsic FrobeniusElement(K::FldNum, p::RngIntElt) -> GrpPermElt
{Compute a Frobenius element at p in the Galois group of the Galois closure
of K. This is a permutation on the roots of a polynomial defining K,
which can be recovered as DefininingPolynomial(A) for any Artin
representation of K; the Frobenius element is well-defined up to conjugacy
and modulo inertia.}
  error if not IsPrime(p),
    "p must be a rational prime or Infinity()";
  ComputeArtinRepresentations(~K);
  Frob:=Frobenius(K,p);
  return K`artinrepdata`CC[Frob];
end intrinsic;


intrinsic FrobeniusElement(K::FldNum, p::Infty) -> GrpPermElt
{Compute a Frobenius element at p in the Galois group of the Galois closure
of K. This is a permutation on the roots of a polynomial defining K,
which can be recovered as DefininingPolynomial(A) for any Artin
representation of K; the Frobenius element is well-defined up to conjugacy
and modulo inertia.}
  ComputeArtinRepresentations(~K);
  Frob:=Frobenius(K,p);
  return K`artinrepdata`CC[Frob];
end intrinsic;


// Standard representation functions copied for Artin representations

intrinsic Group(A::ArtRep) -> GrpPerm
{Returns the Galois group of the field K=Field(A) through
which A factors.} //, and phi specifies the action of G on K.}
  return A`K`artinrepdata`G;
end intrinsic;

intrinsic Character(A::ArtRep) -> AlgChtrElt
{Character of an Artin representation}
  return A`character;
end intrinsic;

intrinsic Degree(A::ArtRep) -> RngIntElt
{Degree (=dimension) of an Artin representation}
  d:=Degree(Character(A));
  if IsCoercible(Integers(),d) then return Integers()!d; end if;
 return d;
end intrinsic;

intrinsic Dimension(A::ArtRep) -> RngIntElt
{Degree (=dimension) of an Artin representation}
  return Degree(Character(A));
end intrinsic;

intrinsic IsIrreducible(A::ArtRep) -> BoolElt
{True if the Artin representation A is irreducible}
  return IsCharacter(ch) and IsIrreducible(ch) where ch is Character(A);
end intrinsic;


intrinsic Decomposition(A::ArtRep) -> SeqEnum[Tup]
{Decompose A into irreducible consituents. Returns a sequence [...<A_i,n_i>...]
  with A_i irreducible and n_i their exponents in A (possibly negative)}
  ch:=Character(A);
  K:=Field(A);
  C:=K`artinrepdata`C;
  return [<K!!c,InnerProduct(ch,c)>: c in C | InnerProduct(ch,c) ne 0];
end intrinsic;


function ArtinRepIsRamified(A,p,wild)
// check whether A is ramified (wild=false) or wildly ramified (wild=true)

  d:=A`K`artinrepdata;
  if (d`disc mod p ne 0) or (assigned d`ramified and not (p in d`ramified))
    then return false; end if;
  if wild and (#d`G mod p ne 0) then return false; end if;
  _,_,_,ramgps,_,_:=LocalData(A`K,p);
  I:=ramgps[1][1];                              // inertia at p
  if #I eq 1 then return false; end if;
  if wild then I:=SylowSubgroup(I,p); end if;   // wild inertia (wild=true)
  if #I eq 1 then return false; end if;
  ch:=LocalRepresentation(A`K,A`character,p); // pullback of A to
                                              // the local Galois group
  invdim:=InnerProduct(Restriction(ch,I),PrincipalCharacter(I));
  return invdim ne Dimension(A);
end function;


intrinsic IsRamified(A::ArtRep, p::RngIntElt) -> BoolElt
{True if the Artin representation A is ramified at p}
  require IsPrime(p): "p must be a prime number";
  require IsCharacter(Character(A)): "Artin representation may not be virtual";
  return ArtinRepIsRamified(A,p,false);
end intrinsic;


intrinsic IsWildlyRamified(A::ArtRep, p::RngIntElt) -> BoolElt
{True if the Artin representation A is wildly ramified at p}
  require IsPrime(p): "p must be a prime number";
  require IsCharacter(Character(A)): "Artin representation may not be virtual";
  return ArtinRepIsRamified(A,p,true);
end intrinsic;


intrinsic Field(A::ArtRep) -> FldNum
{Number field K such that A factors through the Galois closure of K}
  return AbsoluteDegree(A`K) eq AbsoluteDegree(A`K`artinrepdata`K)
    select A`K else A`K`artinrepdata`K;
end intrinsic;


intrinsic '!!'(F:: FldNum, A:: ArtRep: MinPrimes:=20) -> ArtRep, BoolElt
{Given an Artin representation A of some number field, recognize it
if possible as an Artin representation of F and return F!!A,true.
If not possible, return 0,false}

  if A`K cmpeq F then return A; end if;

  K:=Field(A);
  AF:=ArtinRepresentations(F);
  Kdata:=K`artinrepdata;
  Fdata:=F`artinrepdata;
  CCK:=Kdata`CC;
  CCF:=Fdata`CC;

  decA:=Decomposition(A);
  Alist:=[Character(data[1]): data in decA];
  Amult:=[data[2]: data in decA];

  matches:=[[c: c in Fdata`C | (Degree(c) eq Degree(a))
       and (Set(Eltseq(c)) eq Set(Eltseq(a)))]: a in Alist];

  p:=100;
  finished:=0;
  repeat
    p:=NextPrime(p);
    if (Kdata`disc mod p eq 0) or (Fdata`disc mod p eq 0) then continue; end if;

    frobK:=CCK[Kdata`Frob(p)];
    frobF:=CCF[Fdata`Frob(p)];

    finished+:=1;
    for i:=1 to #Alist do
      n:=#matches[i];
      if n eq 0 then return 0, false; end if;
      if n ne 1 then finished:=0; end if;
      //p,Alist[i],finished;
      matches[i]:=[m: m in matches[i] | Alist[i](frobK) eq m(frobF)];
    end for;
  until finished eq MinPrimes;

  return F!! &+[Amult[i]*matches[i][1]: i in [1..#Alist]],true;

end intrinsic;


intrinsic ChangeField(A:: ArtRep, F:: FldNum: MinPrimes:=20) -> ArtRep, BoolElt
{Given an Artin representation A of some number field, recognize it
if possible as an Artin representation of F and return F!!A,true.
If not possible, return 0,false}
  return '!!'(F,A: MinPrimes:=MinPrimes);
end intrinsic;


function VirtualCharacterKernel(a)
  if a eq 0
    then return Group(Parent(a)); end if;
  if not IsCharacter(a) then
    a:=&+[ch: ch in Basis(Parent(a)) | InnerProduct(ch,a) ne 0];
  end if;
  return Kernel(a);
end function;


intrinsic Minimize(A::ArtRep: Optimize:=true) -> ArtRep
{A restricted to the smallest number field K such that A factors through the
      its Galois closure. If Optimize=true (default), attempt to
      minimize the defining polynomial of K using OptimizedRepresentation}

  ch:=Character(A);
  H:=VirtualCharacterKernel(ch);

  F:=A`K;
  d:=AbsoluteDegree(F);
  a:=F`artinrepdata;
  G:=a`G;
  galdata:=a`galoisdata;
  GQ,m:=quo<G|H>;

  prevlimit:=0;
  limits:=[a:a in [3,5,10,20,30,40,60] | a lt d-1] cat 
    (#H eq 1 select [d-1] else [d]);

  ok:=false;
  for indexlimit in limits do
    S:=[H`subgroup: H in Subgroups(GQ: IndexLimit:=indexlimit) |
        Index(GQ,H`subgroup) gt prevlimit];
    for i:=prevlimit+1 to indexlimit do
      ok:=exists(H){H: H in S | (Index(GQ,H) eq i) and
          IsFaithful(PermutationCharacter(GQ,H))};
      if ok then break; end if;
    end for;
    if ok then break; end if;
    prevlimit:=indexlimit;
  end for;

  if not ok then
    if not Optimize then return false; end if;
    K:=OptimizedRepresentation(F);
    if K eq F then return A; end if;
    if not assigned K`artinrepdata then
      K`artinrepdata:=a;
      K`artinrepdata`K:=K;
    end if;
    AMin:=K!!A;
    return A;
  end if;

  K:=NumberField(GaloisSubgroup(galdata,H@@m)); 
  if Type(K) eq FldRat then K:=RationalsAsNumberField(); end if;
  if Optimize then K:=OptimizedRepresentation(K); end if;
  AMin:=K!!A;
  return AMin;

end intrinsic;


intrinsic Kernel(A::ArtRep) -> FldNum
{The smallest Galois extension K/Q such that A factors through Gal(K/Q).
Note that this field may be enormous and uncomputable.}
  K:=SplittingField(DefiningPolynomial(Minimize(A)));
  return K;
end intrinsic;


function GenArtinCompose(A1,opfun,A2: charsonly:=false)

  /*
  //! Minimize?
  A1:=Minimize(A1);
  A2:=Minimize(A2);
  */

  F1A:=A1`K`artinrepdata`K;
  F2A:=A2`K`artinrepdata`K;
  vprint ArtRep,1: "Composing the fields";

  if   IsCoercible(F2A,F1A.1) then F:=F2A;
  elif IsCoercible(F1A,F2A.1) then F:=F1A;
  else
    F:=CompositeFields(AbsoluteField(F1A),AbsoluteField(F2A));
                                        //! use reducible f when it works?
      d,i:=Min([Degree(K): K in F]);
      if   d eq Degree(F1A,Q) then F:=F1A;
      elif d eq Degree(F2A,Q) then F:=F2A;
      else
        F:=F[i];
    end if;
  end if;
  vprint ArtRep,1: "Compositum:",DelSpaces(DefiningPolynomial(F)),
    "; now composing the representations";

  if charsonly
    then return F,F!!A1,F!!A2;
    else return opfun(F!!A1,F!!A2);
  end if;

end function;


intrinsic '+'(A1::ArtRep,A2:: ArtRep) -> ArtRep
{Direct sum of two Artin representations}
  if A1`K cmpeq A2`K
    then return A1`K!!(Character(A1)+Character(A2));
    else return GenArtinCompose(A1,'+',A2);
  end if;
end intrinsic;


intrinsic '-'(A1::ArtRep,A2:: ArtRep) -> ArtRep
{Direct difference of two Artin representations}
  if A1`K cmpeq A2`K
    then return A1`K!!(Character(A1)-Character(A2));
    else return GenArtinCompose(A1,'-',A2);
  end if;
end intrinsic;


intrinsic '*'(A1::ArtRep,A2:: ArtRep) -> ArtRep
{Tensor product of two Artin representations}
  if A1`K cmpeq A2`K
    then return A1`K!!(Character(A1)*Character(A2));
    else return GenArtinCompose(A1,'*',A2);
  end if;
end intrinsic;


intrinsic 'eq'(A1::ArtRep,A2:: ArtRep) -> BoolElt
{Test whether the two Artin representations are equal}
  if A1`K cmpeq A2`K
    then return Character(A1) eq Character(A2);
    else _,C1,C2:=GenArtinCompose(A1,'eq',A2: charsonly:=true);
         return C1 eq C2;
  end if;
end intrinsic;


intrinsic '!!'(K:: FldNum, ch:: AlgChtrElt) -> ArtRep
{The Artin representation of K with character ch}
  ComputeArtinRepresentations(~K);
  A:=ArtRepCreate();
  A`K:=K;
  CR:=Universe(K`artinrepdata`C);
  require IsCoercible(CR,ch): Sprintf("%o is not in %o",ch,CR);
  A`character:=CR!ch;
  return A;
end intrinsic;


intrinsic '!!'(K:: FldNum, ch:: SeqEnum) -> ArtRep
{The Artin representation of K with character ch}
  ComputeArtinRepresentations(~K);
  A:=ArtRepCreate();
  A`K:=K;
  CR:=Universe(K`artinrepdata`C);
  require IsCoercible(CR,ch): Sprintf("%o is not in %o",ch,CR);
  A`character:=CR!ch;
  return A;
end intrinsic;


intrinsic ArtinRepresentations(K::FldRat: Ramification:=false,
  FactorDiscriminant:=false, p0:="default", f:="default") -> SeqEnum
{Artin representations that factor through Q, i.e. just the
principal character of Gal(Qbar/Q)}
  return ArtinRepresentations(RationalsAsNumberField(): Ramification:=Ramification,
      FactorDiscriminant:=FactorDiscriminant, p0:=p0, f:=f);
end intrinsic;

/*
Ramification (default false)
specifies whether to compute the ramification data at all the bad primes.
FactorDiscriminant (default false)
specifies whether
*/

intrinsic ArtinRepresentations(K::FldNum: Ramification:=false,
  FactorDiscriminant:=false, p0:="default", f:="default") -> SeqEnum
{All irreducible Artin representations that factor through
the normal closure of K.}

  require Type(Ramification) eq BoolElt
    : "Ramification must be boolean";
  require (Type(F) eq BoolElt) or
    (Type(F) eq Tup) and (#F eq 5) and (Type(F[5]) eq BoolElt) and
    forall{j: j in [1..4] | IsCoercible(Z,F[j])}
    where F is FactorDiscriminant
    : "FactorDiscriminant must be boolean or a list <Trlim, Plim, Elim, Mlim, Proof>";
  require (p0 cmpeq "default") or                           // "default"
    (p0 cmpeq "galois") or                                  // "galois"
    IsCoercible(Z,p0) and (Z!p0 ge 2) and IsPrime(Z!p0) or  // 37
    (Type(p0) eq SeqEnum) and (#p0 eq 3) and                // [101,200,1]
    forall{x: x in p0 | IsCoercible(Z,x) and (x ge 0)}
    : "p0 must be \"default\", \"galois\", a prime number or a sequence [minp, attempts, mindeg]";
  require (f cmpeq "default") or
    (Type(f) eq RngUPolElt) and IsCoercible(PR(Z),f) and
    (LeadingCoefficient(f) eq 1) /* and IsIrreducible(f) */
    : "f must be \"default\" or a monic polynomial over the integers";

  ComputeArtinRepresentations(~K: Ramification:=Ramification,
      FactorDiscriminant:=FactorDiscriminant, p0:=p0, f:=f);
  return [K!!ch: ch in K`artinrepdata`C];
end intrinsic;


CharacterCharPoly:=func<c,g|
  CharacteristicPolynomialFromTraces([c(g^i): i in [1..c[1]]])
>;


intrinsic LSeries(A:: ArtRep: Precision:=0) -> LSer
{L-series of an Artin representation}
  ch:=Character(A);
  if IsZero(ch) then
    L:=LSeries(1: Precision:=Precision);
    L`parent:=A;
    return L;
  end if;
  require IsCharacter(ch): "Artin representation may not be virtual";
  if not IsIrreducible(ch) then
    dec:=Decomposition(A);
    dec:=&cat([[d[1]: i in [1..d[2]]]: d in dec]);
    L:=&*[LSeries(a: Precision:=Precision): a in dec];
    LSetPrecision(L,Precision);
    L`parent:=A;
    L`name:=A;
    return L;
  end if;
  data:=A`K`artinrepdata;
  G:=data`G;
  if ch eq PrincipalCharacter(G) then
    return RiemannZeta(:Precision:=Precision);
  end if;
  F:=data`Frob;
  bad:=RamifiedPrimes(A`K);

  C:=data`C;
  CC:=data`CC;
  charpolys:=[Eltseq(CharacterCharPoly(ch,g)): g in CC];

  function loc(p,d: Precision:=0)
    R:=PR(ComplexField(Precision));
    if not (p in bad) then
      return R!charpolys[F(p)];
    else
      return EulerFactor(A,p: R:=ComplexField(Precision));
    end if;
  end function;

  // gamma factors depend on the action of complex conjugation
  cctrace:=Z!ch[ComplexConjugation(A`K)];
  gammazeroes:=(Degree(ch)+cctrace)/2;   // dim of +1 eigenspace
  gammaones:=(Degree(ch)-cctrace)/2;     // dim of -1 eigenspace
  gamma:=[0: i in [1..gammazeroes]] cat [1: i in [1..gammaones]];

  // not going to compute the sign (root number)
  // except for orthogonal characters where it is 1 (Frohlich-Queyrut's thm)
  sign:=Schur(ch,2) eq 1 select 1 else 0;
  return LSeries(1,gamma,Conductor(A),loc: Parent:=A, Precision:=Precision, Sign:=sign);

end intrinsic;


/*
intrinsic LSeries(E:: CrvEll, A:: ArtRep: Precision:=0) -> LSer
{L-series L(E,A,s) of an elliptic curve E/Q twisted by an Artin representation A}
  require BaseField(E) cmpeq Q:
    "Elliptic curve must be defined over the rationals";
  commonramification:=PrimeDivisors(GCD(Conductor(E),Conductor(A)));
  require IsEmpty(commonramification):
    Sprintf("E and A are both ramified at p=%o: not yet implemented",
    Eltseq(commonramification)[1]);

  LE:=LSeries(E: Precision:=Precision);
  LA:=LSeries(A: Precision:=Precision);
  L:=TensorProduct(LE,LA,[]);
  if Precision ne 0 then LSetPrecision(L,Precision); end if;
  L`parent:="twist of " cat Sprint(E) cat " by " cat Sprint(A);
  return L;
end intrinsic;
*/


intrinsic PermutationCharacter(K::FldNum) -> ArtRep
{For a number field K with Galois closure F, compute the character of
C[Gal(Qbar/Q)/Gal(Qbar/K)] as an Artin representation. It is of degree [K:Q]
and is the character of the permutation representation of the absolute
Galois group on the embeddings of K in C}
  ComputeArtinRepresentations(~K);
  G:=K`artinrepdata`G;
  H:=Stabiliser(G,1);    //! assumes transitive
  return K!!PermutationCharacter(G,H);
end intrinsic;


intrinsic PermutationCharacter(K::FldRat) -> ArtRep
{The principal character of Gal(Qbar/Q) as an Artin representation}
  return ArtinRepresentations(Q)[1];
end intrinsic;


function RationalArg(x)
  C:=ComplexField();
  x:=C!x;
  if Abs(x) lt 10^(-Precision(C) div 2) then return 9; end if;
  a:=Arg(x)/2/Pi(C);
  n:=BestApproximation(a,10^(Precision(C) div 2));
  if n eq -1/2 then n:=1/2; end if;
  return n;
end function;

intrinsic // should go elsewhere
 EulerFactor(chi::GrpDrchElt,p::RngIntElt : R:=ComplexField()) -> RngUPolElt
{The Euler factor of a Dirichlet character at a prime}
 require IsPrime(p): "p must be prime"; P<t>:=PolynomialRing(R);
 return P!Polynomial([1,-chi(p)]); end intrinsic;

/*
function test_same(A,chi) K<zeta>:=CyclotomicField(Order(chi));
 P:=[p : p in PrimesUpTo(1000) | Valuation(2*Conductor(A),p) eq 0];
 for p in P do
  if EulerFactor(A,p : R:=K) ne EulerFactor(chi,p : R:=K) then return false;
  end if; end for; return true; end function;
// this might not be good, as perhaps EulerFactor computes extra stuff ??!
*/

function test_same(A,chi) ch:=Character(A);
 P:=[p : p in PrimesUpTo(1000) | Valuation(2*Conductor(chi),p) eq 0];
 K:=Field(A); N:=Conductor(chi); CC:=K`artinrepdata`CC;
 F:=K`artinrepdata`Frob; FClass:=Group(A)!1;
 o:=Order(chi); CF:=CyclotomicField(IsOdd(o) select 2*o else o);
 for p in P do if CF!ch(CC[F(p)]) ne CF!chi(p) then return false; end if;
  end for; return true; end function;

function TraceZero(x) E:=Eltseq(x);
 return [E[2]-E[1],E[3]-E[1],E[4]-E[1],-E[1]]; end function;

function TraceOne(x) E:=Eltseq(x); A:=[E[2]-E[1],E[3]-E[1],E[4]-E[1],-E[1]];
 A:=ChangeUniverse(A,Integers());
 if &+A mod 5 ne 4 then A:=[-a : a in A]; end if;
 q0:=-Integers()!((&+A+1)/5); return [q0] cat [e+q0 : e in A]; end function;

function get_field5(Q,p) assert &+Q eq -1; _<zeta5>:=CyclotomicField(5);
 v1:=&+[Q[i+1]*zeta5^(i) : i in [0..4]];
 v2:=&+[Q[i+1]*zeta5^(4*i) : i in [0..4]]; assert v1*v2 eq p;
 q0,q1,q2,q3,q4:=Explode(Q); l:=(p-1) div 5;
 s1:=q0; s2:=q0^2+q1*q2+q2*q4+q4*q3+q3*q1;
 s3:=3*q0^3+2*q0^2-4*l*q0+(q1+q4)*q1*q4+(q2+q3)*q2*q3+
     2*(q1^2*q2+q2^2*q4+q4^2*q3+q3^2*q1); // [s2,s3];
 P2:=-2*l; P3:=-(2*l+p*(1+s1))/5; P4:=(l^2-p*(1+2*s1+s2)/5)/5;
 P5:=(l^2-(p*(s1+s2)+(p*s3+l+1)/5)/5)/25; //[P2,P3,P4,P5];
 K:=NumberField(Polynomial([P5,P4,P3,P2,1,1])); // Tanner PLMS 1887
 assert #GaloisGroup(K) eq 5; return K; end function;

function get_quintic_fields(n) FAC:=Factorization(n); P:=[f[1] : f in FAC];
 assert n gt 1 and &and[p eq 5 or p mod 5 eq 1 : p in P];
 five:=5 in P; if five then Exclude(~P,5); end if; n:=&*P;
 if #P eq 0 then return [NumberField(Polynomial([-1,10,-5,-10,0,1]))]; end if;
 C5<zeta5>:=CyclotomicField(5); Z5:=Integers(C5); d:=#P; A:=[]; B:=[]; RET:=[];
 L:=[[x where _,x:=IsPrincipal(x[1]) : x in Factorization(p*Z5)] : p in P];
 for y in RSpace(Integers(4),d-1) do e:=ChangeUniverse(Eltseq(y),Integers());
  A cat:=[L[1][1]*&*[C5|L[i+1][e[i]+1] : i in [1..d-1]]]; end for;
 for a in A do TZ:=TraceZero(a); p1:=&+[TZ[i]*zeta5^i : i in [1..4]];
  p2:=&+[TZ[i]*zeta5^(2*i) : i in [1..4]]; B cat:=[p1*p2]; end for;
 for a in B do R:=[a,a*zeta5,a*zeta5^2,a*zeta5^3,a*zeta5^4];
  Fs:=[get_field5(TraceOne(r),n) : r in R];
  if five then
   RET cat:=[K : K in Fs | Valuation(Discriminant(Integers(K)),5) ne 0];
  else RET cat:=[K : K in Fs | Valuation(Discriminant(Integers(K)),5) eq 0];
   end if; end for;
  if five then assert &and[Discriminant(Integers(K)) eq n^4*5^8 : K in RET];
  else assert &and[Discriminant(Integers(K)) eq n^4 : K in RET]; end if;
 return RET; end function;

function DCKF_quintic(chi) assert Type(chi) eq GrpDrchElt;
 assert Order(chi) eq 5; N:=Conductor(chi); FIELDS:=get_quintic_fields(N);
 for K in FIELDS do vprint ArtRep,1: "Try",DefiningPolynomial(K);
  A:=ArtinRepresentations(K);
  for k in [2..5] do if test_same(A[k],chi) then return K; end if; end for;
  end for; error "No quintic field detected?!"; return false; end function;

// test_same is a bit silly, as I get the actual rep, not just the field,
// making the rest of the code (Tim's part) rather redundant

function get_quartic_fields(chi)
 D:=Conductor(chi^2); DISC:=Conductor(chi)^2*D; v2:=Valuation(DISC,2);
 if D mod 4 eq 0 then D:=D div 4; end if; RET:=[]; Bs:=[]; ARR:=[];
 F:=NumberField(Polynomial([-D,0,1]));
 A:=Conductor(chi) div D; _,A:=Valuation(A,2);
 P:=[f[1] : f in Factorization(D)];
 C4<i>:=CyclotomicField(4); Z4:=Integers(C4);
 if 2 in P then Exclude(~P,2); mult:=1+i; else mult:=1; end if; d:=#P;
 L:=[[x where _,x:=IsPrincipal(x[1]) : x in Factorization(p*Z4)] : p in P];
 for y in RSpace(Integers(2),d) do e:=ChangeUniverse(Eltseq(y),Integers());
  ARR cat:=[mult*&*[C4|L[i][e[i]+1] : i in [1..d]]]; end for;
 for a in ARR do E:=Eltseq(a);
  Bs cat:=[E[1] ge 0 select E[1] else -E[1]];
  Bs cat:=[E[2] ge 0 select E[2] else -E[2]]; end for;
 Bs:=SetToSequence(ChangeUniverse(Set(Bs),Integers())); Sort(~Bs);
 if v2 eq 6 then Bs:=[b : b in Bs | IsOdd(b)]; end if;
 if v2 eq 4 or v2 eq 0 then Bs:=[b : b in Bs | IsEven(b)]; end if;
 for B in Bs do e:=(A+B) mod 4;
  if v2 ge 6 or (v2 eq 4 and e eq 3) or (v2 eq 0 and e eq 1) then
   E:=ext<F|Polynomial([A*(D+B*Sqrt(F!D)),0,-1])>;
   K:=OptimizedRepresentation(AbsoluteField(E));
   assert Discriminant(Integers(K)) eq DISC;
   assert IsCyclic(GaloisGroup(K)); RET cat:=[K]; end if;
  A:=-A; e:=(A+B) mod 4;
  if v2 ge 6 or (v2 eq 4 and e eq 3) or (v2 eq 0 and e eq 1) then
   E:=ext<F|Polynomial([A*(D+B*Sqrt(F!D)),0,-1])>;
   K:=OptimizedRepresentation(AbsoluteField(E));
   assert Discriminant(Integers(K)) eq DISC;
   assert IsCyclic(GaloisGroup(K)); RET cat:=[K]; end if; end for;
 return RET; end function;

function DCKF_quartic(chi) assert Type(chi) eq GrpDrchElt;
 assert Order(chi) eq 4; FIELDS:=get_quartic_fields(chi);
 for K in FIELDS do vprint ArtRep,1: "Try",DefiningPolynomial(K);
  A:=ArtinRepresentations(K);
  for k in [2,4] do if test_same(A[k],chi) then return K; end if; end for;
  end for; error "No quartic field detected?!"; return false; end function;

function get_cubic_fields(n) FAC:=Factorization(n); P:=[f[1] : f in FAC];
 assert n gt 1 and &and[p eq 3 or p mod 3 eq 1 : p in P];
 three:=3 in P; if three then Exclude(~P,3); end if; n:=&*P;
 if #P eq 0 then return [NumberField(Polynomial([-1,-3,0,1]))];end if;
 C3<zeta3>:=CyclotomicField(3); Z3:=Integers(C3); d:=#P; A:=[];B:=[]; RET:=[];
 L:=[[x where _,x:=IsPrincipal(x[1]) : x in Factorization(p*Z3)] : p in P];
 for y in VectorSpace(GF(2),d-1) do e:=ChangeUniverse(Eltseq(y),Integers());
  A cat:=[L[1][1]*&*[C3|L[i+1][e[i]+1] : i in [1..d-1]]]; end for;
 for a in A do R:=[a,a*zeta3,a*zeta3^2];
  if three then B cat:=[x : x in R | Valuation(Eltseq(x)[2],3) eq 0];
  else B cat:=[x : x in R | Valuation(Eltseq(x)[2],3) gt 0]; end if; end for;
 M:=Matrix(3,3,[Rationals()|1,1,1, 1,-1,0, 0,-1,1]);
 for m in B do a,b:=Explode(Eltseq(m));
  if Valuation(b,3) gt 0 and Valuation(a+1,3) eq 0 then a:=-a; b:=-b; end if;
  v:=Vector([(n-1)/3,b/3,(a+1)/3]);
  f,g,h:=Explode(Eltseq(Solution(Transpose(M),v)));
  K:=NumberField(Polynomial([f*g-h^2,-(n-1)/3,1,1])); // Cayley, PLMS 1879
  assert #GaloisGroup(K) eq 3;
  assert Discriminant(Integers(K)) eq (&*P)^2*(three select 81 else 1);
  RET cat:=[K]; end for; return RET; end function;

function DCKF_cubic(chi) assert Type(chi) eq GrpDrchElt;
 assert Order(chi) eq 3; N:=Conductor(chi); FIELDS:=get_cubic_fields(N);
 for K in FIELDS do vprint ArtRep,1: "Try",DefiningPolynomial(K);
  A:=ArtinRepresentations(K); if test_same(A[2],chi) then return K; end if;
  if test_same(A[3],chi) then return K; end if; end for;
 error "No cubic field detected?!"; return false; end function;

function DCKF_quadratic(chi) assert Type(chi) eq GrpDrchElt;
 assert Order(chi) eq 2; N:=Conductor(chi);
 if Valuation(N,2) ne 3 then
  if N mod 4 eq 1 then return NumberField(Polynomial([-N,0,1])); end if;
  if N mod 4 eq 3 then return NumberField(Polynomial([N,0,1])); end if;
  if N mod 16 eq 4 then return NumberField(Polynomial([N,0,1])); end if;
  if N mod 16 eq 12 then return NumberField(Polynomial([-N,0,1])); end if;
  end if;
 if chi eq KroneckerCharacter(N) then return NumberField(Polynomial([-N,0,1]));
 else return NumberField(Polynomial([N,0,1])); end if; end function;

function claus_solution(chi) o:=Order(chi); Z:=Integers();
 C:=Conductor(chi); ZZ:=Integers(RationalsAsNumberField());
 cc:=DirichletCharacterOverNF(chi); LIM:=0; P:=[];
 R,mR:=RayClassGroup(C*ZZ,[1]); Q,mQ:=quo<R|o*R>;
 while #P lt 10 do LIM:=LIM+100;
  P:=[p : p in PrimesUpTo(LIM) | Valuation(C,p) eq 0]; end while;
 mult:=Characteristic(Parent('@'(Random(P),cc:Raw))) div o;
 h:=[<(p*ZZ)@@mR@mQ, GG![(Z!'@'(p,cc:Raw)) div mult]> :
     p in P | Valuation(C,p) eq 0] where GG := AbelianGroup([o]);
 F:=FreeAbelianGroup(#h);
 m1:=hom<F -> Parent(h[1][1]) | [h[i][1] : i in [1..#h]]>;
 m2:=hom<F -> Parent(h[1][2]) | [h[i][2] : i in [1..#h]]>;
 hh:=Inverse(m1)*m2; K:=Kernel(hh); q,mq:=quo<Q|K>;
 A:=AbelianExtension(Inverse(mq)*Inverse(mQ)*mR);
 return OptimisedRepresentation(AbsoluteField(NumberField(A))); end function;

function DirichletCharacterKernelField(chi) // replaced by class field theory
  chi:=AssociatedPrimitiveCharacter(chi); // MW
  return claus_solution(chi);
         //////////////////////////////////////////////////////
  O:=Order(chi);
  if O eq 2 then return DCKF_quadratic(chi); end if;
  if O eq 3 then return DCKF_cubic(chi); end if;
  if O eq 3 then return DCKF_quartic(chi); end if;
  if O eq 5 then return DCKF_quintic(chi); end if;
  if O eq 6 then K:=DCKF_quadratic(AssociatedPrimitiveCharacter(chi^3));
                 L:=DCKF_cubic(AssociatedPrimitiveCharacter(chi^2));
                 return ext<K|DefiningPolynomial(L)>; end if;
  N:=Norm(Conductor(chi)); // MW, ensures correct for GrpDrchNFElt
  A:=[a: a in [1..N] | GCD(a,N) eq 1];
  chivals:={chi(a): a in A};
  classes:=[[a: a in A| chi(a) eq v]: v in chivals];

  // find a prime which is 1 mod N
  p:="failed";
  for n:=N+1 to N^2+10^10 by N do
    if IsPrime(n) then p:=n; break; end if;
  end for;

  cfbound:=2*Max([Abs(c): c in Eltseq((PR(Z).1-#classes[1])^O)]);
  prec:=Floor(Log(p,cfbound))+2;

  // Find a primitive N-th root of unity in Z_p
  Zp:=pAdicRing(p,prec);

  zetanew:=Zp!((GF(p)!PrimitiveRoot(p))^((p-1) div N));
  repeat                            // Teichmuller lift via x->x^p
    zeta:=zetanew; zetanew:=zeta^p;
  until zeta eq zetanew;

  R<x>:=PR(Zp);
  for j in A do
    roots:=[&+[zeta^(j*a): a in V]: V in classes]; // better way if N is large?
    f:=&*[x-r: r in roots];
    f:=Polynomial([ZpToZ(c) : c in Eltseq(f)]);
    if IsIrreducible(f) then return NumberField(f); end if;
  end for;
  error "DirichletCharacterKernelField failed";
end function;


intrinsic ArtinRepresentation(chi::GrpDrchElt: field:="undefined") -> ArtRep
{Convert a Dirichlet character chi to an Artin representation A.
To avoid recomputation, the minimal field through which A factors may be
specified by the field parameter.}
  if IsTrivial(chi) then return ArtinRepresentations(Q)[1]; end if; // MW
  if field cmpeq "undefined" then
    K:=DirichletCharacterKernelField(chi);
  else
    K:=field;
  end if;
  error if Degree(K,Q) ne Order(chi),
    "The specified field is not the minimal field for chi";
  A:=ArtinRepresentations(K);
  p:=1;
  dK:=K`artinrepdata`disc;
  N:=Conductor(chi);
  o:=Order(chi); CF:=CyclotomicField(IsOdd(o) select 2*o else o);
  repeat
    p:=NextPrime(p);
    if (dK mod p eq 0) or (N mod p eq 0) then continue; end if;
    //vprint ArtRep,1: RationalArg(chi(p)),Denominator(RationalArg(chi(p))),Order(chi);
  until Denominator(RationalArg(chi(p))) eq Order(chi);
  F:=K`artinrepdata`Frob(p);
  F:=K`artinrepdata`CC[F];
  reps:=[a: a in A | CF!Character(a)(F) eq CF!chi(p)];
  error if #reps ne 1, "Failed to convert",chi,"to an Artin representation";
  return reps[1];
end intrinsic;

intrinsic ArtinRepresentation(chi::GrpDrchNFElt: field:="undefined") -> ArtRep
{Convert a Dirichlet character chi to an Artin representation A.
 To avoid recomputation, the minimal field through which A factors
 may be specified by the field parameter.} // copy of above, for GrpDrchNFElt
 require Degree(Parent(chi)`NF) eq 1: "Must be a Dirichlet character over Q";
 if IsTrivial(chi) then return ArtinRepresentations(Q)[1]; end if;
 return ArtinRepresentation(DirichletCharacterOverQ(chi) : field:=field);
 end intrinsic;

intrinsic DirichletCharacter(A::ArtRep) -> GrpDrchElt
{Convert an Artin representation to a Dirichlet character}
  ch:=Character(A);
  error if not IsLinear(ch), "The character of A must be linear";
  K:=Field(A);
  N:=Conductor(A);
  G:=DirichletGroup(N,CyclotomicField(Order(ch)));
  CC:=K`artinrepdata`CC;
  p:=1;
  dK:=K`artinrepdata`disc;
  F:=K`artinrepdata`Frob;
  FClass:=Group(A)!1;
  chis:=Elements(G);
  o:=Order(ch); CF:=CyclotomicField(IsOdd(o) select 2*o else o);
  repeat
    p:=NextPrime(p);
    if (dK mod p eq 0) or (N mod p eq 0) then continue; end if;
    FClass:=CC[F(p)];
    chFClass:=ch(FClass);
    chis:=[chi: chi in chis | CF!ch(FClass) eq CF!chi(p)];
    vprint ArtRep,1: "leftover",#chis;
    error if IsEmpty(chis), "Failed to find a dirichlet character";
  until #chis eq 1;
  return chis[1];
end intrinsic;


intrinsic EulerFactor(A:: ArtRep, p::RngIntElt: R:=ComplexField()) -> RngUPolElt
{Euler factor of A at p. The coefficient field (cyclotomic field or some
complex field) may be specified with the optional parameter R.}
  ch:=Character(A);
  data:=A`K`artinrepdata;
  G:=data`G;
  CC:=data`CC;
  PolR:=PR(R);

  // A unramified
  if not IsRamified(A,p) then
    return PolR!CharacterCharPoly(ch,CC[data`Frob(p)]);
  end if;

  // A ramified
  _,DSn,FPerm,ramgps,DtoG,_:=LocalData(A`K,p);
  chD:=LocalRepresentation(A`K,ch,p);
  frob:=DtoG[ClassMap(DSn)(FPerm)];

  I:=ramgps[1][1];  // inertia group as a permutation sgp of D
  if #I eq 1 then return PolR!CharacterCharPoly(ch,CC[frob]); end if;
  // pick up characters of D that enter ch|D and with full inertia
  // invariants (irreducible rep of D -> inertia invariants are
  // either trivial or the whole representation; we want char poly
  // of Frobenius just on inertia invariants)
  ichars:=[c: c in CharacterTable(DSn) | (InnerProduct(chD,c) ne 0) and
    (InnerProduct(Restriction(c,I),PrincipalCharacter(I)) ne 0)];
  return IsEmpty(ichars) select PolR!1 else     // do NOT coerce individual terms
    PolR!(&*[CharacterCharPoly(c,FPerm)^(Z!InnerProduct(chD,c)): c in ichars]);  
end intrinsic;

function DeterminantCharacter(G,c) d:=Integers()!c[1];
 ccgens:=[x[3]:x in ConjugacyClasses(G)];
 return [(-1)^d*Coefficient(CharacterCharPoly(c,g),d): g in ccgens];
 end function;

intrinsic Determinant(A::ArtRep) -> ArtRep
{Given an Artin representation, return its determinant as an Artin rep}
 if Degree(A) eq 1 then return A; end if; // turn off, for testing
 return Field(A)!!DeterminantCharacter(Group(A),Character(A)); end intrinsic;

intrinsic '^'(A::ArtRep,n::RngIntElt) -> ArtRep
{} require n ge 0 or Degree(A) eq 1: "Negative power of nonlinear ArtRep";
 if n eq 0 then return Field(A)!![1 : i in [1..#Eltseq(Character(A))]]; end if;
 if n eq 1 then return A; end if;
 if Degree(A) eq 1 then return Field(A)!!(Character(A)^n); end if;
 B:=A; R:=IsOdd(n) select B else B^0;
 while n gt 0 do n:=n div 2; B:=B*B;
  if IsOdd(n) then R:=R*B; end if; end while; return R; end intrinsic;

intrinsic '*'(n::RngIntElt,A::ArtRep) -> ArtRep
{} return Field(A)!!(n*Character(A)); end intrinsic;

intrinsic '*'(A::ArtRep,n::RngIntElt) -> ArtRep
{} return Field(A)!!(n*Character(A)); end intrinsic;
