freeze;

///////////////////////////////////////////////////////////////////
//                                                                 
// Heuristic formulae for cost of relation search in ClassGroup
// 
// Steve Donnelly, August 2012 (rewritten)
//
// Still a lot more refinements to be made...
//                                                                 
///////////////////////////////////////////////////////////////////

declare verbose ClassGroupExpected, 3;

// Volume of the hyperbolic region inside standard box in R^n, 
// consisting of (x_1, ..., x_n) in R^n with 0 <= x_i <= 1, 
// and product x_1 ... x_n <= b.  
// Here b must satisfy 0 < b < 1.

function hvol(n, b)
  l := Log(b);
  s := 1;
  for k := 1 to n-1 do 
    s +:= (-l)^k / Factorial(k);
  end for;
  return b * s;
end function;


// For a field of degree d and discriminant D, this is a heuristic formula 
// for the expected number of tries needed to find each relation.

// We assume we are dealing with random elements in a region
// such that their norm is guaranteed to be bounded by sqrtD.
// We infer the distribution of the norms, and then integrate 
// this against the probability of smoothness.

function class_group_expected_tries(d, D, bound)

  assert bound ge 1;

  R := RealField(10);
  sqrtD := R! Sqrt(Abs(D));
  bound := R! bound;
  logbound := Log(bound);

  if d le 1 or bound ge sqrtD then
    return R! 1;
  end if;

  // Dickman is too slow for large values, and
  // DickmanRho(15.0) = 7.6 x 10^-20
  // so we cut off the sum at 15

  function riemann_summands(n, l, u)
    f := R! (u/l)^(1/n);    // increment by factor f
    x := R! l;
    lf := Log(f)/logbound;
    lx := Log(x)/logbound;
    V0 := hvol(d, x/sqrtD);
    X := [R| ];
    S := [R| ];

    for i := 1 to n do
      x *:= f;
      lx +:= lf;
      if lx gt 15 then 
         break;
      end if;
      V := hvol(d, x/sqrtD);
      v := V - V0;
      V0 := V;
      s := DickmanRho(lx);
      Append(~X, x);
      Append(~S, s*v);
      vprintf ClassGroupExpected, 3 : 
        "x = %o, V = %o, v = %o, s = %o, sv = %o\n", x, V, v, s, s*v;
    end for;

    return X, S;
  end function;

  terms := 100;
  xmin  := R! 1/2; // must be < bound
  xmax  := sqrtD;

  // Find the relevant range, by sampling only a few values

  X, S := riemann_summands(10, xmin, xmax);

  smax, i := Max(S);
  s := 10^-4 * smax;
  i0 := Min([j : j in [1..i] | S[j] ge s]);
  i1 := Max([j : j in [i..#S] | S[j] ge s]);
  x0 := i0 eq 1 select xmin else X[i0-1]; 
  x1 := X[i1]; 

  // Sum over [x0 .. x1] with a reasonable number of values,
  // check that the central summands are reasonably equal
  r := 0.8;
  w := 5;

  while true do
     _, S := riemann_summands(terms, x0, x1);
     smax, i := Max(S[2..#S]);
     i +:= 1;
     // &+S, i, S[i], S[Max(1,i-w)], S[Min(#S,i+w)];

     if i-w le 0 and x0 - xmin gt 0.5 then
        x0 := Max(xmin, x0/2);
     elif i+w gt #S and xmax - x1 gt 0.5 then
        x1 := Min(xmax, x1*2);
     elif i-w gt 0 and S[i-w]/S[i] ge r and 
          i+w le #S and S[i+w]/S[i] ge r then
        break;
     end if;

     terms *:= 2;
     vprintf ClassGroupExpected, 2 : 
       "Repeating Riemann sum with %o terms\n", terms;
  end while;

  sum := (&+ S);

  if sum lt 10^-20 then
    // calculation is meaningless, due to truncating Dickman
    return R! 10^20;
  end if;

  return 1/sum;

end function;


// cost = B/Log(B) * expected trials per prime 
// time = const * cost
// The const is empirical: currently a reasonable ball park 
// for small degree fields is 10^4 trials per second.

function class_group_expected_cost(d, D : bound:=0, const:=0.0001);
  assert bound ge 1;
  tries := class_group_expected_tries(d, D, bound);
  // P = number of primes in factor base
  P := RealField(6)! bound/Log(bound); 
  cost := P * tries;
  return cost, Round(tries), const * cost;
end function;


///////////////////////////////////////////////////////////////////////////////
// USER INTRINSICS 
// (also called by the new ClassGroup)
///////////////////////////////////////////////////////////////////////////////

intrinsic ClassGroupExpectedNumberOfTrials
          (O::RngOrd, B::RngIntElt) -> RngIntElt

{The expected number of trials to find each smooth element, 
in the class group computation for the order O using the 
factor base bound B}

  require IsAbsoluteOrder(O) : 
     "ClassGroup not implemented for relative extensions";

  return Ceiling(class_group_expected_tries
                 (Degree(O), Discriminant(O), B));
end intrinsic;


intrinsic ClassGroupSuggestedBound
          (O::RngOrd) -> RngIntElt, RngIntElt

{A suggested value for the factor base norm bound, for the 
class group computation for the order O.  Also returns the 
ClassGroupExpectedNumberOfTrials with that bound}

  require IsAbsoluteOrder(O) : 
     "ClassGroup not implemented for relative extensions";

  d := Degree(O);
  D := Abs(Discriminant(O));
  logD := Ilog(2, D);

  // Do not return a ridiculously small bound
  Bmin := 50;

  // For easy fields, return fast
  // TO DO more
  if logD le 10 then
    return 50, 1;
  elif logD le 20 then
    return 100, 1;
  elif d eq 2 then
    if logD le 25 then
       return 100, 2;
    elif logD le 30 then
       return 100, 3;
    elif logD le 50 then
       return 100 + 30*(logD - 30), logD div 9;
    end if;
  end if;

  // There is usually a wide plateau of values of B
  // whose cost varies only within a small factor r.
  // When small enough, choose a value at the upper end,
  // otherwise keep B moderately small.

  B1 := 1000;    // prefer B1 <= B <= B2
  B2 := 2000;
  r1 := 1.3;     // increase cost by this factor to make B smaller
  r2 := 1.1;     // increase cost by this factor to make B larger
  eps := 0.995;  // refine until we know min_cost this accurately

  function test(B)
     E := class_group_expected_tries(d, D, B);
     vprintf ClassGroupExpected, 2 : "B = %o : E = %o\n", B, E;
     return <B, E, B/Log(B)*E>;
  end function;

  // We assume the cost behaves as a continuous function of c
  // with a single minimum, which we now try to locate.
  // In easy cases the optimal c is low, so start low.

  B0 := Max(Bmin, Round(0.2*Log(D)^2));
  T := [test(B0), test(2*B0)]; 

  // extend left until increase (take care with Bmin)
  while T[1,3] lt T[2,3] do
     if T[2,1] eq Bmin then
        return T[2,1], Max(1, Ceiling(T[2,2]));
     end if;
     B := Round(T[1,1] div 2);
     if B le Bmin and T[1,1] gt Bmin then
        B := Bmin;
     end if;
     Insert(~T, 1, test(B));
  end while;

  // extend right until increase
  while T[#T,3] lt T[#T-1,3] do
     Append(~T, test(Round(2*T[#T,1])) );
  end while;

  // refine near the min: include midpoints of the intervals 
  // on both sides of min, recursively
  while true do
     _, min := Min([t[3] : t in T]);
     assert min gt 1 and min lt #T;
     ldone := T[min,1] - T[min-1,1] lt 2 or T[min,3] ge eps*T[min-1,3];
     rdone := T[min+1,1] - T[min,1] lt 2 or T[min,3] ge eps*T[min+1,3];
     if ldone and rdone then
        break;
     end if;
     if not rdone then
        B := (T[min,1] + T[min+1,1]) div 2;
        Insert(~T, min+1, test(B));
     end if;
     if not ldone then
        B := (T[min-1,1] + T[min,1]) div 2;
        Insert(~T, min, test(B));
     end if;
  end while;

  while true do   //////////////  loop incresing r1  //////////////

    _, min := Min([t[3] : t in T]);

    if T[min,1] in [B1..B2] then
       a := min;

    elif T[min,1] gt B2 then

       // refine in the range where B > B2 and cost ~ r1*min_cost
       while true do
          a := Min([i : i in [1..#T] | T[i,1] ge B2 and T[i,3] lt r1*T[min,3]]);
          if a eq 1 or T[a-1,1] lt B2 then
             B := B2;
          else
             B := Max(B2, (T[a,1] + T[a-1,1]) div 2);
          end if;
          if B eq T[a,1] or a gt 1 and B eq T[a-1,1] 
             or a gt 1 and T[a,3] ge eps*T[a-1,3] 
          then
             break;
          end if;
          Insert(~T, a, test(B));
          _, min := Min([t[3] : t in T]);
       end while;
       
    elif T[min,1] lt B1 then

       // refine in the range where B < B1 and cost ~ r2*min_cost
       while true do
          a := Max([i : i in [1..#T] | T[i,1] le B1 and T[i,3] lt r2*T[min,3]]);
          if a eq #T or T[a+1,1] gt B1 then
             B := B1;
          else
             B := Min(B1, (T[a,1] + T[a+1,1]) div 2);
          end if;
          if B eq T[a,1] or a lt #T and B eq T[a+1,1] 
             or a lt #T and T[a,3] ge eps*T[a+1,3] 
          then
             break;
          end if;
          Insert(~T, a+1, test(B));
          _, min := Min([t[3] : t in T]);
       end while;

    end if;

    assert T eq Sort(T);

    if IsVerbose("ClassGroupExpected", 1) then
       for t in T do print t; end for;
    end if;
    vprintf ClassGroupExpected, 1 : "Best:\n%o\n", T[a];

    // Attempt to include difficulty of linear algebra:
    // go around several times to find out how much the
    // cost must increase in order to make B manageable

    N := Round(T[a,1]/Log(T[a,1])); // size of factor base

    if    N gt 1000 and r1 lt 1.5 then   r1 := 1.5;
    elif  N gt 2000 and r1 lt 2.0 then   r1 := 2.0;
    elif  N gt 3000 and r1 lt 3.0 then   r1 := 3.0;
    elif  N gt 3500 and r1 lt 4.0 then   r1 := 4.0;
    elif  N gt 4000 and r1 lt 5.0 then   r1 := 5.0;
    elif  N gt 5000 and r1 lt 10.0 then  r1 := 10.0;
    elif  N gt 6000 and r1 lt 15.0 then  r1 := 15.0;
    elif  N gt 8000 and r1 lt 20.0 then  r1 := 20.0;
    elif  N gt 10000 and r1 lt 50.0 then r1 := 50.0;
    else  
          break; 
    end if;

    vprintf ClassGroupExpected, 1 : 
           "B = %o is large : repeat with r1 = %o\n", T[a,1], RealField(4)!r1;

  end while;  //////////////  r1 loop  //////////////

  return T[a,1], Max(1, Ceiling(T[a,2]));

end intrinsic;

